#include "lion/CCA.h"
#include "include/CCA_Dcdc.h"
#include "include/CCA_EMS.h"
#include "include/CCA_Gw.h"
#include "include/CCA_Hvac.h"
#include "include/CCA_IccCalls.h"
#include "include/CCA_Internal.h"
#include "include/CCA_Pcs.h"
#include "include/CCA_Pdu.h"
#include "include/CCA_Pmu.h"
#include "include/CCA_Version.h"
#include "lion/rtos/RTOS_Sleep.h"
#include "lion/log/log.h"

LOG_MODULE_REGISTER(CCA, LOG_LEVEL_INF);

#define CCA_ENABLE_STATE_MACHINE_DEBUG 0
#define SECONDS_IN_10MS_LOOP(x)	  (unsigned int) ((x*1000)/10)

/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

typedef enum
{
    MQTT_STATE_ERROR = -1,
    MQTT_STATE_UNKNOWN = 0,
    MQTT_STATE_INIT,
    MQTT_STATE_CONNECTED,
    MQTT_STATE_WAIT_TIMEOUT
} aws_iot_state_t;

FileScop RTOS_Mutex_t cca_str_Mutex;
uint8_t aws_iot_run;
FileScop pthread_t cca_str_Thread;
aws_iot_state_t aws_iot_state = MQTT_STATE_UNKNOWN;

OpensslParams_t str_OpensslParams;
/* Each compilation unit must define the NetworkContext struct. */
struct NetworkContext
{
    OpensslParams_t *pParams;
};
NetworkContext_t str_NetworkContext = {0};
MQTTStatus_t str_MqttStatus;

/**
 * @brief Thread entry point for the CCA module.
 * 
 * @details Started by CCA_Start(), the thread will run until CCA_Stop() is called.
 */
FileScop void *cca_ThreadEntry(void *arg);

/*********************************************************************************************************************
* @cond                                              GLOBAL SCOPE                                             @endcond
*********************************************************************************************************************/

GlobScop void CCA_Init(void)
{
    RTOS_MutexInit(&cca_str_Mutex);
    aws_iot_run = false;

    //Initialize the all other CCA modules
    CCA_Icc_Init();
    CCA_Dcdc_Init();
    CCA_EMS_Init();
    CCA_Gw_Init();
    CCA_Hvac_Init();
    CCA_Pcs_Init();
    CCA_Pdu_Init();
    CCA_Pmu_Init();
    CCA_AWS_Init();

    //Initialize the internal version file buffer
    CCA_Version_Init();
}

GlobScop int CCA_Start()
{
    int int_rc = 0;
    RTOS_MutexLock(&cca_str_Mutex, RTOS_FOREVER);
    if (aws_iot_run)
    {
        int_rc = -EALREADY;
    }
    else
    {
        aws_iot_run = true;
        pthread_create(&cca_str_Thread, NULL, &cca_ThreadEntry, NULL);
    }
    RTOS_MutexUnlock(&cca_str_Mutex);
    return int_rc;
}

GlobScop int CCA_Stop(RTOS_Timeout_t tim_Timeout)
{
    int int_rc = 0;
    RTOS_MutexLock(&cca_str_Mutex, RTOS_FOREVER);
    if (aws_iot_run)
    {
        aws_iot_run = false;
        pthread_join(cca_str_Thread, NULL);
    }
    else
    {
        int_rc = -EALREADY;
    }
    RTOS_MutexUnlock(&cca_str_Mutex);
    return int_rc;
}

GlobScop int CCA_Provision()
{
    return -ENOTSUP;
}

/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

FileScop void *cca_ThreadEntry(void *arg)
{   
    CCA_Icc_RegisterCallbackHandlers();

    int s16_returnStatus = EXIT_SUCCESS;
    static int reconnectWaitTimer = 0;
    bool clientSessionPresent = false, brokerSessionPresent = false;
    struct timespec tp;
    uint8_t *arg_in = (uint8_t *)arg;
    time_t currentTime = 0;      // s
    time_t timeOfLastUpdate = 0; // s
    uint8_t retry = 3;
    void *_ret;

    if (s16_returnStatus == EXIT_SUCCESS)
    {
        while (aws_iot_run)
        {
            RTOS_Sleep(RTOS_MSEC(1));
            /* Get Current Time */
            (void)clock_gettime(CLOCK_REALTIME, &tp);
            currentTime = tp.tv_sec;

            /* Execute State Machine*/
            switch ((int)aws_iot_state)
            {

            case MQTT_STATE_INIT:
                timeOfLastUpdate = 0;
                retry = 3;
                /* Set the pParams member of the network context with desired transport. */
                str_NetworkContext.pParams = &str_OpensslParams;

                /* Seed pseudo random number generator (provided by ISO C standard library) for
                 * use by retry utils library when retrying failed network operations. */

                /* Get current time to seed pseudo random number generator. */
                // (void)clock_gettime(CLOCK_REALTIME, &tp); FIXME - Not needed anymore, being done above
                /* Seed pseudo random number generator with nanoseconds. */
                srand(tp.tv_nsec);

                /* Attempt to connect to the MQTT broker. If connection fails, retry after
                 * a timeout. Timeout value will be exponentially increased till the maximum
                 * attempts are reached or maximum timeout value is reached. The function
                 * returns EXIT_FAILURE if the TCP connection cannot be established to
                 * broker after configured number of attempts. */
                s16_returnStatus = CCA_AWS_ConnectToServerWithBackoffRetries(&str_NetworkContext, &str_MqttContext);

                if (s16_returnStatus == EXIT_FAILURE)
                {
                    /* Log error to indicate connection failure after all
                     * reconnect attempts are over. */
                    LOG_ERR("Failed to connect to MQTT broker %.*s.",
                              strlen(cca_sz_mqtt_endpoint),
                              cca_sz_mqtt_endpoint);
                }
                else
                {
                    LOG_INF("Setting up shadows................\n");
                    /* Set up shadow before going further */
                    while (retry != 0)
                    {
                        if (CCA_EMS_InitShadows() == EXIT_SUCCESS)
                        {
                            LOG_INF("Successfully initialized shadows");
                            break;
                        }
                        else
                        {
                            retry -= 1;
                            LOG_ERR("Shadow intialization error\n");
                            continue;
                        }
                    }
                    if (retry == 0)
                    {
                        aws_iot_state = MQTT_STATE_ERROR;
                        break;
                    }
                    else
                    {
                        aws_iot_state = MQTT_STATE_CONNECTED;
                    }
                    break;
                }
            case MQTT_STATE_CONNECTED:
                {
                /* Wait for acknowledgement packet (SUBACK) from the broker in response to the
                 * subscribe request. */
                s16_returnStatus = MQTT_ProcessLoop(&str_MqttContext, MQTT_PROCESS_LOOP_TIMEOUT_MS);

                if (s16_returnStatus != MQTTSuccess)
                {
                    str_MqttStatus = (MQTTStatus_t)s16_returnStatus;
                    s16_returnStatus = EXIT_FAILURE;
                    LOG_ERR("MQTT_ProcessLoop failed: Status=%s.",
                              MQTT_Status_strerror(str_MqttStatus));
                    aws_iot_state = MQTT_STATE_WAIT_TIMEOUT;
                    reconnectWaitTimer = 0;
                    break;
                }

                /* Send update to AWS every 5 minutes */
                if (currentTime - timeOfLastUpdate >= 5*60)
                {
                    CCA_Version_PublishVersionUpdate();
                    CCA_EMS_UpdateProductData();
                    CCA_EMS_UpdateComponentData();
                    timeOfLastUpdate = currentTime;
                }
                }
                break;
            case MQTT_STATE_WAIT_TIMEOUT:

                s16_returnStatus = MQTT_ProcessLoop(&str_MqttContext, MQTT_PROCESS_LOOP_TIMEOUT_MS);

                if (s16_returnStatus == MQTTSuccess)
                {
                    LOG_INF("MQTT Reconnection Success. Going to Connected State");
                    aws_iot_state = MQTT_STATE_CONNECTED;
                }
                else
                {
                    reconnectWaitTimer++;
                    if (reconnectWaitTimer > SECONDS_IN_10MS_LOOP(10))
                    {
                        LOG_INF("MQTT Reconnection Timeout. Restarting the task");
                        // aws_iot_state = MQTT_STATE_UNKNOWN;
                        // aws_iot_run = false;
                        aws_iot_state = MQTT_STATE_INIT;
                        reconnectWaitTimer = 0;
                    }
                }
                break;
            case MQTT_STATE_ERROR:
                LOG_ERR("Shadow creation failed, restarting state machine!\n");
                aws_iot_state = MQTT_STATE_INIT;
                // return _ret;
            default:
                aws_iot_state = MQTT_STATE_INIT;
                break;
            }

#if CCA_ENABLE_STATE_MACHINE_DEBUG
            // DEBUG PRINT for AWS/MQTT STATE MACHINE
            switch ((int)aws_iot_state)
            {
                case MQTT_STATE_INIT:
                    LOG_INF("\nMQTT_STATE: MQTT_STATE_INIT\n");
                    break;
                case MQTT_STATE_CONNECTED:
                    LOG_INF("\nMQTT_STATE: MQTT_STATE_CONNECTED\n");
                    break;
                case MQTT_STATE_WAIT_TIMEOUT:
                    LOG_INF("\nMQTT_STATE: MQTT_STATE_WAIT_TIMEOUT\n");
                    break;
                case MQTT_STATE_ERROR:
                    LOG_INF("\nMQTT_STATE: MQTT_STATE_ERROR\n");
                    break;
                case MQTT_STATE_UNKNOWN:
                    LOG_INF("\nMQTT_STATE: MQTT_STATE_UNKNOWN\n");
                    break;
                default:
                    LOG_INF("\nMQTT_STATE: Default\n");
                    break;
            }
#endif
        }
        /* End TLS session, then close TCP connection. */
        (void)Openssl_Disconnect(&str_NetworkContext);
    }
    CCA_Icc_UnregisterCallbackHandlers();
    return NULL;
}

//EOF
