#include "include/CCA_AWS.h"
#include "include/CCA_Internal.h"
#include "include/CCA_IccCalls.h"
#include "include/CCA_ReqHandler.h"
#include "include/CCA_Version.h"
#include "include/demo_config.h"
#include "lion/log/log.h"
#include <assert.h>

LOG_MODULE_REGISTER(CCA_aws, LOG_LEVEL_INF);

/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

/**
 * These configuration settings are required to run the mutual auth demo.
 * Throw compilation error if the below configs are not defined.
 */
#ifndef ROOT_CA_CERT_PATH
#error "Please define path to Root CA certificate of the MQTT broker(ROOT_CA_CERT_PATH) in demo_config.h."
#endif
#ifndef CLIENT_IDENTIFIER
#error "Please define a unique client identifier, CLIENT_IDENTIFIER, in demo_config.h."
#endif
/* The AWS IoT message broker requires either a set of client certificate/private key
 * or username/password to authenticate the client. */
#ifndef CLIENT_USERNAME
#ifndef CLIENT_CERT_PATH
#error "Please define path to client certificate(CLIENT_CERT_PATH) in demo_config.h."
#endif
#ifndef CLIENT_PRIVATE_KEY_PATH
#error "Please define path to client private key(CLIENT_PRIVATE_KEY_PATH) in demo_config.h."
#endif
#else

/* If a username is defined, a client password also would need to be defined for
 * client authentication. */
#ifndef CLIENT_PASSWORD
#error "Please define client password(CLIENT_PASSWORD) in demo_config.h for client authentication based on username/password."
#endif

/* AWS IoT MQTT broker port needs to be 443 for client authentication based on
 * username/password. */
#if AWS_MQTT_PORT != 443
#error "Broker port, AWS_MQTT_PORT, should be defined as 443 in demo_config.h for client authentication based on username/password."
#endif
#endif /* ifndef CLIENT_USERNAME */

/**
 * Provide default values for undefined configuration settings.
 */
#ifndef AWS_MQTT_PORT
#define AWS_MQTT_PORT (8883)
#endif

#ifndef NETWORK_BUFFER_SIZE
#define NETWORK_BUFFER_SIZE (1024U)
#endif

#ifndef OS_NAME
#define OS_NAME "Ubuntu"
#endif

#ifndef OS_VERSION
#define OS_VERSION "18.04 LTS"
#endif

#ifndef HARDWARE_PLATFORM_NAME
#define HARDWARE_PLATFORM_NAME "Posix"
#endif

/**
 * @brief Length of client identifier.
 */
#define CLIENT_IDENTIFIER_LENGTH ((uint16_t)(sizeof(CLIENT_IDENTIFIER) - 1))

/**
 * @brief ALPN (Application-Layer Protocol Negotiation) protocol name for AWS IoT MQTT.
 *
 * This will be used if the AWS_MQTT_PORT is configured as 443 for AWS IoT MQTT broker.
 * Please see more details about the ALPN protocol for AWS IoT MQTT endpoint
 * in the link below.
 * https://aws.amazon.com/blogs/iot/mqtt-with-tls-client-authentication-on-port-443-why-it-is-useful-and-how-it-works/
 *
 * @note OpenSSL requires that the protocol string passed to it for configuration be encoded
 * with the prefix of 8-bit length information of the string. Thus, the 14 byte (0x0e) length
 * information is prefixed to the string.
 */
#define AWS_IOT_MQTT_ALPN "\x0ex-amzn-mqtt-ca"

/**
 * @brief Length of ALPN protocol name.
 */
#define AWS_IOT_MQTT_ALPN_LENGTH ((uint16_t)(sizeof(AWS_IOT_MQTT_ALPN) - 1))

/**
 * @brief This is the ALPN (Application-Layer Protocol Negotiation) string
 * required by AWS IoT for password-based authentication using TCP port 443.
 *
 * @note OpenSSL requires that the protocol string passed to it for configuration
 * be encoded with the prefix of 8-bit length information of the string. Thus, the
 * 4 byte (0x04) length information is prefixed to the string.
 */
#define AWS_IOT_PASSWORD_ALPN "\x04mqtt"

/**
 * @brief Length of password ALPN.
 */
#define AWS_IOT_PASSWORD_ALPN_LENGTH ((uint16_t)(sizeof(AWS_IOT_PASSWORD_ALPN) - 1))

/**
 * @brief Length of client identifier.
 */
#define CLIENT_IDENTIFIER_LENGTH ((uint16_t)(sizeof(CLIENT_IDENTIFIER) - 1))

/**
 * @brief The maximum number of retries for connecting to server.
 */
#define CONNECTION_RETRY_MAX_ATTEMPTS (5U)

/**
 * @brief The maximum back-off delay (in milliseconds) for retrying connection to server.
 */
#define CONNECTION_RETRY_MAX_BACKOFF_DELAY_MS (5000U)

/**
 * @brief The base back-off delay (in milliseconds) to use for connection retry attempts.
 */
#define CONNECTION_RETRY_BACKOFF_BASE_MS (500U)

/**
 * @brief Timeout for receiving CONNACK packet in milli seconds.
 */
#define CONNACK_RECV_TIMEOUT_MS (1000U)



/**
 * @brief The maximum time interval in seconds which is allowed to elapse
 * between two Control Packets.
 *
 * It is the responsibility of the Client to ensure that the interval between
 * Control Packets being sent does not exceed the this Keep Alive value. In the
 * absence of sending any other Control Packets, the Client MUST send a
 * PINGREQ Packet.
 */
#define MQTT_KEEP_ALIVE_INTERVAL_SECONDS (60U)

/**
 * @brief Transport timeout in milliseconds for transport send and receive.
 */
#ifndef TRANSPORT_SEND_RECV_TIMEOUT_MS
#define TRANSPORT_SEND_RECV_TIMEOUT_MS (500)
#endif

/*-----------------------------------------------------------*/

/**
 * @brief Packet Identifier generated when Subscribe request was sent to the broker;
 * it is used to match received Subscribe ACK to the transmitted subscribe.
 */
FileScop uint16_t u16_LastSubscribePacketIdentifier = 0U;

/**
 * @brief Packet Identifier generated when Unsubscribe request was sent to the broker;
 * it is used to match received Unsubscribe ACK to the transmitted unsubscribe
 * request.
 */
FileScop uint16_t u16_LastUnsubscribePacketIdentifier = 0U;

/**
 * @brief The network buffer must remain valid for the lifetime of the MQTT context.
 */
FileScop uint8_t u8_Buffer[NETWORK_BUFFER_SIZE];

/**
 * @brief Status of latest Subscribe ACK;
 * it is updated every time the callback function processes a Subscribe ACK
 * and accounts for subscription to a single topic.
 */
FileScop MQTTSubAckStatus_t str_GlobalSubAckStatus = MQTTSubAckFailure;

/*-----------------------------------------------------------*/

/**
 * @brief The random number generator to use for exponential backoff with
 * jitter retry logic.
 *
 * @return The generated random number.
 */
FileScop inline uint32_t aws_GenerateRandomNumber() { return (rand());}

/**
 * @brief Callback function for the AWS MQTT client to handle incoming command messages.
 * 
 * @param ptr_str_Context MQTT context
 * @param ptr_str_PublishInfo Information about the incoming message
 */
FileScop void aws_CommandHandler(MQTTContext_t *ptr_str_Context, MQTTPublishInfo_t *ptr_str_PublishInfo);

/**
 * @brief Callback function for the AWS MQTT client to handle incoming delta messages.
 * 
 * @param ptr_str_Context MQTT context
 * @param ptr_str_PublishInfo Information about the incoming message
 */
FileScop void aws_UpdateDeltaHandler(MQTTContext_t *ptr_str_Context, MQTTPublishInfo_t *ptr_str_PublishInfo);

/**
 * @brief Callback function for the AWS MQTT client to handle incoming accepted messages.
 * 
 * @param ptr_str_Context MQTT context
 * @param ptr_str_PublishInfo Information about the incoming message
 */
FileScop void aws_UpdateAcceptedHandler(MQTTContext_t *ptr_str_Context, MQTTPublishInfo_t *ptr_str_PublishInfo);

/**
 * @brief Callback function for the AWS MQTT client to handle incoming rejected messages.
 * 
 * @param ptr_str_Context MQTT context
 * @param ptr_str_PublishInfo Information about the incoming message
 */
FileScop void aws_UpdateRejectedHandler(MQTTContext_t *ptr_str_Context, MQTTPublishInfo_t *ptr_str_PublishInfo);

/**
 * @brief The function to handle the incoming publishes.
 *
 * @param[in] ptr_str_PublishInfo Pointer to publish info of the incoming publish.
 */
FileScop void aws_HandleIncomingPublish(MQTTPublishInfo_t *ptr_str_PublishInfo);

/**
 * @brief The callback function that is registered with the MQTT connection
 * for dispatching incoming publish messages to their registered callbacks,
 * and handling incoming acks reported from the MQTT library.
 *
 * @param[in] ptr_str_MqttContext MQTT context pointer.
 * @param[in] ptr_str_PacketInfo Packet Info pointer for the incoming packet.
 * @param[in] ptr_str_DeserializedInfo Deserialized information from the incoming packet.
 * packet.
 */
FileScop void aws_CommonEventHandler(MQTTContext_t *ptr_str_MqttContext, MQTTPacketInfo_t *ptr_str_PacketInfo,
                                        MQTTDeserializedInfo_t *ptr_str_DeserializedInfo);

/**
 * @brief Sends an MQTT CONNECT packet over the already connected TCP socket.
 *
 * @param[in] ptr_str_MqttContext MQTT context pointer.
 * @param[in] ptr_str_NetworkContext Pointer to the network context created using Openssl_Connect.
 * @param[in] bol_CreateCleanSession Creates a new MQTT session if true.
 * If false, tries to establish the existing session if there was session
 * already present in broker.
 * @param[out] ptr_bol_SessionPresent Session was already present in the broker or not.
 * Session present response is obtained from the CONNACK from broker.
 *
 * @return EXIT_SUCCESS if an MQTT session is established;
 * EXIT_FAILURE otherwise.
 */
FileScop int aws_EstablishMqttSession(MQTTContext_t *ptr_str_MqttContext, NetworkContext_t *ptr_str_NetworkContext, 
                                        bool bol_CreateCleanSession, bool *ptr_bol_SessionPresent);

/**
 * @brief Subscribes to the passed topic filter by sending an MQTT SUBSCRIBE
 * packet and waiting for a SUBACK acknowledgement response from the broker.
 *
 * @param[in] ptr_str_MqttContext MQTT context pointer.
 * @param[in] sz_TopicFilter The topic filter to subscribe to.
 * @param[in] u16_TopicFilterLength The length of the topic filter.
 *
 * @return EXIT_SUCCESS if SUBSCRIBE was successfully sent;
 * EXIT_FAILURE otherwise.
 */
FileScop int aws_SubscribeToTopic(MQTTContext_t *ptr_str_MqttContext, const char *sz_TopicFilter, 
                                    uint16_t u16_TopicFilterLength);

/**
 * @brief The application callback function for getting the incoming publish
 * and incoming acks reported from MQTT library.
 *
 * @param[in] ptr_str_MqttContext MQTT context pointer.
 * @param[in] ptr_str_PacketInfo Packet Info pointer for the incoming packet.
 * @param[in] ptr_str_DeserializedInfo Deserialized information from the incoming packet.
 */
FileScop void aws_EventCallback(MQTTContext_t *ptr_str_MqttContext, MQTTPacketInfo_t *ptr_str_PacketInfo,
                            MQTTDeserializedInfo_t *ptr_str_DeserializedInfo);

/**
 * @brief Function to update variable str_GlobalSubAckStatus with status
 * information from Subscribe ACK. Called by aws_EventCallback after processing
 * incoming subscribe echo.
 *
 * @param[in] ptr_str_PacketInfo Packet Info pointer for the incoming packet.
 */
FileScop void aws_UpdateSubAckStatus(MQTTPacketInfo_t *ptr_str_PacketInfo);

/*********************************************************************************************************************
* @cond                                              MODULE SCOPE                                             @endcond
*********************************************************************************************************************/

ModScop MQTTContext_t str_MqttContext = {0};

ModScop void CCA_AWS_Init(void)
{
    //Set Default AWS Environemnt
    strcpy(cca_sz_mqtt_endpoint, CCA_sz_PROD_ENDPOINT);
    // Pull AWS Configurations
    FILE *fd_AwsConfig = fopen(CCA_sz_CONFIG_FILE_PATH, "r");
    if (fd_AwsConfig != NULL)
    {
        fseek(fd_AwsConfig, 0, SEEK_END);
        uint64_t u64_awsConfigFileSize = ftell(fd_AwsConfig);
        rewind(fd_AwsConfig);
        char sz_awsConfigBuffer[u64_awsConfigFileSize + 1];
        size_t siz_read = fread(sz_awsConfigBuffer, u64_awsConfigFileSize, 1, fd_AwsConfig);
        (void)siz_read; //Suppress unused variable warning. (void)fread is not enough to suppress the warning.
        ArduinoJson::StaticJsonDocument<CCA_u16_CONFIG_JSON_DOC_SIZE> jd_AwsConfig;
        ArduinoJson::deserializeJson(jd_AwsConfig, sz_awsConfigBuffer);
        // Set Endpoint based on Development Environment
        if (!jd_AwsConfig["aws"]["env"].isNull())
        {
            ArduinoJson::JsonVariant jv_env = jd_AwsConfig["aws"]["env"];
            if (strcmp(jv_env.as<const char*>(), "dev") == EXIT_SUCCESS)
            {
                strcpy(cca_sz_mqtt_endpoint, CCA_sz_DEV_ENDPOINT);
            }
            else if (strcmp(jv_env.as<const char*>(), "stage") == EXIT_SUCCESS)
            {
                strcpy(cca_sz_mqtt_endpoint, CCA_sz_STAGE_ENDPOINT);
            }
            else {  // default to Prod
                strcpy(cca_sz_mqtt_endpoint, CCA_sz_PROD_ENDPOINT);
            }
        }
        // Pull component configurations
        ArduinoJson::JsonArray ja_Pcs = jd_AwsConfig[CCA_sz_UNIT_TYPE_PCS];
        for (ArduinoJson::JsonVariant jv_PcsObject : ja_Pcs) 
        {
            if (jv_PcsObject.containsKey("id") && jv_PcsObject.containsKey("uuid"))
            {
                uint8_t u8_id = jv_PcsObject["id"].as<uint8_t>();
                if (u8_id < CCA_u8_MAX_COMPONENT_QTY) 
                {
                    strcpy(cca_arr_pcs_component_uuids[u8_id], jv_PcsObject["uuid"].as<const char*>());
                }
                else 
                {
                    LOG_INF("Initialization Error: Invalid AWS Config File Format - Too many Components for type '%s'",
                            CCA_sz_UNIT_TYPE_PCS);
                }
            }
            else 
            {
                LOG_INF("Initialization Error: Invalid AWS Config File Format");
            }
        }
        ArduinoJson::JsonArray ja_Pdu = jd_AwsConfig[CCA_sz_UNIT_TYPE_PDU];
        for (ArduinoJson::JsonVariant jv_PduObject : ja_Pdu) 
        {
            if (jv_PduObject.containsKey("id") && jv_PduObject.containsKey("uuid"))
            {
                uint8_t u8_id = jv_PduObject["id"].as<uint8_t>();
                if (u8_id < CCA_u8_MAX_COMPONENT_QTY) 
                {
                    strcpy(cca_arr_pdu_component_uuids[u8_id], jv_PduObject["uuid"].as<const char*>());
                }
                else 
                {
                    LOG_INF("Initialization Error: Invalid AWS Config File Format - Too many Components for type '%s'",
                            CCA_sz_UNIT_TYPE_PDU);
                }
            }
            else 
            {
                LOG_INF("Initialization Error: Invalid AWS Config File Format");
            }
        }
        ArduinoJson::JsonArray ja_Dcdc = jd_AwsConfig[CCA_sz_UNIT_TYPE_DCDC];
        for (ArduinoJson::JsonVariant jv_DcdcObject : ja_Dcdc) 
        {
            if (jv_DcdcObject.containsKey("id") && jv_DcdcObject.containsKey("uuid"))
            {
                uint8_t u8_id = jv_DcdcObject["id"].as<uint8_t>();
                if (u8_id < CCA_u8_MAX_COMPONENT_QTY) 
                {
                    strcpy(cca_arr_dcdc_component_uuids[u8_id], jv_DcdcObject["uuid"].as<const char*>());
                }
                else 
                {
                    LOG_INF("Initialization Error: Invalid AWS Config File Format - Too many Components for type '%s'",
                            CCA_sz_UNIT_TYPE_DCDC);
                }
            }
            else 
            {
                LOG_INF("Initialization Error: Invalid AWS Config File Format");
            }
        }
        ArduinoJson::JsonArray ja_Pmu = jd_AwsConfig[CCA_sz_UNIT_TYPE_PMU];
        for (ArduinoJson::JsonVariant jv_PmuObject : ja_Pmu) 
        {
            if (jv_PmuObject.containsKey("id") && jv_PmuObject.containsKey("uuid"))
            {
                uint8_t u8_id = jv_PmuObject["id"].as<uint8_t>();
                if (u8_id < CCA_u8_MAX_COMPONENT_QTY) 
                {
                    strcpy(cca_arr_pmu_component_uuids[u8_id], jv_PmuObject["uuid"].as<const char*>());
                }
                else 
                {
                    LOG_INF("Initialization Error: Invalid AWS Config File Format - Too many Components for type '%s'",
                            CCA_sz_UNIT_TYPE_PMU);
                }
            }
            else 
            {
                LOG_INF("Initialization Error: Invalid AWS Config File Format");
            }
        }
        ArduinoJson::JsonArray ja_Hvac = jd_AwsConfig[CCA_sz_UNIT_TYPE_HVAC];
        for (ArduinoJson::JsonVariant jv_HvacObject : ja_Hvac) 
        {
            if (jv_HvacObject.containsKey("id") && jv_HvacObject.containsKey("uuid"))
            {
                uint8_t u8_id = jv_HvacObject["id"].as<uint8_t>();
                if (u8_id < CCA_u8_MAX_COMPONENT_QTY) 
                {
                    strcpy(cca_arr_hvac_component_uuids[u8_id], jv_HvacObject["uuid"].as<const char*>());
                }
                else 
                {
                    LOG_INF("Initialization Error: Invalid AWS Config File Format - Too many Components for type '%s'",
                            CCA_sz_UNIT_TYPE_HVAC);
                }
            }
            else 
            {
                LOG_INF("Initialization Error: Invalid AWS Config File Format");
            }
        }
        ArduinoJson::JsonObject jo_Ems = jd_AwsConfig[CCA_sz_UNIT_TYPE_EMS];
        if (jo_Ems.containsKey("uuid") && jo_Ems.containsKey("name"))
        {
            strcpy(cca_sz_thing_name, jo_Ems["uuid"].as<const char*>());
            strcpy(cca_sz_product_name, jo_Ems["name"].as<const char*>());
            // Set All shadow topic names for PCS Settings
            strcat(cca_sz_pcs_update_topic, cca_sz_thing_name);
            strcat(cca_sz_pcs_update_topic, CCA_sz_PCS_UPDATE_TOPIC);
            strcat(cca_sz_pcs_update_delta_topic, cca_sz_thing_name);
            strcat(cca_sz_pcs_update_delta_topic, CCA_sz_PCS_UPDATE_DELTA_TOPIC);
            strcat(cca_sz_pcs_update_accepted_topic, cca_sz_thing_name);
            strcat(cca_sz_pcs_update_accepted_topic, CCA_sz_PCS_UPDATE_ACCEPTED_TOPIC);
            strcat(cca_sz_pcs_update_rejected_topic, cca_sz_thing_name);
            strcat(cca_sz_pcs_update_rejected_topic, CCA_sz_PCS_UPDATE_REJECTED_TOPIC);
            // Set All shadow topic names for DCDC Settings
            strcat(cca_sz_dcdc_update_topic, cca_sz_thing_name);
            strcat(cca_sz_dcdc_update_topic, CCA_sz_DCDC_UPDATE_TOPIC);
            strcat(cca_sz_dcdc_update_delta_topic, cca_sz_thing_name);
            strcat(cca_sz_dcdc_update_delta_topic, CCA_sz_DCDC_UPDATE_DELTA_TOPIC);
            strcat(cca_sz_dcdc_update_accepted_topic, cca_sz_thing_name);
            strcat(cca_sz_dcdc_update_accepted_topic, CCA_sz_DCDC_UPDATE_ACCEPTED_TOPIC);
            strcat(cca_sz_dcdc_update_rejected_topic, cca_sz_thing_name);
            strcat(cca_sz_dcdc_update_rejected_topic, CCA_sz_DCDC_UPDATE_REJECTED_TOPIC);
            // Set All shadow topic names for EMS Settings
            strcat(cca_sz_ems_update_topic, cca_sz_thing_name);
            strcat(cca_sz_ems_update_topic, CCA_sz_EMS_UPDATE_TOPIC);
            strcat(cca_sz_ems_update_delta_topic, cca_sz_thing_name);
            strcat(cca_sz_ems_update_delta_topic, CCA_sz_EMS_UPDATE_DELTA_TOPIC);
            strcat(cca_sz_ems_update_accepted_topic, cca_sz_thing_name);
            strcat(cca_sz_ems_update_accepted_topic, CCA_sz_EMS_UPDATE_ACCEPTED_TOPIC);
            strcat(cca_sz_ems_update_rejected_topic, cca_sz_thing_name);
            strcat(cca_sz_ems_update_rejected_topic, CCA_sz_EMS_UPDATE_REJECTED_TOPIC);
            // Set All shadow topic names for PMU Settings
            strcat(cca_sz_pmu_update_topic, cca_sz_thing_name);
            strcat(cca_sz_pmu_update_topic, CCA_sz_PMU_UPDATE_TOPIC);
            strcat(cca_sz_pmu_update_delta_topic, cca_sz_thing_name);
            strcat(cca_sz_pmu_update_delta_topic, CCA_sz_PMU_UPDATE_DELTA_TOPIC);
            strcat(cca_sz_pmu_update_accepted_topic, cca_sz_thing_name);
            strcat(cca_sz_pmu_update_accepted_topic, CCA_sz_PMU_UPDATE_ACCEPTED_TOPIC);
            strcat(cca_sz_pmu_update_rejected_topic, cca_sz_thing_name);
            strcat(cca_sz_pmu_update_rejected_topic, CCA_sz_PMU_UPDATE_REJECTED_TOPIC);
            // Set All shadow topic names for HVAC Settings
            strcat(cca_sz_hvac_update_topic, cca_sz_thing_name);
            strcat(cca_sz_hvac_update_topic, CCA_sz_HVAC_UPDATE_TOPIC);
            strcat(cca_sz_hvac_update_delta_topic, cca_sz_thing_name);
            strcat(cca_sz_hvac_update_delta_topic, CCA_sz_HVAC_UPDATE_DELTA_TOPIC);
            strcat(cca_sz_hvac_update_accepted_topic, cca_sz_thing_name);
            strcat(cca_sz_hvac_update_accepted_topic, CCA_sz_HVAC_UPDATE_ACCEPTED_TOPIC);
            strcat(cca_sz_hvac_update_rejected_topic, cca_sz_thing_name);
            strcat(cca_sz_hvac_update_rejected_topic, CCA_sz_HVAC_UPDATE_REJECTED_TOPIC);
            // Set all topic names for version info
            strcat(cca_sz_version_update_topic, cca_sz_thing_name);
            strcat(cca_sz_version_update_topic, CCA_sz_VERSION_UPDATE_TOPIC);
            strcat(cca_sz_version_update_delta_topic, cca_sz_thing_name);
            strcat(cca_sz_version_update_delta_topic, CCA_sz_VERSION_UPDATE_DELTA_TOPIC);
            strcat(cca_sz_version_update_accepted_topic, cca_sz_thing_name);
            strcat(cca_sz_version_update_accepted_topic, CCA_sz_VERSION_UPDATE_ACCEPTED_TOPIC);
            strcat(cca_sz_version_update_rejected_topic, cca_sz_thing_name);
            strcat(cca_sz_version_update_rejected_topic, CCA_sz_VERSION_UPDATE_REJECTED_TOPIC);
            // Set All topic names for Command Request/Response 
            strcat(cca_sz_command_request_topic, cca_sz_thing_name);
            strcat(cca_sz_command_response_topic, cca_sz_thing_name);
            // Set topic name for Alerts
            strcat(cca_sz_alerts_topic, cca_sz_thing_name);
        }
        else 
        {
            LOG_INF("Initialization Error: Invalid AWS Config File Format - Missing uuid or name for '%s'",
                    CCA_sz_UNIT_TYPE_EMS);
        }
    }
    else 
    {
        LOG_INF("Initialization Error: Failed to open AWS Config File");
    }
}

ModScop int CCA_AWS_ConnectToServerWithBackoffRetries(NetworkContext_t *ptr_str_NetworkContext,
                                                        MQTTContext_t *ptr_str_MqttContext)
{
    int int_rc = EXIT_FAILURE;
    BackoffAlgorithmStatus_t str_BackoffAlgStatus = BackoffAlgorithmSuccess;
    OpensslStatus_t str_OpensslStatus = OPENSSL_SUCCESS;
    BackoffAlgorithmContext_t str_ReconnectParams;
    ServerInfo_t str_ServerInfo;
    OpensslCredentials_t str_OpensslCredentials;
    uint16_t u16_NextRetryBackOff;
    bool bol_SessionPresent = false;
    /* Initialize information to connect to the MQTT broker. */
    str_ServerInfo.pHostName = cca_sz_mqtt_endpoint;
    str_ServerInfo.hostNameLength = strlen(cca_sz_mqtt_endpoint);
    str_ServerInfo.port = AWS_MQTT_PORT;

    /* Initialize credentials for establishing TLS session. */
    memset(&str_OpensslCredentials, 0, sizeof(OpensslCredentials_t));
    str_OpensslCredentials.pRootCaPath = ROOT_CA_CERT_PATH;
    str_OpensslCredentials.pClientCertPath = CLIENT_CERT_PATH;
    str_OpensslCredentials.pPrivateKeyPath = CLIENT_PRIVATE_KEY_PATH;

    /* AWS IoT requires devices to send the Server Name Indication (SNI)
     * extension to the Transport Layer Security (TLS) protocol and provide
     * the complete endpoint address in the host_name field. Details about
     * SNI for AWS IoT can be found in the link below.
     * https://docs.aws.amazon.com/iot/latest/developerguide/transport-security.html */
    str_OpensslCredentials.sniHostName = cca_sz_mqtt_endpoint;

    if (AWS_MQTT_PORT == 443)
    {
/* Pass the ALPN protocol name depending on the port being used.
 * Please see more details about the ALPN protocol for the AWS IoT MQTT
 * endpoint in the link below.
 * https://aws.amazon.com/blogs/iot/mqtt-with-tls-client-authentication-on-port-443-why-it-is-useful-and-how-it-works/
 *
 * For username and password based authentication in AWS IoT,
 * #AWS_IOT_PASSWORD_ALPN is used. More details can be found in the
 * link below.
 * https://docs.aws.amazon.com/iot/latest/developerguide/custom-authentication.html
 */
#ifdef CLIENT_USERNAME
        str_OpensslCredentials.pAlpnProtos = AWS_IOT_PASSWORD_ALPN;
        str_OpensslCredentials.alpnProtosLen = AWS_IOT_PASSWORD_ALPN_LENGTH;
#else
        str_OpensslCredentials.pAlpnProtos = AWS_IOT_MQTT_ALPN;
        str_OpensslCredentials.alpnProtosLen = AWS_IOT_MQTT_ALPN_LENGTH;
#endif
    }

    /* Initialize reconnect attempts and interval */
    BackoffAlgorithm_InitializeParams(&str_ReconnectParams,
                                      CONNECTION_RETRY_BACKOFF_BASE_MS,
                                      CONNECTION_RETRY_MAX_BACKOFF_DELAY_MS,
                                      CONNECTION_RETRY_MAX_ATTEMPTS);

    /* Attempt to connect to MQTT broker. If connection fails, retry after
     * a timeout. Timeout value will exponentially increase until maximum
     * attempts are reached.
     */
    do
    {
        /* Establish a TLS session with the MQTT broker. This example connects
         * to the MQTT broker as specified in CCA_sz_mqtt_endpoint and AWS_MQTT_PORT
         * at the demo config header. */
        LOG_INF("Establishing a TLS session to %.*s:%d.",
                 strlen(cca_sz_mqtt_endpoint),
                 cca_sz_mqtt_endpoint,
                 AWS_MQTT_PORT);
        str_OpensslStatus = Openssl_Connect(ptr_str_NetworkContext,
                                        &str_ServerInfo,
                                        &str_OpensslCredentials,
                                        TRANSPORT_SEND_RECV_TIMEOUT_MS,
                                        TRANSPORT_SEND_RECV_TIMEOUT_MS);

        if (str_OpensslStatus == OPENSSL_SUCCESS)
        {

            /* Sends an MQTT Connect packet using the established TLS session,
             * then waits for connection acknowledgment (CONNACK) packet. */
            int_rc = aws_EstablishMqttSession(ptr_str_MqttContext,
                                                ptr_str_NetworkContext,
                                                true, /* clean session */
                                                &bol_SessionPresent);

            if (int_rc == EXIT_FAILURE)
            {
                /* End TLS session, then close TCP connection. */
                (void)Openssl_Disconnect(ptr_str_NetworkContext);
            }
        }

        if (int_rc == EXIT_FAILURE)
        {
            /* Generate a random number and get back-off value (in milliseconds) for the next connection retry. */
            str_BackoffAlgStatus = BackoffAlgorithm_GetNextBackoff(&str_ReconnectParams, aws_GenerateRandomNumber(), 
                                                                    &u16_NextRetryBackOff);

            if (str_BackoffAlgStatus == BackoffAlgorithmRetriesExhausted)
            {
                LOG_ERR("Connection to the broker failed, all attempts exhausted.");
                int_rc = EXIT_FAILURE;
            }
            else if (str_BackoffAlgStatus == BackoffAlgorithmSuccess)
            {
                LOG_WRN("Connection to the broker failed. Retrying connection after %hu ms backoff.",
                        (unsigned short)u16_NextRetryBackOff);
                Clock_SleepMs(u16_NextRetryBackOff);
            }
        }
    } while ((int_rc == EXIT_FAILURE) && (str_BackoffAlgStatus == BackoffAlgorithmSuccess));

    return int_rc;
}

ModScop int CCA_AWS_SubscribeToAndRegisterTopicFilter(MQTTContext_t *ptr_str_Context, const char *sz_TopicFilter,
                                             uint16_t u16_TopicFilterLength, SubscriptionManagerCallback_t fn_Callback)
{
    int int_rc = EXIT_SUCCESS;
    SubscriptionManagerStatus_t str_ManagerStatus = (SubscriptionManagerStatus_t)0u;

    /* Register the topic filter and its callback with subscription manager.
     * On an incoming PUBLISH message whose topic name that matches the topic filter
     * being registered, its callback will be invoked. */
    str_ManagerStatus = SubscriptionManager_RegisterCallback(sz_TopicFilter, u16_TopicFilterLength, fn_Callback);

    if (str_ManagerStatus != SUBSCRIPTION_MANAGER_SUCCESS)
    {
        int_rc = EXIT_FAILURE;
    }
    else
    {
        LOG_INF("Subscribing to the MQTT topic %.*s.", u16_TopicFilterLength, sz_TopicFilter);

        int_rc = aws_SubscribeToTopic(ptr_str_Context, sz_TopicFilter, u16_TopicFilterLength);
    }

    if (int_rc != EXIT_SUCCESS)
    {
        /* Remove the registered callback for the temperature topic filter as
         * the subscription operation for the topic filter did not succeed. */
        (void)SubscriptionManager_RemoveCallback(sz_TopicFilter,
                                                 u16_TopicFilterLength);
    }

    return int_rc;
}

ModScop int CCA_AWS_PublishToTopic(MQTTContext_t *ptr_str_MqttContext, const char *sz_Topic, uint16_t u16_TopicLength,
                                    const char *sz_Message, uint16_t u16_MessageLength)
{
    int int_rc = EXIT_SUCCESS;

    MQTTStatus_t str_MqttStatus = MQTTSuccess;
    MQTTPublishInfo_t str_PublishInfo;
    uint16_t u16_PubPacketId = MQTT_PACKET_ID_INVALID;

    assert(ptr_str_MqttContext != NULL);

    (void)memset(&str_PublishInfo, 0x00, sizeof(MQTTPublishInfo_t));

    if (int_rc == EXIT_FAILURE)
    {
        LOG_ERR("Unable to find a free spot for outgoing PUBLISH message.\n\n");
    }
    else
    {
        /* This example publishes with QOS1. */
        str_PublishInfo.qos = MQTTQoS0;
        str_PublishInfo.pTopicName = sz_Topic;
        str_PublishInfo.topicNameLength = u16_TopicLength;
        str_PublishInfo.pPayload = sz_Message;
        str_PublishInfo.payloadLength = u16_MessageLength;

        /* Get a new packet ID for the publish. */
        u16_PubPacketId = MQTT_GetPacketId(ptr_str_MqttContext);

        /* Send PUBLISH packet. */
        str_MqttStatus = MQTT_Publish(ptr_str_MqttContext, &str_PublishInfo, u16_PubPacketId);

        if (str_MqttStatus != MQTTSuccess)
        {
            LOG_ERR("Failed to send PUBLISH packet to broker with error = %u.", str_MqttStatus);
            int_rc = EXIT_FAILURE;
        }
        else
        {
            LOG_INF("PUBLISH sent for topic %.*s to broker with packet ID %u.\n\n", u16_TopicLength, sz_Topic, u16_PubPacketId);
        }
    }

    return int_rc;
}

ModScop int CCA_AWS_SubscribeToShadowSubtopics(const char* sz_ShadowName, const char *sz_UpdateDeltaTopic, 
                                                const char *sz_UpdateAcceptedTopic, const char *sz_UpdateRejectedTopic)
{
    int int_rc = EXIT_SUCCESS;
    // Subscribe to neccessary topics
    if (CCA_AWS_SubscribeToAndRegisterTopicFilter(&str_MqttContext, sz_UpdateDeltaTopic, strlen(sz_UpdateDeltaTopic), 
                                                    aws_UpdateDeltaHandler) == EXIT_SUCCESS)
    {
        if (CCA_AWS_SubscribeToAndRegisterTopicFilter(&str_MqttContext, sz_UpdateAcceptedTopic,
            strlen(sz_UpdateAcceptedTopic), aws_UpdateAcceptedHandler) == EXIT_SUCCESS)
        {
            if (CCA_AWS_SubscribeToAndRegisterTopicFilter(&str_MqttContext, sz_UpdateRejectedTopic,
                strlen(sz_UpdateRejectedTopic), aws_UpdateRejectedHandler) == EXIT_SUCCESS)
            {
                LOG_INF("Subcribed to all %s shdadow topics", sz_ShadowName);
            }
            else
            {
                LOG_ERR("ERROR: Failed to subcribe to %s topic.", sz_UpdateRejectedTopic);
                int_rc = EXIT_FAILURE;
            }
        }
        else
        {
            LOG_ERR("ERROR: Failed to subcribe to %s topic.", sz_UpdateAcceptedTopic);
            int_rc = EXIT_FAILURE;
        }
    }
    else
    {
        LOG_ERR("ERROR: Failed to subcribe to %s topic.", sz_UpdateDeltaTopic);
        int_rc = EXIT_FAILURE;
    }
    return int_rc;
}

ModScop int CCA_AWS_SubscribeToCommandRequestTopic(void)
{
    return CCA_AWS_SubscribeToAndRegisterTopicFilter(&str_MqttContext, cca_sz_command_request_topic,
                                                strlen(cca_sz_command_request_topic), aws_CommandHandler);
}

/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

FileScop void aws_CommandHandler(MQTTContext_t *ptr_str_Context, MQTTPublishInfo_t *ptr_str_PublishInfo)
{
    LOG_INF("CCA_aws_CommandHandler() triggered.\n");
    // Convert the raw payload into a json document and prepare for Analysis
    ArduinoJson::StaticJsonDocument<CCA_u16_COMMAND_JSON_DOC_SIZE> jd_Request;
    ArduinoJson::deserializeJson(jd_Request, ptr_str_PublishInfo->pPayload);
    ArduinoJson::JsonVariant id = jd_Request["id"];
    ArduinoJson::JsonObject requestBody = jd_Request["body"];

    // Frame the Response JSON document
    ArduinoJson::StaticJsonDocument<CCA_u16_COMMAND_JSON_DOC_SIZE> jd_Response;
    jd_Response["id"] = jd_Request["id"];
    ArduinoJson::JsonObject jo_ResponseBody = jd_Response.createNestedObject("body");
    jo_ResponseBody["st"] = "";

    //Validate the request
    cca_HandlerFunction_t fn_handler;
    int int_rc = CCA_Rqh_ValidateRequest(id.as<const char*>(), &fn_handler);
    if (int_rc == EXIT_SUCCESS)
    {
        // Call the handler function
        fn_handler(jd_Request, jd_Response);
    }
    else
    {
        //Convert int_rc to a string and add to the response body
        jo_ResponseBody["st"] = "fail";
        jo_ResponseBody["fault"] = strerror(-int_rc);
    }

    //Serialize and publish the response
    char sz_Response[CCA_u16_COMMAND_JSON_DOC_SIZE];
    ArduinoJson::serializeJson(jd_Response, sz_Response);
    if (CCA_AWS_PublishToTopic(&str_MqttContext, cca_sz_command_response_topic,strlen(cca_sz_command_response_topic),
                        sz_Response, strlen(sz_Response))== EXIT_SUCCESS)
    {
        LOG_INF("Successfully handled command request id:%s and published response st:%s.", 
                id.as<const char *>(), jo_ResponseBody["st"].as<const char *>());
    }
    else {
        LOG_INF("ERROR: Failed to handle command request id:%s and published response st:%s.", 
                id.as<const char *>(), jo_ResponseBody["st"].as<const char *>());
    }
}

FileScop void aws_UpdateDeltaHandler(MQTTContext_t *ptr_str_Context, MQTTPublishInfo_t *ptr_str_PublishInfo)
{
    assert(ptr_str_PublishInfo != NULL);
    assert(ptr_str_PublishInfo->pPayload != NULL);
    LOG_INF("/update/delta json payload:%s.", (const char *)ptr_str_PublishInfo->pPayload);
    // Extract the requested parameter changes from the desired state
    ArduinoJson::StaticJsonDocument<CCA_u16_SETTINGS_JSON_DOC_SIZE> jd_Payload;
    ArduinoJson::deserializeJson(jd_Payload, ptr_str_PublishInfo->pPayload);
    ArduinoJson::JsonObject jo_Datum = jd_Payload["state"]["product"]["datum"];
    uint8_t u8_NumberOfRequestedKeys = 0;
    char *arr_sz_RequestedKeys[128];
    char *arr_sz_requestedValues[128];
    int int_RequestSatuses[128];
    for (ArduinoJson::JsonPair keyValue : jo_Datum) // For every key value pair in datum
    {  
        const char *sz_key = keyValue.key().c_str();
        ArduinoJson::JsonVariant value = keyValue.value();
        LOG_INF("Key: %s, Value: %i", sz_key, value.as<int>());
        arr_sz_RequestedKeys[u8_NumberOfRequestedKeys] = (char*)sz_key;
        arr_sz_requestedValues[u8_NumberOfRequestedKeys] = (char*)value.as<const char*>();
        u8_NumberOfRequestedKeys++;
    }
    // pcs_settings Shadow
    if (strcmp(ptr_str_PublishInfo->pTopicName, cca_sz_pcs_update_delta_topic) == EXIT_SUCCESS)    
    {
        // Send a request to change the settings on each PCS in the system
        CCA_Icc_SetUnitConfiguration(CCA_sz_UNIT_TYPE_PCS, 0, u8_NumberOfRequestedKeys, (const char**)arr_sz_RequestedKeys, 
                                (const char**)arr_sz_requestedValues, int_RequestSatuses);

        // **Clear the Desired state to avoid infinite looping upon failure
        ArduinoJson::StaticJsonDocument<CCA_u16_DELTA_JSON_DOC_SIZE> jd_ClearDesiredState;
        jd_ClearDesiredState["state"]["desired"] = (char*)NULL;
        char sz_ClearDesiredState[CCA_u16_DELTA_JSON_DOC_SIZE];
        ArduinoJson::serializeJson(jd_ClearDesiredState, sz_ClearDesiredState);
        if (CCA_AWS_PublishToTopic(&str_MqttContext, cca_sz_pcs_update_topic,strlen(cca_sz_pcs_update_topic),
                            sz_ClearDesiredState, strlen(sz_ClearDesiredState)) == EXIT_SUCCESS) 
        {
            LOG_INF("Successfully clear desired state of %s topic.", cca_sz_pcs_update_topic);
        }        
        else {
            LOG_INF("Failed to clear desired state of %s topic.", cca_sz_pcs_update_topic);
        }                                               
        // Evaluate the success of each setting and send update to setting shadow if it was successfull
        ArduinoJson::StaticJsonDocument<CCA_u16_SETTINGS_JSON_DOC_SIZE> jd_Reported;
        ArduinoJson::JsonObject jo_ReportedDatum = jd_Reported["state"]["reported"]["product"].createNestedObject("datum");
        for (uint i = 0; i < u8_NumberOfRequestedKeys; i++) 
        {
            if (int_RequestSatuses[i] == EXIT_SUCCESS) // All PCS settings were successfully set for this key
            {   
                jo_ReportedDatum[arr_sz_RequestedKeys[i]] = arr_sz_requestedValues[i];
            }
        }
        char sz_PcsSettingsReport[CCA_u16_SETTINGS_JSON_DOC_SIZE];
        ArduinoJson::serializeJson(jd_Reported, sz_PcsSettingsReport);
        // Publish with new data from all successfully written values
        if (CCA_AWS_PublishToTopic(&str_MqttContext, cca_sz_pcs_update_topic, strlen(cca_sz_pcs_update_topic),
                            sz_PcsSettingsReport, strlen(sz_PcsSettingsReport)) == EXIT_SUCCESS) 
        {
            LOG_INF("Successfully published reported state of %s topic.", cca_sz_pcs_update_topic);
        }        
        else {
            LOG_INF("Failed to published reported state of %s topic.", cca_sz_pcs_update_topic);
        }     
    }
    // dcdc_settings Shadow
    else if (strcmp(ptr_str_PublishInfo->pTopicName, cca_sz_dcdc_update_delta_topic) == EXIT_SUCCESS)  
    {
        // Send a request to change the settings on each DCDC in the system
        uint8_t u8_DcdcQty = CCA_Int_GetUnitQuantity(DCDC);
        for (uint8_t i = 0; i < u8_DcdcQty; i++)
        {
            CCA_Icc_SetUnitConfiguration(CCA_sz_UNIT_TYPE_DCDC, i, u8_NumberOfRequestedKeys, 
                                            (const char**)arr_sz_RequestedKeys, 
                                            (const char**)arr_sz_requestedValues, int_RequestSatuses);
        }

        // **Clear the Desired state to avoid infinite looping upon failure
        ArduinoJson::StaticJsonDocument<CCA_u16_DELTA_JSON_DOC_SIZE> jd_ClearDesiredState;
        jd_ClearDesiredState["state"]["desired"] = (char*)NULL;
        char sz_ClearDesiredState[CCA_u16_DELTA_JSON_DOC_SIZE];
        ArduinoJson::serializeJson(jd_ClearDesiredState, sz_ClearDesiredState);
        if (CCA_AWS_PublishToTopic(&str_MqttContext, cca_sz_dcdc_update_topic, strlen(cca_sz_dcdc_update_topic),
                            sz_ClearDesiredState, strlen(sz_ClearDesiredState)) == EXIT_SUCCESS) 
        {
            LOG_INF("Successfully clear desired state of %s topic.", cca_sz_dcdc_update_topic);
        }        
        else {
            LOG_INF("Failed to clear desired state of %s topic.", cca_sz_dcdc_update_topic);
        }                                              
        // Evaluate the success of each setting and send update to setting shadow if it was successfull
        ArduinoJson::StaticJsonDocument<CCA_u16_SETTINGS_JSON_DOC_SIZE> jo_Reported;
        ArduinoJson::JsonObject jo_ReportedDatum = jo_Reported["state"]["reported"]["product"].createNestedObject("datum");
        for (uint i = 0; i < u8_NumberOfRequestedKeys; i++) 
        {
            if (int_RequestSatuses[i] == EXIT_SUCCESS)     // All DC/DC settings were successfully set for this key
            {   
                jo_ReportedDatum[arr_sz_RequestedKeys[i]] = arr_sz_requestedValues[i];
            }
        }
        char sz_DcdcSettingsReport[CCA_u16_SETTINGS_JSON_DOC_SIZE];
        ArduinoJson::serializeJson(jo_Reported, sz_DcdcSettingsReport);
        // Publish with new data from all successfully written values
        if (CCA_AWS_PublishToTopic(&str_MqttContext, cca_sz_dcdc_update_topic, strlen(cca_sz_dcdc_update_topic),
                            sz_DcdcSettingsReport, strlen(sz_DcdcSettingsReport)) == EXIT_SUCCESS) 
        {
            LOG_INF("Successfully published reported state of %s topic.", cca_sz_dcdc_update_topic);
        }        
        else {
            LOG_INF("Failed to published reported state of %s topic.", cca_sz_dcdc_update_topic);
        } 
    }
    // hvac_settings Shadow
    else if (strcmp(ptr_str_PublishInfo->pTopicName, cca_sz_hvac_update_delta_topic) == EXIT_SUCCESS)  
    {
        // Send a request to change the settings on each HVAC in the system
        CCA_Icc_SetUnitConfiguration(CCA_sz_UNIT_TYPE_HVAC, 0, u8_NumberOfRequestedKeys, (const char**)arr_sz_RequestedKeys, 
                                (const char**)arr_sz_requestedValues, int_RequestSatuses);

        // **Clear the Desired state to avoid infinite looping upon failure
        ArduinoJson::StaticJsonDocument<CCA_u16_DELTA_JSON_DOC_SIZE> jd_ClearDesiredState;
        jd_ClearDesiredState["state"]["desired"] = (char*)NULL;
        char sz_ClearDesiredState[CCA_u16_DELTA_JSON_DOC_SIZE];
        ArduinoJson::serializeJson(jd_ClearDesiredState, sz_ClearDesiredState);
        if (CCA_AWS_PublishToTopic(&str_MqttContext, cca_sz_hvac_update_topic,strlen(cca_sz_hvac_update_topic),
                            sz_ClearDesiredState, strlen(sz_ClearDesiredState)) == EXIT_SUCCESS) 
        {
            LOG_INF("Successfully clear desired state of %s topic.", cca_sz_hvac_update_topic);
        }        
        else {
            LOG_INF("Failed to clear desired state of %s topic.", cca_sz_hvac_update_topic);
        }                                              
        // Evaluate the success of each setting and send update to setting shadow if it was successfull
        ArduinoJson::StaticJsonDocument<CCA_u16_SETTINGS_JSON_DOC_SIZE> jo_Reported;
        ArduinoJson::JsonObject jo_ReportedDatum = jo_Reported["state"]["reported"]["product"].createNestedObject("datum");
        for (uint i = 0; i < u8_NumberOfRequestedKeys; i++) 
        {
            if (int_RequestSatuses[i] == EXIT_SUCCESS)     // All HVAC settings were successfully set for this key
            {   
                jo_ReportedDatum[arr_sz_RequestedKeys[i]] = arr_sz_requestedValues[i];
            }
        }
        char sz_HvacSettingsReport[CCA_u16_SETTINGS_JSON_DOC_SIZE];
        ArduinoJson::serializeJson(jo_Reported, sz_HvacSettingsReport);
        // Publish with new data from all successfully written values
        if (CCA_AWS_PublishToTopic(&str_MqttContext, cca_sz_hvac_update_topic, strlen(cca_sz_hvac_update_topic),
                            sz_HvacSettingsReport, strlen(sz_HvacSettingsReport)) == EXIT_SUCCESS) 
        {
            LOG_INF("Successfully published reported state of %s topic.", cca_sz_hvac_update_topic);
        }        
        else {
            LOG_INF("Failed to published reported state of %s topic.", cca_sz_hvac_update_topic);
        } 
    }
    // pmu_settings Shadow
    else if (strcmp(ptr_str_PublishInfo->pTopicName, cca_sz_pmu_update_delta_topic) == EXIT_SUCCESS)  
    {
        // Send a request to change the settings on each PMU in the system
        CCA_Icc_SetEMSConfiguration(CCA_sz_UNIT_TYPE_PMU, COD_u8_NULL_UNIT_ID, u8_NumberOfRequestedKeys, (const char**)arr_sz_RequestedKeys, 
                                (const char**)arr_sz_requestedValues, int_RequestSatuses);

        // **Clear the Desired state to avoid infinite looping upon failure
        ArduinoJson::StaticJsonDocument<CCA_u16_DELTA_JSON_DOC_SIZE> jd_ClearDesiredState;
        jd_ClearDesiredState["state"]["desired"] = (char*)NULL;
        char sz_ClearDesiredState[CCA_u16_DELTA_JSON_DOC_SIZE];
        ArduinoJson::serializeJson(jd_ClearDesiredState, sz_ClearDesiredState);
        if (CCA_AWS_PublishToTopic(&str_MqttContext, cca_sz_pmu_update_topic,strlen(cca_sz_pmu_update_topic),
                            sz_ClearDesiredState, strlen(sz_ClearDesiredState)) == EXIT_SUCCESS) 
        {
            LOG_INF("Successfully clear desired state of %s topic.", cca_sz_pmu_update_topic);
        }        
        else {
            LOG_INF("Failed to clear desired state of %s topic.", cca_sz_pmu_update_topic);
        }                                              
        // Evaluate the success of each setting and send update to setting shadow if it was successfull
        ArduinoJson::StaticJsonDocument<CCA_u16_SETTINGS_JSON_DOC_SIZE> jo_Reported;
        ArduinoJson::JsonObject jo_ReportedDatum = jo_Reported["state"]["reported"]["product"].createNestedObject("datum");
        for (uint i = 0; i < u8_NumberOfRequestedKeys; i++) 
        {
            if (int_RequestSatuses[i] == EXIT_SUCCESS)     // All PMU settings were successfully set for this key
            {   
                jo_ReportedDatum[arr_sz_RequestedKeys[i]] = arr_sz_requestedValues[i];
            }
        }
        char sz_PmuSettingsReport[CCA_u16_SETTINGS_JSON_DOC_SIZE];
        ArduinoJson::serializeJson(jo_Reported, sz_PmuSettingsReport);
        // Publish with new data from all successfully written values
        if (CCA_AWS_PublishToTopic(&str_MqttContext, cca_sz_pmu_update_topic, strlen(cca_sz_pmu_update_topic),
                            sz_PmuSettingsReport, strlen(sz_PmuSettingsReport)) == EXIT_SUCCESS) 
        {
            LOG_INF("Successfully published reported state of %s topic.", cca_sz_pmu_update_topic);
        }        
        else {
            LOG_INF("Failed to published reported state of %s topic.", cca_sz_pmu_update_topic);
        } 
    }
    // ems_settings Shadow
    else if (strcmp(ptr_str_PublishInfo->pTopicName, cca_sz_ems_update_delta_topic) == EXIT_SUCCESS)  
    {   
        ArduinoJson::StaticJsonDocument<CCA_u16_SETTINGS_JSON_DOC_SIZE> jo_Reported;
        ArduinoJson::JsonObject jo_ReportedDatum = jo_Reported["state"]["reported"]["product"].createNestedObject("datum");
        // Check for and send requests for settings changes to the PCS communication
        if (jo_Datum.containsKey(CCA_sz_UNIT_TYPE_PCS)) 
        {
            ArduinoJson::JsonArray ja_Pcs = jo_Datum[CCA_sz_UNIT_TYPE_PCS];
            // Iterate through all of the object in the pcs array
            ArduinoJson::JsonArray ja_PcsReported = jo_ReportedDatum.createNestedArray(CCA_sz_UNIT_TYPE_PCS);
            for (ArduinoJson::JsonVariant jv_PcsObject : ja_Pcs) 
            {
                if (jv_PcsObject.containsKey("datum"))
                {
                    uint8_t u8_id = jv_PcsObject["id"].as<uint8_t>();
                    ArduinoJson::JsonObject pcsDatum = jv_PcsObject["datum"];
                    uint8_t u8_numberOfRequestedPCSKeys = 0;
                    char *sz_requestedPCSKeys[64];
                    char *sz_requestedPCSValues[64];
                    int s16_requestPCSSatuses[64];
                    for (ArduinoJson::JsonPair keyValue : pcsDatum)     // For every key value pair in datum
                    {  
                        const char *sz_key = keyValue.key().c_str();
                        ArduinoJson::JsonVariant value = keyValue.value();
                        LOG_INF("Key: %s, Value: %s", sz_key, value.as<const char*>());
                        sz_requestedPCSKeys[u8_numberOfRequestedPCSKeys] = (char*)sz_key;
                        sz_requestedPCSValues[u8_numberOfRequestedPCSKeys] = (char*)value.as<const char*>();
                        u8_numberOfRequestedPCSKeys++;
                    }
                    if (CCA_Icc_SetEMSConfiguration(CCA_sz_UNIT_TYPE_PCS, u8_id, u8_numberOfRequestedPCSKeys, (const char**)sz_requestedPCSKeys,
                                                (const char**)sz_requestedPCSValues, s16_requestPCSSatuses) == EXIT_SUCCESS)
                    {
                        // Evaluate the success of each setting and send update to setting shadow if it was successfull
                        ArduinoJson::JsonObject pcsReportedObject = ja_PcsReported.createNestedObject();
                        pcsReportedObject["id"] = jv_PcsObject["id"];
                        ArduinoJson::JsonObject pcsReportedDatum = pcsReportedObject.createNestedObject("datum");
                        for (uint i = 0; i < u8_numberOfRequestedPCSKeys; i++) 
                        {
                            if (s16_requestPCSSatuses[i] == EXIT_SUCCESS)  // All PCS settings were successfully set for this key
                            {   
                                pcsReportedDatum[sz_requestedPCSKeys[i]] = sz_requestedPCSValues[i];
                            }
                        }
                    }
                }
            }
        }
        // Check for and send requests for settings changes to the DCDC communication
        if (jo_Datum.containsKey(CCA_sz_UNIT_TYPE_DCDC)) 
        {
            ArduinoJson::JsonArray ja_Dcdc = jo_Datum[CCA_sz_UNIT_TYPE_DCDC];
            // Iterate through all of the object in the dcdc array
            ArduinoJson::JsonArray ja_DcdcReported = jo_ReportedDatum.createNestedArray(CCA_sz_UNIT_TYPE_DCDC);
            for (ArduinoJson::JsonVariant jv_DcdcObject : ja_Dcdc) 
            {
                if (jv_DcdcObject.containsKey("datum"))
                {
                    uint8_t u8_id = jv_DcdcObject["id"].as<uint8_t>();
                    ArduinoJson::JsonObject dcdcDatum = jv_DcdcObject["datum"];
                    uint8_t u8_numberOfRequestedDCDCKeys = 0;
                    char *sz_requestedDCDCKeys[64];
                    char *sz_requestedDCDCValues[64];
                    int s16_requestDCDCSatuses[64];
                    for (ArduinoJson::JsonPair keyValue : dcdcDatum)     // For every key value pair in datum
                    {  
                        const char *sz_key = keyValue.key().c_str();
                        ArduinoJson::JsonVariant value = keyValue.value();
                        LOG_INF("Key: %s, Value: %i", sz_key, value.as<int>());
                        sz_requestedDCDCKeys[u8_numberOfRequestedDCDCKeys] = (char*)sz_key;
                        sz_requestedDCDCValues[u8_numberOfRequestedDCDCKeys] = (char*)value.as<const char*>();
                        u8_numberOfRequestedDCDCKeys++;
                    }
                    if (CCA_Icc_SetEMSConfiguration(CCA_sz_UNIT_TYPE_DCDC, u8_id, u8_numberOfRequestedDCDCKeys, 
                                                (const char**)sz_requestedDCDCKeys, (const char**)sz_requestedDCDCValues, 
                                                s16_requestDCDCSatuses) == EXIT_SUCCESS)
                    {
                        // Evaluate the success of each setting and send update to setting shadow if it was successfull
                        ArduinoJson::JsonObject dcdcReportedObject = ja_DcdcReported.createNestedObject();
                        dcdcReportedObject["id"] = jv_DcdcObject["id"];
                        ArduinoJson::JsonObject dcdcReportedDatum = dcdcReportedObject.createNestedObject("datum");
                        for (uint i = 0; i < u8_numberOfRequestedDCDCKeys; i++) 
                        {
                            if (s16_requestDCDCSatuses[i] == EXIT_SUCCESS)  // All DCDC settings were successfully set for this key
                            {   
                                dcdcReportedDatum[sz_requestedDCDCKeys[i]] = sz_requestedDCDCValues[i];
                            }
                        }
                    }
                }
            }
        }
        // Check for and send requests for settings changes to the HVAC communication
        if (jo_Datum.containsKey(CCA_sz_UNIT_TYPE_HVAC)) 
        {
            ArduinoJson::JsonArray ja_Hvac = jo_Datum[CCA_sz_UNIT_TYPE_HVAC];
            // Iterate through all of the object in the hvac array
            ArduinoJson::JsonArray ja_hvacReported = jo_ReportedDatum.createNestedArray(CCA_sz_UNIT_TYPE_HVAC);
            for (ArduinoJson::JsonVariant jv_HvacObject : ja_Hvac) 
            {
                if (jv_HvacObject.containsKey("datum"))
                {
                    uint8_t u8_id = jv_HvacObject["id"].as<uint8_t>();
                    ArduinoJson::JsonObject hvacDatum = jv_HvacObject["datum"];
                    uint8_t u8_numberOfRequestedHVACKeys = 0;
                    char *sz_requestedHVACKeys[64];
                    char *sz_requestedHVACValues[64];
                    int s16_requestHVACSatuses[64];
                    for (ArduinoJson::JsonPair keyValue : hvacDatum)     // For every key value pair in datum
                    {  
                        const char *sz_key = keyValue.key().c_str();
                        ArduinoJson::JsonVariant value = keyValue.value();
                        LOG_INF("Key: %s, Value: %i", sz_key, value.as<int>());
                        sz_requestedHVACKeys[u8_numberOfRequestedHVACKeys] = (char*)sz_key;
                        sz_requestedHVACValues[u8_numberOfRequestedHVACKeys] = (char*)value.as<const char*>();
                        u8_numberOfRequestedHVACKeys++;
                    }
                    if (CCA_Icc_SetEMSConfiguration(CCA_sz_UNIT_TYPE_HVAC, u8_id, u8_numberOfRequestedHVACKeys, 
                                                (const char**)sz_requestedHVACKeys, (const char**)sz_requestedHVACValues, 
                                                s16_requestHVACSatuses) == EXIT_SUCCESS)
                    {
                        // Evaluate the success of each setting and send update to setting shadow if it was successfull
                        ArduinoJson::JsonObject hvacReportedObject = ja_hvacReported.createNestedObject();
                        hvacReportedObject["id"] = jv_HvacObject["id"];
                        ArduinoJson::JsonObject hvacReportedDatum = hvacReportedObject.createNestedObject("datum");
                        for (uint i = 0; i < u8_numberOfRequestedHVACKeys; i++) 
                        {
                            if (s16_requestHVACSatuses[i] == EXIT_SUCCESS)  // All HVAC settings were successfully set for this key
                            {   
                                hvacReportedDatum[sz_requestedHVACKeys[i]] = sz_requestedHVACValues[i];
                            }
                        }
                    }
                }
            }
        }
        // Check for and send requests for settings changes to the GW communication
        if (jo_Datum.containsKey(CCA_sz_UNIT_TYPE_GW)) 
        {
            ArduinoJson::JsonArray gwArray = jo_Datum[CCA_sz_UNIT_TYPE_GW];
            // Iterate through all of the object in the gw array
            ArduinoJson::JsonArray ja_GwReported = jo_ReportedDatum.createNestedArray(CCA_sz_UNIT_TYPE_GW);
            for (ArduinoJson::JsonVariant gwObject : gwArray) 
            {
                if (gwObject.containsKey("datum"))
                {
                    uint8_t u8_id = gwObject["id"].as<uint8_t>();
                    ArduinoJson::JsonObject gwDatum = gwObject["datum"];
                    uint8_t u8_numberOfRequestedGWKeys = 0;
                    char *sz_requestedGWKeys[64];
                    char *sz_requestedGWValues[64];
                    int s16_requestGWSatuses[64];
                    for (ArduinoJson::JsonPair keyValue : gwDatum)     // For every key value pair in datum
                    {  
                        const char *sz_key = keyValue.key().c_str();
                        ArduinoJson::JsonVariant value = keyValue.value();
                        LOG_INF("Key: %s, Value: %i", sz_key, value.as<int>());
                        sz_requestedGWKeys[u8_numberOfRequestedGWKeys] = (char*)sz_key;
                        sz_requestedGWValues[u8_numberOfRequestedGWKeys] = (char*)value.as<const char*>();
                        u8_numberOfRequestedGWKeys++;
                    }
                    if (CCA_Icc_SetEMSConfiguration(CCA_sz_UNIT_TYPE_GW, u8_id, u8_numberOfRequestedGWKeys, 
                                                (const char**)sz_requestedGWKeys, (const char**)sz_requestedGWValues, 
                                                s16_requestGWSatuses) == EXIT_SUCCESS)
                    {
                        // Evaluate the success of each setting and send update to setting shadow if it was successfull
                        ArduinoJson::JsonObject gwReportedObject = ja_GwReported.createNestedObject();
                        gwReportedObject["id"] = gwObject["id"];
                        ArduinoJson::JsonObject gwReportedDatum = gwReportedObject.createNestedObject("datum");
                        for (uint i = 0; i < u8_numberOfRequestedGWKeys; i++) 
                        {
                            if (s16_requestGWSatuses[i] == EXIT_SUCCESS)  // All GW settings were successfully set for this key
                            {   
                                gwReportedDatum[sz_requestedGWKeys[i]] = sz_requestedGWValues[i];
                            }
                        }
                    }
                }
            }
        }
        // Check for and send requests for settings changes to the Time of Use Profiles
        if (jo_Datum.containsKey(CCA_sz_TOU_KEY)) 
        {
            // Create Request Block
            ArduinoJson::StaticJsonDocument<CCA_u16_SETTINGS_JSON_DOC_SIZE> jd_Request;
            char sz_Request[CCA_u16_SETTINGS_JSON_DOC_SIZE];
            uint8_t u8_ProfileID = COD_u8_NULL_UNIT_ID;
            ArduinoJson::JsonObject jo_TouSettingParent = jo_Datum[CCA_sz_TOU_KEY];
            int s16_TouSettingSt = EXIT_SUCCESS;
            // Iterate through all of the settings in the Time of Use Settings
            for (ArduinoJson::JsonPair jp_TouSetting : jo_TouSettingParent) 
            {
                // Check if the Setting is a Profile
                const char *sz_TouSettingKey = jp_TouSetting.key().c_str();
                if (isdigit(sz_TouSettingKey[0]))
                {
                    ArduinoJson::JsonObject jo_TouProfile = jo_TouSettingParent[sz_TouSettingKey];
                    for (ArduinoJson::JsonPair jp_TouProfileSetting : jo_TouProfile)
                    {
                        uint8_t u8_ProfileID = strtol(sz_TouSettingKey, NULL, 10);
                        // Check if the Setting is an array
                        const char *sz_TouProfileSettingKey = jp_TouProfileSetting.key().c_str();
                        ArduinoJson::JsonArray ja_TouArray = jo_TouProfile[sz_TouProfileSettingKey];
                        if (!ja_TouArray.isNull())  // Array Settings
                        {
                            // jd_Request.add(ja_TouArray);
                            ArduinoJson::JsonArray ja_RequestArray = jd_Request.createNestedArray(sz_TouProfileSettingKey);
                            ja_RequestArray.set(ja_TouArray);
                            ArduinoJson::serializeJson(jd_Request, sz_Request);
                            if (s16_TouSettingSt = CCA_Icc_SetTouSettings(u8_ProfileID, sz_Request, sz_TouProfileSettingKey) != EXIT_SUCCESS)
                            {
                                LOG_INF("Delta Handler failed  to set time of use setting: %s", sz_TouProfileSettingKey);
                            }
                        }
                        else // Top Level Profile Setting
                        {
                            ArduinoJson::JsonVariant jv_TouProfileSettingValue = jp_TouProfileSetting.value();
                            const char *sz_TouProfileSettingValue = jv_TouProfileSettingValue.as<const char*>();
                            jd_Request[sz_TouProfileSettingKey] = sz_TouProfileSettingValue;
                            ArduinoJson::serializeJson(jd_Request, sz_Request);
                            if(s16_TouSettingSt = CCA_Icc_SetTouSettings(u8_ProfileID, sz_Request, sz_TouProfileSettingKey) != EXIT_SUCCESS)
                            {
                                LOG_INF("Delta Handler failed  to set time of use setting: %s", sz_TouProfileSettingKey);
                            }
                        }
                    }
                }
                else // Top Level Time of Use Setting
                {
                    ArduinoJson::JsonVariant jv_TouSettingValue = jp_TouSetting.value();
                    const char *sz_TouSettingValue = jv_TouSettingValue.as<const char*>();
                    jd_Request[sz_TouSettingKey] = sz_TouSettingValue;
                    ArduinoJson::serializeJson(jd_Request, sz_Request);
                    if (CCA_Icc_SetTouSettings(u8_ProfileID, sz_Request, sz_TouSettingKey) != EXIT_SUCCESS)
                    {
                        LOG_INF("Delta Handler failed  to set time of use setting: %s", sz_TouSettingKey);
                    }

                }
            }
            // Retreive the current Time of Use setting to update the reported state
            CCA_Int_AppendTouSettingsToJsonObject(jo_ReportedDatum);
        }

        // Set all other EMS configuration settings
        uint8_t u8_NumberOfRequestedEMSKeys = 0;
        char *sz_RequestedEMSKeys[128];
        char *sz_RequestedEMSValues[128];
        int s16_RequestEMSSatuses[128];
        for (ArduinoJson::JsonPair keyValue : jo_Datum) // For every key value pair in datum
        {  
            const char *sz_key = keyValue.key().c_str();
            // Skip component objects that were already handled
            if ((strcmp(sz_key, CCA_sz_UNIT_TYPE_PCS) != EXIT_SUCCESS) && (strcmp(sz_key, CCA_sz_UNIT_TYPE_DCDC) != EXIT_SUCCESS) 
                && (strcmp(sz_key, CCA_sz_UNIT_TYPE_HVAC) != EXIT_SUCCESS) && (strcmp(sz_key, CCA_sz_UNIT_TYPE_GW) != EXIT_SUCCESS)
                && (strcmp(sz_key, CCA_sz_TOU_KEY) != EXIT_SUCCESS))    
            {
                ArduinoJson::JsonVariant value = keyValue.value();
                LOG_INF("Key: %s, Value: %i", sz_key, value.as<int>());
                sz_RequestedEMSKeys[u8_NumberOfRequestedEMSKeys] = (char*)sz_key;
                sz_RequestedEMSValues[u8_NumberOfRequestedEMSKeys] = (char*)value.as<const char*>();
                u8_NumberOfRequestedEMSKeys++;
            }
        }

        // **Clear the Desired state to avoid infinite looping upon failure
        ArduinoJson::StaticJsonDocument<CCA_u16_DELTA_JSON_DOC_SIZE> jd_ClearDesiredState;
        jd_ClearDesiredState["state"]["desired"] = (char*)NULL;
        char sz_ClearDesiredState[CCA_u16_DELTA_JSON_DOC_SIZE];
        ArduinoJson::serializeJson(jd_ClearDesiredState, sz_ClearDesiredState);
        if (CCA_AWS_PublishToTopic(&str_MqttContext, cca_sz_ems_update_topic,strlen(cca_sz_ems_update_topic),
                            sz_ClearDesiredState, strlen(sz_ClearDesiredState)) == EXIT_SUCCESS) 
        {
            LOG_INF("Successfully clear desired state of %s topic.", cca_sz_ems_update_topic);
        }        
        else {
            LOG_INF("Failed to clear desired state of %s topic.", cca_sz_ems_update_topic);
        }


        CCA_Int_SetCCUSettings(u8_NumberOfRequestedEMSKeys, (const char**)sz_RequestedEMSKeys, 
                            (const char**)sz_RequestedEMSValues);
        if (CCA_Icc_SetEMSConfiguration(CCA_sz_UNIT_TYPE_EMS, COD_u8_NULL_UNIT_ID, u8_NumberOfRequestedEMSKeys, 
                                    (const char**)sz_RequestedEMSKeys, (const char**)sz_RequestedEMSValues, 
                                    s16_RequestEMSSatuses) == EXIT_SUCCESS)
        {                  
            // Evaluate the success of each setting and send update to setting shadow if it was successfull
            for (uint i = 0; i < u8_NumberOfRequestedEMSKeys; i++) 
            {
                if (s16_RequestEMSSatuses[i] == EXIT_SUCCESS)  // All EMS settings were successfully set for this key
                {   
                    jo_ReportedDatum[sz_RequestedEMSKeys[i]] = sz_RequestedEMSValues[i];
                }
            }
        }
        char sz_EmsSettingsReport[CCA_u16_SETTINGS_JSON_DOC_SIZE];
        ArduinoJson::serializeJson(jo_Reported, sz_EmsSettingsReport);
        // Publish with new data from all successfully written values
        if (CCA_AWS_PublishToTopic(&str_MqttContext, cca_sz_ems_update_topic, strlen(cca_sz_ems_update_topic),
                            sz_EmsSettingsReport, strlen(sz_EmsSettingsReport)) == EXIT_SUCCESS) 
        {
            LOG_INF("Successfully published reported state of %s topic.", cca_sz_ems_update_topic);
        }        
        else {
            LOG_INF("Failed to published reported state of %s topic.", cca_sz_ems_update_topic);
        } 
    }
    // version Shadow
    else if (strcmp(ptr_str_PublishInfo->pTopicName, cca_sz_version_update_delta_topic) == EXIT_SUCCESS)
    {
        //Changing the version info is not supported
        // **Clear the Desired state to avoid infinite looping upon failure
        ArduinoJson::StaticJsonDocument<CCA_u16_DELTA_JSON_DOC_SIZE> jd_ClearDesiredState;
        jd_ClearDesiredState["state"]["desired"] = (char*)NULL;
        char sz_ClearDesiredState[CCA_u16_DELTA_JSON_DOC_SIZE];
        ArduinoJson::serializeJson(jd_ClearDesiredState, sz_ClearDesiredState);
        if (CCA_AWS_PublishToTopic(&str_MqttContext, cca_sz_version_update_topic, strlen(cca_sz_version_update_topic),
                            sz_ClearDesiredState, strlen(sz_ClearDesiredState)) == EXIT_SUCCESS) 
        {
            LOG_INF("Successfully cleared desired state of %s topic.", cca_sz_version_update_topic);
        }        
        else
        {
            LOG_INF("Failed to clear desired state of %s topic.", cca_sz_version_update_topic);
        }
        if (CCA_Version_PublishVersionUpdate() == EXIT_SUCCESS)
        {
            LOG_INF("Successfully published reported state of %s topic.", cca_sz_version_update_topic);
        }        
        else
        {
            LOG_INF("Failed to publish reported state of %s topic.", cca_sz_version_update_topic);
        }
    }
    else {  
        LOG_INF("ERROR: Unsupported delta topic received: %s", ptr_str_PublishInfo->pTopicName);
    }
}

FileScop void aws_UpdateAcceptedHandler(MQTTContext_t *ptr_str_Context, MQTTPublishInfo_t *ptr_str_PublishInfo)
{
    (void)ptr_str_Context;
    (void)ptr_str_PublishInfo;
    LOG_INF("Update to shadow success\n");
}

FileScop void aws_UpdateRejectedHandler(MQTTContext_t *ptr_str_Context, MQTTPublishInfo_t *ptr_str_PublishInfo)
{
    (void)ptr_str_Context;
    (void)ptr_str_PublishInfo;
    /* print the error cause */
    LOG_INF("Update to shadow failed\n");
}

FileScop void aws_HandleIncomingPublish(MQTTPublishInfo_t *ptr_str_PublishInfo)
{
    assert(ptr_str_PublishInfo != NULL);

#if 0
    /* Process incoming Publish. */
    LOG_INF("Incoming QOS : %d.", ptr_str_PublishInfo->qos);

    LOG_INF("Incoming Publish Topic Name: %.*s matches subscribed topic.\n"
             "Incoming Publish Message : %.*s.\n\n",
             ptr_str_PublishInfo->topicNameLength,
             ptr_str_PublishInfo->pTopicName,
             (int)ptr_str_PublishInfo->payloadLength,
             (const char *)ptr_str_PublishInfo->pPayload);
#endif
}

FileScop void aws_CommonEventHandler(MQTTContext_t *ptr_str_MqttContext, MQTTPacketInfo_t *ptr_str_PacketInfo,
                                        MQTTDeserializedInfo_t *ptr_str_DeserializedInfo)
{
    assert(ptr_str_MqttContext != NULL);
    assert(ptr_str_PacketInfo != NULL);
    assert(ptr_str_DeserializedInfo != NULL);

    /* Suppress unused parameter warning when asserts are disabled in build. */
    (void)ptr_str_MqttContext;

    /* Handle incoming publish. The lower 4 bits of the publish packet
     * type is used for the dup, QoS, and retain flags. Hence masking
     * out the lower bits to check if the packet is publish. */
    if ((ptr_str_PacketInfo->type & 0xF0U) == MQTT_PACKET_TYPE_PUBLISH)
    {
        assert(ptr_str_DeserializedInfo->pPublishInfo != NULL);
        /* Handle incoming publish. */
        SubscriptionManager_DispatchHandler(ptr_str_MqttContext, ptr_str_DeserializedInfo->pPublishInfo);
    }
    else
    {
        /* Handle other packets. */
        switch (ptr_str_PacketInfo->type)
        {
        case MQTT_PACKET_TYPE_SUBACK:
            LOG_INF("Received SUBACK.\n\n");
            /* Make sure ACK packet identifier matches with Request packet identifier. */
            // assert(u16_LastSubscribePacketIdentifier == ptr_str_DeserializedInfo->packetIdentifier);
            break;

        case MQTT_PACKET_TYPE_UNSUBACK:
            LOG_INF("Received UNSUBACK.\n\n");
            /* Make sure ACK packet identifier matches with Request packet identifier. */
            assert(u16_LastUnsubscribePacketIdentifier == ptr_str_DeserializedInfo->packetIdentifier);
            break;

        case MQTT_PACKET_TYPE_PINGRESP:

            /* Nothing to be done from application as library handles
             * PINGRESP. */
            LOG_WRN("PINGRESP should not be handled by the application "
                     "callback when using MQTT_ProcessLoop.\n\n");
            break;

        case MQTT_PACKET_TYPE_PUBACK:
            LOG_INF("PUBACK received for packet id %u.\n\n",
                     ptr_str_DeserializedInfo->packetIdentifier);
            break;

        /* Any other packet type is invalid. */
        default:
            LOG_ERR("Unknown packet type received:(%02x).\n\n",
                      ptr_str_PacketInfo->type);
        }
    }
}

FileScop int aws_EstablishMqttSession(MQTTContext_t *ptr_str_MqttContext, NetworkContext_t *ptr_str_NetworkContext,
                                        bool bol_CreateCleanSession, bool *ptr_bol_SessionPresent)
{
    int int_rc = EXIT_SUCCESS;
    MQTTStatus_t str_MqttStatus;
    MQTTConnectInfo_t str_ConnectInfo;
    MQTTFixedBuffer_t str_NetworkBuffer;
    TransportInterface_t str_Transport;

    assert(ptr_str_MqttContext != NULL);
    assert(ptr_str_NetworkContext != NULL);

    /* Fill in TransportInterface send and receive function pointers.
     * For this demo, TCP sockets are used to send and receive data
     * from network. Network context is SSL context for OpenSSL.*/
    str_Transport.pNetworkContext = ptr_str_NetworkContext;
    str_Transport.send = Openssl_Send;
    str_Transport.recv = Openssl_Recv;

    /* Fill the values for network buffer. */
    str_NetworkBuffer.pBuffer = u8_Buffer;
    str_NetworkBuffer.size = NETWORK_BUFFER_SIZE;

    /* Initialize MQTT library. */
    str_MqttStatus = MQTT_Init(ptr_str_MqttContext,
                           &str_Transport,
                           Clock_GetTimeMs,
                           aws_CommonEventHandler,
                           &str_NetworkBuffer);

    if (str_MqttStatus != MQTTSuccess)
    {
        int_rc = EXIT_FAILURE;
        LOG_ERR("MQTT init failed with status %s.", MQTT_Status_strerror(str_MqttStatus));
    }
    else
    {
        /* Establish MQTT session by sending a CONNECT packet. */

        /* If #bol_CreateCleanSession is true, start with a clean session
         * i.e. direct the MQTT broker to discard any previous session data.
         * If #bol_CreateCleanSession is false, directs the broker to attempt to
         * reestablish a session which was already present. */
        str_ConnectInfo.cleanSession = bol_CreateCleanSession;

        /* The client identifier is used to uniquely identify this MQTT client to
         * the MQTT broker. In a production device the identifier can be something
         * unique, such as a device serial number. */
        str_ConnectInfo.pClientIdentifier = cca_sz_thing_name;
        str_ConnectInfo.clientIdentifierLength = sizeof(cca_sz_thing_name)-1;

        /* The maximum time interval in seconds which is allowed to elapse
         * between two Control Packets.
         * It is the responsibility of the Client to ensure that the interval between
         * Control Packets being sent does not exceed the this Keep Alive value. In the
         * absence of sending any other Control Packets, the Client MUST send a
         * PINGREQ Packet. */
        str_ConnectInfo.keepAliveSeconds = MQTT_KEEP_ALIVE_INTERVAL_SECONDS;

        /* Username and password for authentication. Not used in this demo. */
        str_ConnectInfo.pUserName = NULL;
        str_ConnectInfo.userNameLength = 0U;
        str_ConnectInfo.pPassword = NULL;
        str_ConnectInfo.passwordLength = 0U;

        /* Send MQTT CONNECT packet to broker. */
        str_MqttStatus = MQTT_Connect(ptr_str_MqttContext, &str_ConnectInfo, NULL, CONNACK_RECV_TIMEOUT_MS, ptr_bol_SessionPresent);

        if (str_MqttStatus != MQTTSuccess)
        {
            int_rc = EXIT_FAILURE;
            LOG_ERR("Connection with MQTT broker failed with status %u.", str_MqttStatus);
        }
        else
        {
            LOG_INF("MQTT connection successfully established with broker.\n\n");
        }
    }

    return int_rc;
}

FileScop int aws_SubscribeToTopic(MQTTContext_t *ptr_str_MqttContext,const char *sz_TopicFilter,
                                    uint16_t u16_TopicFilterLength)
{
    int int_rc = EXIT_SUCCESS;
    MQTTStatus_t str_MqttStatus;
    MQTTSubscribeInfo_t arr_str_SubscriptionList[1];

    assert(ptr_str_MqttContext != NULL);

    /* Start with everything at 0. */
    (void)memset((void *)arr_str_SubscriptionList, 0x00, sizeof(arr_str_SubscriptionList));

    /* This demo subscribes and publishes to topics at Qos1, so the publish
     * messages received from the broker should have QoS1 as well. */
    arr_str_SubscriptionList[0].qos = MQTTQoS1;
    arr_str_SubscriptionList[0].pTopicFilter = sz_TopicFilter;
    arr_str_SubscriptionList[0].topicFilterLength = u16_TopicFilterLength;

    /* Generate packet identifier for the SUBSCRIBE packet. */
    u16_LastSubscribePacketIdentifier = MQTT_GetPacketId(ptr_str_MqttContext);

    /* Send SUBSCRIBE packet. */
    str_MqttStatus = MQTT_Subscribe(ptr_str_MqttContext,
                                arr_str_SubscriptionList,
                                sizeof(arr_str_SubscriptionList) / sizeof(MQTTSubscribeInfo_t),
                                u16_LastSubscribePacketIdentifier);

    if (str_MqttStatus != MQTTSuccess)
    {
        LOG_ERR("Failed to send SUBSCRIBE packet to broker with error = %u.",
                  str_MqttStatus);
        int_rc = EXIT_FAILURE;
    }
    else
    {
        LOG_INF("SUBSCRIBE sent for topic %.*s to broker.\n\n",
                 u16_TopicFilterLength,
                 sz_TopicFilter);

        /* Wait for acknowledgement packet (SUBACK) from the broker in response to the
         * subscribe request. */
        str_MqttStatus = MQTT_ProcessLoop(ptr_str_MqttContext, MQTT_PROCESS_LOOP_TIMEOUT_MS);

        if (str_MqttStatus != MQTTSuccess)
        {
            int_rc = EXIT_FAILURE;
            LOG_ERR("MQTT_ProcessLoop failed: Status=%s.",
                      MQTT_Status_strerror(str_MqttStatus));
        }
    }

    return int_rc;
}

FileScop void aws_EventCallback(MQTTContext_t *ptr_str_MqttContext, MQTTPacketInfo_t *ptr_str_PacketInfo,
                                MQTTDeserializedInfo_t *ptr_str_DeserializedInfo)
{
    uint16_t u16_PacketIdentifier;

    assert(ptr_str_MqttContext != NULL);
    assert(ptr_str_PacketInfo != NULL);
    assert(ptr_str_DeserializedInfo != NULL);

    /* Suppress unused parameter warning when asserts are disabled in build. */
    (void)ptr_str_MqttContext;

    u16_PacketIdentifier = ptr_str_DeserializedInfo->packetIdentifier;

    /* Handle incoming publish. The lower 4 bits of the publish packet
     * type is used for the dup, QoS, and retain flags. Hence masking
     * out the lower bits to check if the packet is publish. */
    if ((ptr_str_PacketInfo->type & 0xF0U) == MQTT_PACKET_TYPE_PUBLISH)
    {
        assert(ptr_str_DeserializedInfo->pPublishInfo != NULL);
        /* Handle incoming publish. */
        aws_HandleIncomingPublish(ptr_str_DeserializedInfo->pPublishInfo);
    }
    else
    {
        /* Handle other packets. */
        switch (ptr_str_PacketInfo->type)
        {
        case MQTT_PACKET_TYPE_SUBACK:

            /* A SUBACK from the broker, containing the server response to our subscription request, has been received.
             * It contains the status code indicating server approval/rejection for the subscription to the single topic
             * requested. The SUBACK will be parsed to obtain the status code, and this status code will be stored in global
             * variable str_GlobalSubAckStatus. */
            aws_UpdateSubAckStatus(ptr_str_PacketInfo);

            break;

        case MQTT_PACKET_TYPE_UNSUBACK:

            break;

        case MQTT_PACKET_TYPE_PINGRESP:

            /* Nothing to be done from application as library handles
             * PINGRESP. */
            LOG_WRN("PINGRESP should not be handled by the application "
                     "callback when using MQTT_ProcessLoop.\n\n");
            break;

        case MQTT_PACKET_TYPE_PUBACK:
            LOG_INF("PUBACK received for packet id %u.\n\n",
                     u16_PacketIdentifier);

            break;

        /* Any other packet type is invalid. */
        default:
            LOG_ERR("Unknown packet type received:(%02x).\n\n",
                      ptr_str_PacketInfo->type);
        }
    }
}

FileScop void aws_UpdateSubAckStatus(MQTTPacketInfo_t *ptr_str_PacketInfo)
{
    uint8_t *ptr_Payload = NULL;
    size_t siz_Payload = 0;

    MQTTStatus_t str_MqttStatus = MQTT_GetSubAckStatusCodes(ptr_str_PacketInfo, &ptr_Payload, &siz_Payload);

    /* MQTT_GetSubAckStatusCodes always returns success if called with packet info
     * from the event callback and non-NULL parameters. */
    assert(str_MqttStatus == MQTTSuccess);

    /* Suppress unused variable warning when asserts are disabled in build. */
    (void)str_MqttStatus;

    /* Demo only subscribes to one topic, so only one status code is returned. */
    str_GlobalSubAckStatus = (MQTTSubAckStatus_t)ptr_Payload[0];
}

//EOF
