#include "include/CCA_EMS.h"
#include "include/CCA_Dcdc.h"
#include "include/CCA_Gw.h"
#include "include/CCA_Hvac.h"
#include "include/CCA_IccCalls.h"
#include "include/CCA_Internal.h"
#include "include/CCA_Pcs.h"
#include "include/CCA_Pdu.h"
#include "include/CCA_Pmu.h"
#include "include/CCA_Version.h"
#include "lion/log/log.h"

LOG_MODULE_REGISTER(CCA_ems, LOG_LEVEL_INF);

/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

//---------------------------------------//
// Section: Defines/Enums/Typedefs       //
//---------------------------------------//

#define VALUE_SIZE 20

#define cca_siz_EMS_PARAMS_SIZE sizeof(SCH_EmsParameters_t)

/**
 * @brief Union containing the exit statuses of the EMS component operations.
 */
typedef union
{
    struct
    {
        uint8_t pcs:  1; //Exit status for PCS operations
        uint8_t pdu:  1; //Exit status for PDU operations
        uint8_t pmu:  1; //Exit status for PMU operations
        uint8_t dcdc: 1; //Exit status for DCDC operations
        uint8_t hvac: 1; //Exit status for HVAC operations
        uint8_t ems:  1; //Exit status for EMS operations
        uint8_t vrsn: 1; //Exit status for Version Info operations
        uint8_t gnrl: 1; //General status, used for error cases not covered by the other statuses.
    };
    uint8_t all;
} cca_ems_OpStatus_t;

//---------------------------------------//
// Section: Private Variables/Constants  //
//---------------------------------------//

FileScop SCH_EmsParameters_t str_EmsParameters = {0};

FileScop const CCA_KeyToSchemaParamMap_t arr_str_EmsCoreData[] = {
    #define X_EMS(param) {#param, (int32_t*)&str_EmsParameters.coreData.param, NULL},
    EMS_CORE_DATA_XMACRO
    #undef X_EMS
};
FileScop const size_t siz_NUM_CORE_KEYS = sizeof(arr_str_EmsCoreData) / sizeof(arr_str_EmsCoreData[0]);

FileScop const CCA_KeyToSchemaCodeMap_t arr_str_EmsAlarmCodes[] = {
    #define X_EMS(code) {#code, offsetof(SCH_EmsCodes_t, alarm.code)},
    EMS_CODES_ALARM_XMACRO
    #undef X_EMS
};
FileScop const size_t siz_NUM_ALARM_KEYS = sizeof(arr_str_EmsAlarmCodes) / sizeof(arr_str_EmsAlarmCodes[0]);

FileScop const CCA_KeyToSchemaCodeMap_t arr_str_EmsFaultCodes[] = {
    #define X_EMS(code) {#code, offsetof(SCH_EmsCodes_t, fault.code)},
    EMS_CODES_FAULT_XMACRO
    #undef X_EMS
};
FileScop const size_t siz_NUM_FAULT_KEYS = sizeof(arr_str_EmsFaultCodes) / sizeof(arr_str_EmsFaultCodes[0]);

//---------------------------------------//
// Section: Function Prototypes          //
//---------------------------------------//

/**
 * @brief Retrieves the EMS parallel cabinet configuration from DCU and 
 * stores it in @ref cca_sz_ems_parallel_cfg.
 * 
 * @return EXIT_SUCCESS or 0 upon success, standard POSIX error code upon failure.
*/
FileScop int cca_ems_GetParallelConfiguration(void);

/**
 * @brief Triggers product data for the EMS (CORE) to be read and populated in the JSON document provided.
 * 
 * @param jd_ProductData JSON document to store the product data in.
 * @param ja_Alarms JSON array to store the alarms in.
 * @param ja_Faults JSON array to store the faults in.
 * @returns EXIT_SUCCESS or 0 upon success, standard POSIX error code upon failure.
 */
FileScop int cca_ems_GetProductData(ArduinoJson::JsonDocument& jd_ProductData, ArduinoJson::JsonArray& ja_Alarms,
                                    ArduinoJson::JsonArray& ja_Faults);

/**
 * @brief Publishes the EMS settings to AWS.
 * 
 * @return EXIT_SUCCESS or 0 upon success, standard POSIX error code upon failure.
 */
FileScop int cca_ems_PublishEMSSettings(void);

/*********************************************************************************************************************
* @cond                                              MODULE SCOPE                                             @endcond
*********************************************************************************************************************/

ModScop void CCA_EMS_Init(void)
{
    memset(&str_EmsParameters, 0, sizeof(str_EmsParameters));
    return;
}

ModScop int CCA_EMS_InitShadows(void)
{
    if (CCA_Int_InitUnitQuantities() != EXIT_SUCCESS)
    {
        LOG_ERR("Failed to initialize unit quantities.");
        return -ECANCELED;
    }

    StackScop cca_ems_OpStatus_t str_InitStatus = {0};

    //Initialize the PCS shadows
    if (CCA_Int_GetUnitQuantity(PCS) > 0)
    {
        str_InitStatus.pcs = (CCA_Pcs_InitShadows() != EXIT_SUCCESS);
    }
    else
    {
        str_InitStatus.pcs = EXIT_SUCCESS;
        LOG_INF("pcs_settings shadow initialization skipped (No PCS units)");
    }

    //Initialize the DCDC shadows
    if (CCA_Int_GetUnitQuantity(DCDC) > 0)
    {
        str_InitStatus.dcdc = (CCA_Dcdc_InitShadows() != EXIT_SUCCESS);
    }
    else
    {
        str_InitStatus.dcdc = EXIT_SUCCESS;
        LOG_INF("dcdc_settings shadow initialization skipped (No DCDC units)");
    }

    //Initialize the HVAC shadows
    if (CCA_Int_GetUnitQuantity(HVAC) > 0)
    {
        str_InitStatus.hvac = (CCA_Hvac_InitShadows() != EXIT_SUCCESS);
    }
    else
    {
        str_InitStatus.hvac = EXIT_SUCCESS;
        LOG_INF("hvac_settings shadow initialization skipped (No HVAC units)");
    }

    //Initialize the EMS shadows
    // Subscribe to EMS_Settings Shadow sub-topics
    str_InitStatus.ems = (CCA_AWS_SubscribeToShadowSubtopics(CCA_sz_EMS_SHADOW_NAME, cca_sz_ems_update_delta_topic,
                            cca_sz_ems_update_accepted_topic, cca_sz_ems_update_rejected_topic) != EXIT_SUCCESS);
    // Publish EMS Settings
    if (str_InitStatus.ems == EXIT_SUCCESS)
    {
        str_InitStatus.ems = (cca_ems_PublishEMSSettings() != EXIT_SUCCESS);
    }
    else
    {
        LOG_ERR("Failed to initialize ems_settings shdadow.\n");
    }

    //Initialize the PMU shadows
    bool bol_IsEmsParallelSlave = false;
    if (cca_ems_GetParallelConfiguration() == EXIT_SUCCESS)
    {
        bol_IsEmsParallelSlave = !strcmp(cca_sz_ems_parallel_cfg, CCA_sz_EMS_PARALLEL_SLAVE);
    }
    if (!bol_IsEmsParallelSlave) //Go ahead and initialize PMU shadow if configured as anything besides parallel slave
    {
        if (CCA_Int_GetUnitQuantity(PMU) > 0)
        {
            str_InitStatus.pmu = (CCA_Pmu_InitShadows() != EXIT_SUCCESS);
        }
        else
        {
            str_InitStatus.pmu = EXIT_SUCCESS;
            LOG_INF("pmu_settings shadow initialization skipped (No PMU units)");
        }
    }
    else
    {
        LOG_INF("pmu_settings shadow initialization skipped (EMS Parallel Slave)");
    }

    //Initialize the Version shadows
    str_InitStatus.vrsn = (CCA_Version_InitShadows() != EXIT_SUCCESS);

    // Subscribe to Command Request Topic
    str_InitStatus.gnrl = (CCA_AWS_SubscribeToCommandRequestTopic() != EXIT_SUCCESS);
    if (str_InitStatus.gnrl == EXIT_SUCCESS)
    {
        LOG_INF("Subcribed to Command Request topic.");
    }
    else
    {
        LOG_ERR("Failed to subcribe to Command Request topic");
    }
    
    if (str_InitStatus.all != EXIT_SUCCESS)
    {
        LOG_INF("Shadow initialization(s) failed: 0x%02X", str_InitStatus.all);
        return -EXIT_FAILURE;
    }
    return EXIT_SUCCESS;
}

ModScop int CCA_EMS_UpdateComponentData(void)
{
    int int_rc = 0;
    /* PCS Data */
    if ((int_rc = CCA_Pcs_UpdateComponentData(COD_u8_NULL_UNIT_ID)) != EXIT_SUCCESS)
    {
        LOG_ERR("Failed to update PCS component data with error: %i", int_rc);
    }

    /* PDU Data */
    if ((int_rc = CCA_Pdu_UpdateComponentData(COD_u8_NULL_UNIT_ID)) != EXIT_SUCCESS)
    {
        LOG_ERR("Failed to update PDU component data with error: %i", int_rc);
    }

    /* DC/DC Data */
    if ((int_rc = CCA_Dcdc_UpdateComponentData(COD_u8_NULL_UNIT_ID)) != EXIT_SUCCESS)
    {
        LOG_ERR("Failed to update DC/DC component data with error: %i", int_rc);
    }

    /* PMU Data */
    if ((int_rc = CCA_Pmu_UpdateComponentData(COD_u8_NULL_UNIT_ID)) != EXIT_SUCCESS)
    {
        LOG_ERR("Failed to update PMU component data with error: %i", int_rc);
    }

    /* HVAC Data */
    if ((int_rc = CCA_Hvac_UpdateComponentData(COD_u8_NULL_UNIT_ID)) != EXIT_SUCCESS)
    {
        LOG_ERR("Failed to update HVAC component data with error: %i", int_rc);
    }

    return 0;
}

ModScop int CCA_EMS_UpdateProductData(void)
{  
    char sz_MqttPublishTopic[60];
    sprintf(sz_MqttPublishTopic, "powersave/product/%s", cca_sz_thing_name);

    ArduinoJson::StaticJsonDocument<CCA_u16_PRODUCT_JSON_DOC_SIZE> jd_Temp; //We want alarms and faults at the bottom
    ArduinoJson::JsonArray ja_TempAlarms = jd_Temp.createNestedArray();     //of the JSON Document, so we need an 
    ArduinoJson::JsonArray ja_TempFaults = jd_Temp.createNestedArray();     //intermediary so we can get codes as we 
                                                                            //read data from devices.
    ArduinoJson::StaticJsonDocument<CCA_u16_PRODUCT_JSON_DOC_SIZE> jd_ProductData;
    jd_ProductData["id"] = cca_sz_thing_name;
    jd_ProductData["name"] = cca_sz_product_name;

    // Get EMS (Core) Data
    cca_ems_GetProductData(jd_ProductData, ja_TempAlarms, ja_TempFaults);
    
    // Get Total PCS Data
    jd_ProductData["pcs_qty"] = CCA_Int_GetUnitQuantity(PCS);
    CCA_Pcs_GetProductData(jd_ProductData, ja_TempAlarms, ja_TempFaults);

    // Get Total PDU Data
    jd_ProductData["pdu_qty"] = CCA_Int_GetUnitQuantity(PDU);
    CCA_Pdu_GetProductData(jd_ProductData, ja_TempAlarms, ja_TempFaults);

    // Get Total DC/DC Data
    jd_ProductData["dcdc_qty"] = CCA_Int_GetUnitQuantity(DCDC);
    CCA_Dcdc_GetProductData(jd_ProductData, ja_TempAlarms, ja_TempFaults);

    // Get Total PMU Data
    CCA_Pmu_GetProductData(jd_ProductData, ja_TempAlarms, ja_TempFaults);
    
    // Send the current PeaK Shaving and Power Factor Correction Set Points
    static const uint8_t SET_POINT_NUM_KEYS = 2;
    const char *arr_sz_SetPointKeys[] = { "ps_sp", "pfc_sp"};
    char *arr_sz_SetPointResponses[SET_POINT_NUM_KEYS] = { NULL };
    if (CCA_Icc_GetEMSConfiguration(CCA_sz_UNIT_TYPE_EMS, COD_u8_NULL_UNIT_ID, SET_POINT_NUM_KEYS, arr_sz_SetPointKeys, 
                                    arr_sz_SetPointResponses) != EXIT_SUCCESS)
    {
        LOG_INF("Failed to CCA_Icc_GetEMSConfiguration for %s, index: %u", CCA_sz_UNIT_TYPE_EMS, COD_u8_NULL_UNIT_ID);
    }
    for (uint8_t j = 0; j < SET_POINT_NUM_KEYS; j++)
    {
        char sz_TempValue[VALUE_SIZE];
        if (arr_sz_SetPointResponses[j] != NULL)
        {
            sprintf(sz_TempValue, "%s", arr_sz_SetPointResponses[j]);
        }
        else 
        {
            sprintf(sz_TempValue, "%s", "");
        }
        jd_ProductData[arr_sz_SetPointKeys[j]] = sz_TempValue;
    }

    // Get Cabinet Data from the Gateway
    jd_ProductData["hvac_qty"] = CCA_Int_GetUnitQuantity(HVAC);
    CCA_Gw_GetProductData(jd_ProductData, ja_TempAlarms, ja_TempFaults);

    // Add the temp code arrays to the productDataDoc
    jd_ProductData["alarms"] = ja_TempAlarms;
    jd_ProductData["faults"] = ja_TempFaults;

    char sz_ProductData[CCA_u16_PRODUCT_JSON_DOC_SIZE];
    ArduinoJson::serializeJson(jd_ProductData, sz_ProductData);
    // Validate the JSON prior to publish
    if (!jd_ProductData.isNull())
    {
        CCA_AWS_PublishToTopic(&str_MqttContext, sz_MqttPublishTopic, strlen(sz_MqttPublishTopic),
                       sz_ProductData, strlen(sz_ProductData));
    }
    else
    {
        // Print the error messages
        printf("ERROR: INVALID JSON during AWS_PublishProductData()");
        printf("%s\n", sz_ProductData);
    }
    return 0; //TODO Return something meaningful
}

/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

FileScop int cca_ems_GetParallelConfiguration(void)
{
    const char *sz_EmsParallelKey[] = {"parallel_cfg"};
    char *sz_EmsResponseBuffer[1] = {NULL};
    // Retrieve the current settings from the EMS
    int int_rc = CCA_Icc_GetEMSConfiguration(CCA_sz_UNIT_TYPE_EMS, COD_u8_NULL_UNIT_ID, 1, sz_EmsParallelKey, sz_EmsResponseBuffer);
    if (int_rc != 0)
    {
        LOG_ERR("Failed to get EMS parallel configuration, with Error: %i", int_rc);
        return int_rc;
    }

    if (sz_EmsResponseBuffer[0] == NULL)
    {
        int_rc = -ENOENT;
        LOG_ERR("Failed to get EMS parallel configuration, with Error: %i", int_rc);
    }
    else
    {
        sprintf(cca_sz_ems_parallel_cfg, "%s", sz_EmsResponseBuffer[0]);
        LOG_INF("Retrieved EMS parallel configuration: %s", cca_sz_ems_parallel_cfg);
    }
    return int_rc;
}

FileScop int cca_ems_GetProductData(ArduinoJson::JsonDocument& jd_ProductData, ArduinoJson::JsonArray& ja_Alarms,
                                    ArduinoJson::JsonArray& ja_Faults)
{
    StackScop uint32_t u32_EmsDataSize = cca_siz_EMS_PARAMS_SIZE;
    StackScop bool bol_CommFailure = false;
    StackScop int int_rc = CCA_Icc_GetBufferedData(CCA_sz_UNIT_TYPE_EMS, 0, &str_EmsParameters, &u32_EmsDataSize);
    if (int_rc != 0 || u32_EmsDataSize != cca_siz_EMS_PARAMS_SIZE)
    {
        LOG_INF("Failed to getEMSData for %s, index: %u, bytes: %d, with Error: %i",
                    CCA_sz_UNIT_TYPE_EMS, 0, u32_EmsDataSize, int_rc);
        bol_CommFailure = true;
    }

    //Populate the JSON Document with the EMS (Core) Data
    CCA_Int_AppendToJsonDoc(jd_ProductData, arr_str_EmsCoreData, siz_NUM_CORE_KEYS, bol_CommFailure);
    //Add the EMS (Core) codes to their respective arrays
    if (!bol_CommFailure)
    {
        CCA_Int_AppendCodesToJsonArray(ja_Alarms, arr_str_EmsAlarmCodes, siz_NUM_ALARM_KEYS, &str_EmsParameters.codes);
        CCA_Int_AppendCodesToJsonArray(ja_Faults, arr_str_EmsFaultCodes, siz_NUM_FAULT_KEYS, &str_EmsParameters.codes);
    }
    return (int_rc < 0) ? int_rc : (bol_CommFailure) ? -EIO : 0;
}

FileScop int cca_ems_PublishEMSSettings(void)
{
    ArduinoJson::StaticJsonDocument<CCA_u16_BIG_SETTINGS_JSON_DOC_SIZE> jd_EmsSettingsReported;
    ArduinoJson::JsonObject jo_Product = jd_EmsSettingsReported["state"]["reported"].createNestedObject("product");
    jo_Product["id"] = cca_sz_thing_name;
    jo_Product["name"] = cca_sz_product_name;
    ArduinoJson::JsonObject jo_ProductDatum = jo_Product.createNestedObject("datum");
    const char *settingKeys[] = {
        "sys_work_md", "cmsn_md_en", "diag_md_en", "on_grid_dod", "off_grid_dod", 
        "chg_hyst", "dchg_hyst", "grid_sell", "grid_bf_kW", "mstr_grid_ch", "ps_sp", "ps_prop", "ps_integ", 
        "ps_deriv", "pfc_en", "pfc_p_min", "pfc_sp", "pfc_prop", "pfc_integ", "pfc_deriv", 
        "ch_i_max", "ch_i_derat_slope", "ch_pro_cell_v_1", "ch_pro_i_1", "ch_pro_cell_v_2", 
        "ch_pro_i_2", "ch_pro_cell_v_3", "ch_pro_i_3", "ch_pro_prop", "ch_pro_integ", 
        "ch_pro_deriv", "sleep_time_lim", "sleep_md_en", "pcs_qty", "dcdc_qty", "pdu_qty", 
        "hvac_qty", "pmu_qty", "ip_addr", "netmask", "parallel_cfg", "parallel_qty", "ls_en", "timezone"
    };
    uint8_t NUM_KEYS = sizeof(settingKeys) / sizeof(settingKeys[0]);
    char* sz_EmsResponseBuffer[NUM_KEYS] = {NULL};
    // Retrieve the current settings from the EMS
    int int_rc = CCA_Icc_GetEMSConfiguration(CCA_sz_UNIT_TYPE_EMS, COD_u8_NULL_UNIT_ID, NUM_KEYS, settingKeys, 
                                                sz_EmsResponseBuffer);
    if (int_rc != 0)
    {
        LOG_INF("Failed to getEMSData for %s, index: %u, with Error: %i",
                CCA_sz_UNIT_TYPE_EMS, COD_u8_NULL_UNIT_ID, int_rc);
    }
    for (uint8_t i = 0; i < NUM_KEYS; i++)
    {
        char sz_TempValue[VALUE_SIZE];
        if (sz_EmsResponseBuffer[i] != NULL)
        {
            sprintf(sz_TempValue, "%s", sz_EmsResponseBuffer[i]);
        }
        else 
        {
            sprintf(sz_TempValue, "%s", "");
        }
        jo_ProductDatum[settingKeys[i]] = sz_TempValue;
    }

    // Get Time of Use Profile Settings
    CCA_Int_AppendTouSettingsToJsonObject(jo_ProductDatum);

    const char* arr_sz_ModbusUnits[] = { 
        CCA_sz_UNIT_TYPE_PCS, CCA_sz_UNIT_TYPE_DCDC, CCA_sz_UNIT_TYPE_HVAC, CCA_sz_UNIT_TYPE_GW
    };
    for (uint8_t i = 0; i < sizeof(arr_sz_ModbusUnits)/sizeof(arr_sz_ModbusUnits[0]); i++)
    {
        // Build Communication Settings based on Unit Type
        const char* sz_UnitType = arr_sz_ModbusUnits[i];
        ArduinoJson::JsonArray ja_Unit = jo_ProductDatum.createNestedArray(sz_UnitType);
        uint8_t u8_SettingsQty = CCA_Int_GetUnitQuantity_sz(arr_sz_ModbusUnits[i]);
        
        //Get communications settings for all components of unit type
        for (uint8_t j = 0; j < u8_SettingsQty; j++)
        {
            ArduinoJson::JsonObject jo_UnitObject = ja_Unit.createNestedObject();
            char sz_UnitID[VALUE_SIZE];
            sprintf(sz_UnitID, "%u", j);
            jo_UnitObject["id"] = sz_UnitID;
            ArduinoJson::JsonObject jo_UnitDatum = jo_UnitObject.createNestedObject("datum");
            const char *unitSettingKeys[] = {"ip_addr", "port", "mb_addr"};
            uint8_t NUM_KEYS = sizeof(unitSettingKeys) / sizeof(unitSettingKeys[0]);
            char* sz_UnitResponseBuffer[NUM_KEYS] = {NULL};
            int_rc = CCA_Icc_GetEMSConfiguration(sz_UnitType, j, NUM_KEYS, unitSettingKeys, sz_UnitResponseBuffer);
            if (int_rc != 0)
            {
                LOG_INF("Failed to getEMSData for %s, index: %u, with Error: %i", sz_UnitType, j, int_rc);
            }
            for (uint8_t k = 0; k < NUM_KEYS; k++)
            {
                char sz_TempValue[VALUE_SIZE];
                if (sz_UnitResponseBuffer[k] != NULL)
                {
                    sprintf(sz_TempValue, "%s", sz_UnitResponseBuffer[k]);
                }
                else 
                {
                    sprintf(sz_TempValue, "%s", "");
                }
                jo_UnitDatum[unitSettingKeys[k]] = sz_TempValue;
            }
        }
    }
    
    char sz_EmsSettingsReport[CCA_u16_BIG_SETTINGS_JSON_DOC_SIZE];
    ArduinoJson::serializeJson(jd_EmsSettingsReported, sz_EmsSettingsReport);
    LOG_INF("emsSettingsReport: %s", sz_EmsSettingsReport);
    int_rc = CCA_AWS_PublishToTopic(&str_MqttContext, cca_sz_ems_update_topic, strlen(cca_sz_ems_update_topic),
                                        sz_EmsSettingsReport, strlen(sz_EmsSettingsReport));
    return int_rc;
}

//EOF
