#include "include/CCA_IccCalls.h"
#include "include/CCA_Version.h"
#include "lion/log/log.h"
#include <ArduinoJson.h>
#include <zcbor_decode.h>
#include <zcbor_encode.h>

LOG_MODULE_REGISTER(CCA_IccCalls, LOG_LEVEL_INF);

/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

FileScop ICC_Function_Handler_t str_AlertStatusHandler;
FileScop ICC_Function_Handler_t str_VersionInfoHandler;

/**
 * @brief   Packages and sends any new and/or cleared alert codes to AWS for the respective component requested, based on 
 *          an array of strings
 * 
 * @param enm_UnitType          "pcs", "pdu", "dcdc", "pmu", "hvac", "gw"
 * @param u8_UnitIndex          (0-2), 255 (Null)
 * @param sz_arr_NewAlerts      Array of new alert codes that have opened since the last update
 * @param u8_NewAlertsQty       Quantity of new alerts to be sent
 * @param sz_arr_ClearedAlerts  Array of cleared alert codes that have closed since the last update
 * @param u8_ClearedAlertsQty   Quantity of cleared alerts to be sent
 * 
 * 
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
FileScop int cca_Icc_SendAlerts(unit_type_t enm_UnitType, uint8_t u8_UnitIndex, char **sz_arr_NewAlerts, 
                                uint8_t u8_NewAlertsQty, char **sz_arr_ClearedAlerts, uint8_t u8_ClearedAlertsQty);


/**
 * @brief The ICC callback for sending new and cleared alert status data to AWS
 *
 * @details Decodes CBOR data and calls the appropriate function
 *
 * @param u32_ID ICC Function ID
 * @param vptr_Data Pointer to CBOR data
 * @param u32_DataLength Length of CBOR data
 * @param vptr_Response Pointer to response buffer
 * @param ptr_u32_responseLength Pointer to response length
 * @param vptr_UserData Pointer to user data
 *
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
FileScop int cca_Icc_AlertStatusCallback(uint32_t u32_ID, const void *vptr_Data, uint32_t u32_DataLength, 
                                        void *vptr_Response, uint32_t *ptr_u32_responseLength, void *vptr_UserData);

/**
 * @brief The ICC callback for sending device version info
 *
 * @details Decodes CBOR data and saves the version info into the version shadow JSON document
 *
 * @param u32_ID ICC Function ID
 * @param vptr_Data Pointer to CBOR data
 * @param u32_DataLength Length of CBOR data
 * @param vptr_Response Pointer to response buffer
 * @param ptr_u32_ResponseLength Pointer to response length
 * @param vptr_UserData Pointer to user data
 *
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
FileScop int cca_Icc_VersionInfoCallback(uint32_t u32_ID, const void *vptr_Data, uint32_t u32_DataLength,
                                            void *vptr_Response, uint32_t *ptr_u32_ResponseLength, void *vptr_UserData);

/*********************************************************************************************************************
* @cond                                              MODULE SCOPE                                             @endcond
*********************************************************************************************************************/

ModScop void CCA_Icc_Init(void)
{
    return; //Nothing to initialize
}

ModScop void CCA_Icc_RegisterCallbackHandlers(void)
{
    IAL_RegisterCallback(IAL_ALERT_STATUS_ID, &str_AlertStatusHandler, cca_Icc_AlertStatusCallback);
    IAL_RegisterCallback(IAL_VERSION_INFO_ID, &str_VersionInfoHandler, cca_Icc_VersionInfoCallback);
}
ModScop void CCA_Icc_UnregisterCallbackHandlers(void)
{
    IAL_UnregisterCallback(&str_AlertStatusHandler);
    IAL_UnregisterCallback(&str_VersionInfoHandler);
}

/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

FileScop int cca_Icc_SendAlerts(unit_type_t enm_UnitType, uint8_t u8_UnitIndex, char **arr_sz_NewAlerts, 
                                uint8_t u8_NewAlertsQty, char **arr_sz_ClearedAlerts, uint8_t u8_ClearedAlertsQty)
{
    ArduinoJson::StaticJsonDocument<CCA_u16_ALERT_JSON_DOC_SIZE> jd_AlertDoc;
    ArduinoJson::JsonArray ja_AlertingComponentsArray = jd_AlertDoc.createNestedArray();
    ArduinoJson::JsonObject jo_AlertingComponent = ja_AlertingComponentsArray.createNestedObject();

    // Create and fill in the header of the alert message
    switch (enm_UnitType) 
    {
        case PCS:
        {
            jo_AlertingComponent["id"] = cca_arr_pcs_component_uuids[u8_UnitIndex];
            jo_AlertingComponent["id_type"] = "component";
            jo_AlertingComponent["type"] = CCA_sz_UNIT_TYPE_PCS;
            break;
        }
        case PDU:
        {
            jo_AlertingComponent["id"] = cca_arr_pdu_component_uuids[u8_UnitIndex];
            jo_AlertingComponent["id_type"] = "component";
            jo_AlertingComponent["type"] = CCA_sz_UNIT_TYPE_PDU;
            break;
        }
        case PMU:
        {
            jo_AlertingComponent["id"] = cca_arr_pmu_component_uuids[u8_UnitIndex];
            jo_AlertingComponent["id_type"] = "component";
            jo_AlertingComponent["type"] = CCA_sz_UNIT_TYPE_PMU;
            break;
        }
        case DCDC:
        {
            jo_AlertingComponent["id"] = cca_arr_dcdc_component_uuids[u8_UnitIndex];
            jo_AlertingComponent["id_type"] = "component";
            jo_AlertingComponent["type"] = CCA_sz_UNIT_TYPE_DCDC;
            break;
        }
        case HVAC:
        {
            jo_AlertingComponent["id"] = cca_arr_hvac_component_uuids[u8_UnitIndex];
            jo_AlertingComponent["id_type"] = "component";
            jo_AlertingComponent["type"] = CCA_sz_UNIT_TYPE_HVAC;
            break;
        }
        case EMS:   // Product (System level) alerts
        {
            jo_AlertingComponent["id"] = cca_sz_thing_name;
            jo_AlertingComponent["id_type"] = "product";
            jo_AlertingComponent["type"] = "cabinet";
            break;
        }
        case GW:    // Product (System level) alerts
        {
            jo_AlertingComponent["id"] = cca_sz_thing_name;
            jo_AlertingComponent["id_type"] = "product";
            jo_AlertingComponent["type"] = "cabinet";
            break;
        }
        default:
        {
            return -ENODEV;
        }
    }
    ArduinoJson::JsonArray ja_AlertsArray = jo_AlertingComponent.createNestedArray("alerts");

    // Add the New Alerts
    for (uint8_t i = 0; i < u8_NewAlertsQty; i++) 
    {
        ArduinoJson::JsonObject jo_NewAlert = ja_AlertsArray.createNestedObject();
        jo_NewAlert["code"] = arr_sz_NewAlerts[i];
        jo_NewAlert["status"] = "open";
    }

    // Add the Cleared Alerts
    for (uint8_t i = 0; i < u8_ClearedAlertsQty; i++) 
    {
        ArduinoJson::JsonObject jo_ClearedAlert = ja_AlertsArray.createNestedObject();
        jo_ClearedAlert["code"] = arr_sz_ClearedAlerts[i];
        jo_ClearedAlert["status"] = "closed";
    }

    // Send the alert to AWS
    char sz_AlertMessage[CCA_u16_ALERT_JSON_DOC_SIZE];
    ArduinoJson::serializeJson(jd_AlertDoc, sz_AlertMessage);
    int int_rc = CCA_AWS_PublishToTopic(&str_MqttContext, cca_sz_alerts_topic, strlen(cca_sz_alerts_topic), 
                                        sz_AlertMessage, strlen(sz_AlertMessage));
    if (int_rc == EXIT_SUCCESS)
    {
        MQTTStatus_t mqttStatus = MQTT_ProcessLoop(&str_MqttContext, MQTT_PROCESS_LOOP_TIMEOUT_MS);

        if (mqttStatus != MQTTSuccess)
        {
            LOG_WRN("MQTT_ProcessLoop failed: Error = %s.", MQTT_Status_strerror(mqttStatus));
            int_rc = -EIO;
        }
        else
        {
            LOG_INF("Successfully sent alert for %s id:%d.", jo_AlertingComponent["type"].as<const char*>(), u8_UnitIndex);
            int_rc = EXIT_SUCCESS;
        }
    }
    
    if (int_rc != EXIT_SUCCESS)
    {
        LOG_WRN("Failed to send alert for %s id:%d, Error = %d", 
                jo_AlertingComponent["type"].as<const char*>(), u8_UnitIndex, int_rc);
    }
    return int_rc;
}

FileScop int cca_Icc_AlertStatusCallback(uint32_t u32_ID, const void *ptr_Data, uint32_t u32_DataLength, 
                                        void *vptr_Response, uint32_t *ptr_u32_ResponseLength, void *vptr_UserData)
{
    ZCBOR_STATE_D(CBORDecodeState, 1, (uint8_t*)ptr_Data, u32_DataLength, 4);
    int int_Rc = EXIT_SUCCESS;
    *ptr_u32_ResponseLength = 0;

    // Decode Unit Type
    struct zcbor_string sz_zcborUnitType; 
    char* sz_ReqUnitType;
    unit_type_t enm_reqUnitType;
    if (zcbor_tstr_decode(CBORDecodeState, &sz_zcborUnitType))
    {
        sz_ReqUnitType = (char *)sz_zcborUnitType.value;
        if (int_Rc = CCA_Int_ValidateUnitType(sz_ReqUnitType, &enm_reqUnitType) != EXIT_SUCCESS)
        {
            return int_Rc;
        }
    }
    else
    {
        return -EINVAL;
    }

    // Decode Unit Index
    uint32_t u32_ReqUnitIndex;
    if (zcbor_uint32_decode(CBORDecodeState, &u32_ReqUnitIndex))
    {
        // Check if device index is in range
        if ((int_Rc = CCA_Int_ValidateUnitIndex(enm_reqUnitType, u32_ReqUnitIndex)) != EXIT_SUCCESS)
        {
            return int_Rc;
        }
    }
    else
    {
        return -EINVAL;
        
    }

    // Decode List of New Alerts
    char *arr_sz_NewAlerts[IAL_DATA_u8_ALERT_BUFFER_SIZE];
    uint8_t u8_ReqNewAlertsQty = 0;
    if (!zcbor_list_start_decode(CBORDecodeState))
    {
        return -EINVAL;
    }
    struct zcbor_string sz_zcborNewAlertCode; 
    for (uint32_t i = 0; zcbor_tstr_decode(CBORDecodeState, &sz_zcborNewAlertCode); i++)
    {
        arr_sz_NewAlerts[i] = (char *)sz_zcborNewAlertCode.value;
        u8_ReqNewAlertsQty++;
    }
    zcbor_list_end_decode(CBORDecodeState);

    // Decode List of Cleared Alerts
    char *arr_sz_ClearedAlerts[IAL_DATA_u8_ALERT_BUFFER_SIZE];
    uint8_t u8_ReqClearedAlertsQty = 0;
    if (!zcbor_list_start_decode(CBORDecodeState))
    {
        return -EINVAL;
    }
    struct zcbor_string sz_zcborClearedAlertCode; 
    for (uint32_t i = 0; zcbor_tstr_decode(CBORDecodeState, &sz_zcborClearedAlertCode); i++)
    {
        arr_sz_ClearedAlerts[i] = (char *)sz_zcborClearedAlertCode.value;
        u8_ReqClearedAlertsQty++;
    }
    zcbor_list_end_decode(CBORDecodeState);

    // Send All Alerts to AWS
    return cca_Icc_SendAlerts(enm_reqUnitType, (uint8_t)u32_ReqUnitIndex, arr_sz_NewAlerts, u8_ReqNewAlertsQty, 
                            arr_sz_ClearedAlerts, u8_ReqClearedAlertsQty);
}


FileScop int cca_Icc_VersionInfoCallback(uint32_t u32_ID, const void *vptr_Data, uint32_t u32_DataLength,
                                            void *vptr_Response, uint32_t *ptr_u32_ResponseLength, void *vptr_UserData)
{
    LOG_INF("Callback for Version Info");
    //Decode keys/values from the CBOR encoded data
    ZCBOR_STATE_D(CBORDecodeState, 1, (uint8_t*)vptr_Data, u32_DataLength, 1);
    if (zcbor_map_start_decode(CBORDecodeState) == 0)
    {
        LOG_INF("Decoding CBOR Error");
        return -EINVAL;
    }
    // Decode the key value pairs and store them into a json object
    struct zcbor_string sz_zcborKey;
    struct zcbor_string sz_zcborValue;
    ArduinoJson::StaticJsonDocument<CCA_u16_SETTINGS_JSON_DOC_SIZE> jd_VersionInfo;
    for (uint32_t i = 0; zcbor_tstr_decode(CBORDecodeState, &sz_zcborKey); i++)
    {
        zcbor_tstr_decode(CBORDecodeState, &sz_zcborValue);
        jd_VersionInfo[sz_zcborKey.value] = sz_zcborValue.value;
    }
    zcbor_map_end_decode(CBORDecodeState);

    ArduinoJson::JsonObject jo_VersionInfoAsObj = jd_VersionInfo.as<ArduinoJson::JsonObject>();
    int int_rc = CCA_Version_UpdateVersionInfoWith(jo_VersionInfoAsObj);
    *ptr_u32_ResponseLength = 0;
    return int_rc;
}

//EOF
