#include "include/CCA_Pdu.h"
#include "include/CCA_Internal.h"
#include "include/CCA_IccCalls.h"
#include "lion/log/log.h"

#define CCA_PDU_PARAMS_SIZE sizeof(SCH_PduParameters_t)
#define VALUE_SIZE 10

LOG_MODULE_REGISTER(CCA_pdu, LOG_LEVEL_INF);

/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

FileScop SCH_PduParameters_t str_PduParameters;

FileScop const CCA_KeyToSchemaParamMap_t arr_str_PduProdData[] = {
    {"pdu_v", (int32_t *)&str_PduParameters.signals.pdu_v, NULL},
    {"pdu_soc", (int32_t *)&str_PduParameters.signals.pdu_soc, NULL},
    {"pdu_i", (int32_t *)&str_PduParameters.signals.pdu_i, NULL},
    {"pdu_soh", (int32_t *)&str_PduParameters.signals.pdu_soh, NULL},
    {"ch_cap_day", (int32_t *)&str_PduParameters.signals.ch_cap_day, NULL},
    {"dch_cap_day", (int32_t *)&str_PduParameters.signals.dch_cap_day, NULL},
    {"ch_cap_total", (int32_t *)&str_PduParameters.signals.ch_cap_total, NULL},
    {"dch_cap_total", (int32_t *)&str_PduParameters.signals.dch_cap_total, NULL},
    {"bms_qty", (int32_t *)&str_PduParameters.signals.bms_qty, NULL}
};
FileScop const size_t siz_NUM_PROD_KEYS = sizeof(arr_str_PduProdData) / sizeof(arr_str_PduProdData[0]);

FileScop const CCA_KeyToSchemaParamMap_t arr_str_PduCompData[] = {
    #define X_PDU(key) {#key, (int32_t *)&str_PduParameters.signals.key, NULL},
    PDU_SIGNALS_XMACRO
    #undef X_PDU
};
FileScop const size_t siz_NUM_COMP_KEYS = sizeof(arr_str_PduCompData) / sizeof(arr_str_PduCompData[0]);

FileScop const CCA_KeyToSchemaCodeMap_t arr_str_PduAlarmCodes[] = {
    #define X_PDU(code) {#code, offsetof(SCH_PduCodes_t, alarm.code)},
    PDU_CODES_ALARM_XMACRO
    #undef X_PDU
};
FileScop const size_t siz_NUM_ALARM_KEYS = sizeof(arr_str_PduAlarmCodes) / sizeof(arr_str_PduAlarmCodes[0]);

FileScop const CCA_KeyToSchemaCodeMap_t arr_str_PduFaultCodes[] = {
    #define X_PDU(code) {#code, offsetof(SCH_PduCodes_t, fault.code)},
    PDU_CODES_FAULT_XMACRO
    #undef X_PDU
};
FileScop const size_t siz_NUM_FAULT_KEYS = sizeof(arr_str_PduFaultCodes) / sizeof(arr_str_PduFaultCodes[0]);

FileScop const CCA_KeyToSchemaCodeMap_t arr_str_PduStatusCodes[] = {
    #define X_PDU(code) {#code, offsetof(SCH_PduCodes_t, status.code)},
    PDU_CODES_STATUS_XMACRO
    #undef X_PDU
};
FileScop const size_t siz_NUM_STATUS_KEYS = sizeof(arr_str_PduStatusCodes) / sizeof(arr_str_PduStatusCodes[0]);

/**
 * @brief Scales the raw SOC value to a "user" value.
 * 
 * @param i32_RawSOC The raw SOC value to scale.
 * @return int32_t The scaled SOC value.
 */
FileScop int32_t cca_pdu_ScaleSOC(int32_t i32_RawSOC);

/**
 * @brief Calculates the power based on the voltage and current.
 * 
 * @param u16_Voltage_dV The voltage in dV.
 * @param s16_Current_dA The current in dA.
 * @param sz_Power_daW Destination string for the power in daW.
 * @return int 0 upon success, standard POSIX error code upon failure.
 */
FileScop int cca_pdu_CalculatePower(uint16_t u16_Voltage_dV, int16_t s16_Current_dA, char* sz_Power_daW);

/*********************************************************************************************************************
* @cond                                              MODULE SCOPE                                             @endcond
*********************************************************************************************************************/

ModScop void CCA_Pdu_Init(void)
{
    memset(&str_PduParameters, 0, CCA_PDU_PARAMS_SIZE);
}

ModScop int CCA_Pdu_InitShadows(void)
{
    return -ENOTSUP;
}

ModScop int CCA_Pdu_UpdateComponentData(uint8_t u8_UnitIndex)
{
    if (u8_UnitIndex != COD_u8_NULL_UNIT_ID && CCA_Int_ValidateUnitIndex(PDU, u8_UnitIndex) != EXIT_SUCCESS)
    {
        return -EINVAL;
    }

    const char *sz_ProductUUID = cca_sz_thing_name;
    char sz_MqttPublishTopic[60];
    const uint8_t u8_PduQty = CCA_Int_GetUnitQuantity(PDU);
    int int_rc[u8_PduQty] = {0};
    bool bol_CommFailure[u8_PduQty] = {false};

    uint8_t u8_EndIndex = (u8_UnitIndex == COD_u8_NULL_UNIT_ID) ? u8_PduQty : u8_UnitIndex + 1;
    uint8_t i = (u8_UnitIndex == COD_u8_NULL_UNIT_ID) ? 0 : u8_UnitIndex;
    for (i; i < u8_EndIndex; i++)
    {
        ArduinoJson::StaticJsonDocument<CCA_u16_COMPONENT_JSON_DOC_SIZE> jd_PduComponentData;
        sprintf(sz_MqttPublishTopic, "powersave/component/%s", cca_arr_pdu_component_uuids[i]);
        // Build PDU Component Data document
        jd_PduComponentData["id"] = cca_arr_pdu_component_uuids[i];
        jd_PduComponentData["product_id"] = sz_ProductUUID;
        jd_PduComponentData["type"] = CCA_sz_UNIT_TYPE_PDU;
        jd_PduComponentData["serial"] = cca_arr_PDU_SERIALS[i];
        jd_PduComponentData["model"] = cca_sz_PDU_MODEL;
        jd_PduComponentData["mfr"] = cca_sz_PDU_MFR;
        bol_CommFailure[i] = false;
        uint32_t u32_PduParametersSize = CCA_PDU_PARAMS_SIZE;
         
        int_rc[i] = CCA_Icc_GetBufferedData(CCA_sz_UNIT_TYPE_PDU, i, &str_PduParameters, &u32_PduParametersSize);
        if (int_rc[i] != 0 || u32_PduParametersSize != CCA_PDU_PARAMS_SIZE)
        {
            LOG_INF("Failed to getEMSData for %s, index: %u, bytes: %d, with Error: %i", 
                        CCA_sz_UNIT_TYPE_PDU, i, u32_PduParametersSize, int_rc[i]);
            bol_CommFailure[i] = true;
        }
        //We need to scale the pdu_soc before adding it to the JSON Document
        if (!bol_CommFailure[i])
        {
            str_PduParameters.signals.pdu_soc = cca_pdu_ScaleSOC(str_PduParameters.signals.pdu_soc);
        }
        //Populate the JSON Document with the PDU Data
        CCA_Int_AppendToJsonDoc(jd_PduComponentData, arr_str_PduCompData, siz_NUM_COMP_KEYS, bol_CommFailure[i]);
        // Calculate PDU Power (not available as readable parameter)
        if (jd_PduComponentData["pdu_v"].as<const char*>() != "" || jd_PduComponentData["pdu_i"].as<const char*>() != "")
        {
            char tempValue[VALUE_SIZE];
            cca_pdu_CalculatePower(jd_PduComponentData["pdu_v"].as<uint16_t>(), jd_PduComponentData["pdu_i"].as<int16_t>(), tempValue);
            jd_PduComponentData["pdu_p"] = tempValue;
        }
        else 
        {
            jd_PduComponentData["pdu_p"] = "";
        }
        // Check for alarms, faults, and statuses
        ArduinoJson::JsonArray ja_Alarms = jd_PduComponentData.createNestedArray("alarms");
        ArduinoJson::JsonArray ja_Faults = jd_PduComponentData.createNestedArray("faults");
        ArduinoJson::JsonArray ja_Status = jd_PduComponentData.createNestedArray("status");
        CCA_Int_AppendCodesToJsonArray(ja_Alarms, arr_str_PduAlarmCodes, siz_NUM_ALARM_KEYS, &str_PduParameters.codes);
        CCA_Int_AppendCodesToJsonArray(ja_Faults, arr_str_PduFaultCodes, siz_NUM_FAULT_KEYS, &str_PduParameters.codes);
        CCA_Int_AppendCodesToJsonArray(ja_Status, arr_str_PduStatusCodes, siz_NUM_STATUS_KEYS, &str_PduParameters.codes);

        char sz_PduComponentData[CCA_u16_COMPONENT_JSON_DOC_SIZE];
        ArduinoJson::serializeJson(jd_PduComponentData, sz_PduComponentData);
        // Validate the JSON prior to publish
        if (!jd_PduComponentData.isNull())
        {
            CCA_AWS_PublishToTopic(&str_MqttContext, sz_MqttPublishTopic, strlen(sz_MqttPublishTopic),
                           sz_PduComponentData, strlen(sz_PduComponentData));
        }
        else
        {
            // Print the error messages
            LOG_ERR("ERROR: INVALID JSON during AWS_PublishComponentData()");
            LOG_ERR("%s\n", sz_PduComponentData);
        }
    }

    int update_rc = 0;
    for (uint8_t i = 0; i < u8_PduQty; i++)
    {
        if (int_rc[i] != 0)
        {
            update_rc = int_rc[i];
            break;
        }
        else if (bol_CommFailure[i])
        {
            update_rc = -EIO;
            break;
        }
    }
    return update_rc;
}

ModScop int CCA_Pdu_GetProductData(ArduinoJson::JsonDocument& jd_ProductData, ArduinoJson::JsonArray& ja_Alarms,
                                    ArduinoJson::JsonArray& ja_Faults)
{
    uint32_t u32_PduParametersSize = CCA_PDU_PARAMS_SIZE;
    const uint8_t u8_PduQty = CCA_Int_GetUnitQuantity(PDU);
    bool bol_CommFailure = false;
    int int_rc = -ENXIO;
    if (u8_PduQty > 0)
    {
        int_rc = CCA_Icc_GetBufferedData(CCA_sz_UNIT_TYPE_PDU, 0, &str_PduParameters, &u32_PduParametersSize);
    }
    if (int_rc != 0 || u32_PduParametersSize != CCA_PDU_PARAMS_SIZE)
    {
        LOG_INF("Failed to getEMSData for %s, index: %u, bytes: %d, with Error: %i", 
                    CCA_sz_UNIT_TYPE_PDU, 0, u32_PduParametersSize, int_rc);
        bol_CommFailure = true;
    }
    //We need to scale the pdu_soc before adding it to the JSON Document
    if (!bol_CommFailure)
    {
        str_PduParameters.signals.pdu_soc = cca_pdu_ScaleSOC(str_PduParameters.signals.pdu_soc);
    }
    //Populate the JSON Document with the PDU Data, but only for the following keys:
    CCA_Int_AppendToJsonDoc(jd_ProductData, arr_str_PduProdData, siz_NUM_PROD_KEYS, bol_CommFailure);
    // Calculate PDU Power (not available as readable parameter)
    if (jd_ProductData["pdu_v"].as<const char*>() != "" || jd_ProductData["pdu_i"].as<const char*>() != "")
    {
        char tempValue[VALUE_SIZE];
        cca_pdu_CalculatePower(jd_ProductData["pdu_v"].as<uint16_t>(), jd_ProductData["pdu_i"].as<int16_t>(), tempValue);
        jd_ProductData["pdu_p"] = tempValue;
    }
    else 
    {
        jd_ProductData["pdu_p"] = "";
    }
    //Add the PDU Alarms and Faults to the temp arrays
    if (!bol_CommFailure)
    {
        CCA_Int_AppendCodesToJsonArray(ja_Alarms, arr_str_PduAlarmCodes, siz_NUM_ALARM_KEYS, &str_PduParameters.codes);
        CCA_Int_AppendCodesToJsonArray(ja_Faults, arr_str_PduFaultCodes, siz_NUM_FAULT_KEYS, &str_PduParameters.codes);
    }
    return (int_rc < 0) ? int_rc : (bol_CommFailure) ? -EIO : 0;
}

/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

FileScop int32_t cca_pdu_ScaleSOC(int32_t i32_RawSOC)
{
    int32_t i32_scaledSOC = 0;
    // No need (yet) to scale data less than 90%
    if (i32_RawSOC <= 90)
    {
        i32_scaledSOC = i32_RawSOC;
    }
    else 
    {
        // 91%-94% -> 92%-98%
        if (i32_RawSOC < 95) 
        {
            i32_scaledSOC = ((i32_RawSOC - 90)*2) + 90;
        }
        else 
        {
            // 95%-100% -> 100%
            i32_scaledSOC = 100;
        }
    }
    // Safety check for invalid values
    if (i32_scaledSOC < 0) 
    {
        i32_scaledSOC = 0;
    }
    else if (i32_scaledSOC > 100)
    {
        i32_scaledSOC = 100;
    }
    return i32_scaledSOC;
}

FileScop int cca_pdu_CalculatePower(uint16_t u16_Voltage_dV, int16_t s16_Current_dA, char* sz_Power_daW)
{
    // 0.01kW = 10W = deka-Watts
    // P = IV, (0.1V*0.1A) = (0.01W) -> *0.001 = (0.01kw)
    int32_t s16_power_daW = (u16_Voltage_dV * s16_Current_dA) * 0.001;
    sprintf(sz_Power_daW, "%i", s16_power_daW);
    return EXIT_SUCCESS;
}

//EOF
