#include "include/CCA_ReqHandler.h"
#include "include/CCA_IccCalls.h"
#include "include/CCA_Dcdc.h"
#include "include/CCA_EMS.h"
#include "include/CCA_Hvac.h"
#include "include/CCA_Pcs.h"
#include "include/CCA_Pdu.h"
#include "include/CCA_Pmu.h"
#include "lion/log/log.h"
#include <time.h>

LOG_MODULE_REGISTER(CCA_ReqHandler, LOG_LEVEL_INF);

/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

#define CCA_sz_JD_ID        "id"
#define CCA_sz_JD_BODY      "body"
#define CCA_sz_JD_BD_TYPE   "type"
#define CCA_sz_JD_BD_UNIT   "unit"
#define CCA_sz_JD_BD_VALUE  "value"
#define CCA_sz_JD_BD_ST     "st"
#define CCA_sz_JD_BD_ST_SUCCESS     "succ"
#define CCA_sz_JD_BD_ST_FAIL        "fail"
#define CCA_sz_JD_BD_FAULT  "fault"
#define CCA_sz_JD_BD_FAULT_COMM     "communication failure"
#define CCA_sz_JD_BD_FAULT_EMS_COMM "ems communication failure"
#define CCA_sz_JD_BD_FAULT_ID       "unsupported 'id'"
#define CCA_sz_JD_BD_FAULT_NO_TYPE  "'type' expected"
#define CCA_sz_JD_BD_FAULT_NO_UNIT  "'unit' expected"
#define CCA_sz_JD_BD_FAULT_NO_VAL   "'value' expected"
#define CCA_sz_JD_BD_FAULT_PARSE    "Could not parse request body"
#define CCA_sz_JD_BD_FAULT_TYPE     "unsupported 'type'"
#define CCA_sz_JD_BD_FAULT_UNIT     "unsupported 'unit'"
#define CCA_sz_JD_BD_FAULT_TIME     "minimum request interval not met"

#define CCA_RQH_REQ_ID      jd_Request[CCA_sz_JD_ID]                //Request ID
#define CCA_RQH_REQ_BODY    jd_Request[CCA_sz_JD_BODY]              //Request Body
#define CCA_RQH_REQ_TYPE    CCA_RQH_REQ_BODY[CCA_sz_JD_BD_TYPE]     //Unit Type
#define CCA_RQH_REQ_UNIT    CCA_RQH_REQ_BODY[CCA_sz_JD_BD_UNIT]     //Unit Index
#define CCA_RQH_REQ_VALUE   CCA_RQH_REQ_BODY[CCA_sz_JD_BD_VALUE]    //Requested Value

#define CCA_RQH_RESP_BODY   jd_Response[CCA_sz_JD_BODY]             //Response Body   
#define CCA_RQH_RESP_ST     CCA_RQH_RESP_BODY[CCA_sz_JD_BD_ST]      //Return Status
#define CCA_RQH_RESP_FAULT  CCA_RQH_RESP_BODY[CCA_sz_JD_BD_FAULT]   //Fault Key, only used if the request fails (st = fail)

#define CCA_RQH_REQ_LIMIT_NONE      0
#define CCA_RQH_REQ_LIMIT_UPD_DATA  10

#define CCA_RQH_COMM_FAILURE_UNIT_LEN  6 //On component comm failure, append that unit to string with ", <unit>", 6 bytes

typedef struct 
{
    const char *sz_ID;                  //Request ID
    cca_HandlerFunction_t fn_Handler;   //Function to handle the request
    time_t t_RequestLimit_s;            //Minimum time between requests
    time_t t_LastRequestTime_s;         //Time of the last request (measured in seconds since the epoch)
} cca_rqh_Handler_t;

/**
 * @brief   Hash Table Look up function that will match the Request ID to the 
 *          corresponding function definiton. 
 * @param   sz_ID   Request ID to look up
 * 
 * @returns Pointer to the function handler, or NULL if not found
*/
FileScop cca_rqh_Handler_t *cca_rqh_HandlerSearch(const char *sz_ID);

/**
 * @brief   Validates the type of the request body
 * 
 * @details This function will check if the request body contains the 'type' key. If it does not,
 *          it will return false and set the response body to indicate that the 'type' key was expected.
 * 
 * @param   jd_Request  JSON Document containing the request
 * @param   jd_Response JSON Document to store the response
 * @param   jv_Type     JSON Variant to store the type
 * 
 * @returns True if the type is valid, false otherwise
*/
FileScop inline bool cca_rqh_ValidateType(ArduinoJson::JsonDocument &jd_Request, ArduinoJson::JsonDocument &jd_Response,
                                            ArduinoJson::JsonVariant &jv_Type);

/**
 * @brief   Reports the status of the request
 * 
 * @param   jd_Response JSON Document to store the response
 * @param   bol_st      Status of the request
 */
FileScop inline void cca_rqh_ReportStatus(ArduinoJson::JsonDocument &jd_Response, bool bol_st);

/**
 * @brief   Request Handler Functions
 * 
 * @param   jd_Request  JSON Document containing the request
 * @param   jd_Response JSON Document to store the response
 */
FileScop void cca_rqh_Restart(ArduinoJson::JsonDocument &jd_Request, ArduinoJson::JsonDocument &jd_Response);
FileScop void cca_rqh_ClearFaults(ArduinoJson::JsonDocument &jd_Request, ArduinoJson::JsonDocument &jd_Response);
FileScop void cca_rqh_Start(ArduinoJson::JsonDocument &jd_Request, ArduinoJson::JsonDocument &jd_Response);
FileScop void cca_rqh_Stop(ArduinoJson::JsonDocument &jd_Request, ArduinoJson::JsonDocument &jd_Response);
FileScop void cca_rqh_ReinitCom(ArduinoJson::JsonDocument &jd_Request, ArduinoJson::JsonDocument &jd_Response);
FileScop void cca_rqh_Close(ArduinoJson::JsonDocument &jd_Request, ArduinoJson::JsonDocument &jd_Response);
FileScop void cca_rqh_Open(ArduinoJson::JsonDocument &jd_Request, ArduinoJson::JsonDocument &jd_Response);
FileScop void cca_rqh_TargetActW(ArduinoJson::JsonDocument &jd_Request, ArduinoJson::JsonDocument &jd_Response);
FileScop void cca_rqh_TargetReactW(ArduinoJson::JsonDocument &jd_Request, ArduinoJson::JsonDocument &jd_Response);
FileScop void cca_rqh_OffGridMd(ArduinoJson::JsonDocument &jd_Request, ArduinoJson::JsonDocument &jd_Response);
FileScop void cca_rqh_TestAlerts(ArduinoJson::JsonDocument &jd_Request, ArduinoJson::JsonDocument &jd_Response);
FileScop void cca_rqh_UpdData(ArduinoJson::JsonDocument &jd_Request, ArduinoJson::JsonDocument &jd_Response);
FileScop void cca_rqh_RefreshVersions(ArduinoJson::JsonDocument &jd_Request, ArduinoJson::JsonDocument &jd_Response);

/**
 * @brief   Handles the generic commands that are sent to the DCDC, PCS, and PDU units.
 * 
 * @param   jd_Request  JSON Document containing the request
 * @param   jd_Response JSON Document to store the response
 * 
 * @returns True if the request was successful, false otherwise
 */
FileScop bool cca_rqh_genericCmd(ArduinoJson::JsonDocument &jd_Request, ArduinoJson::JsonDocument &jd_Response);

/**
 * @brief   Checks if the minimum time between requests has elapsed.
 * 
 * @details If the minimum time between requests has been met, the function 
 *          will set the last request time to the current time.
 * 
 * @param   ptr_Handler Pointer to the handler that contains the request time
 * @returns True if the minimum time has elapsed, false otherwise
 */
FileScop bool cca_rqh_IsMinTimeElapsed(cca_rqh_Handler_t *ptr_Handler);

/**
 * @brief   This function will set the value of the PCS unit based on the request.
 * 
 * @param   jd_Request  JSON Document containing the request
 * @param   jd_Response JSON Document to store the response
 * 
 * @returns True if the request was successful, false otherwise
 */
FileScop bool cca_rqh_setPcsValue(ArduinoJson::JsonDocument &jd_Request, ArduinoJson::JsonDocument &jd_Response);

FileScop cca_rqh_Handler_t arr_Handlers[] = {
    {"restart", cca_rqh_Restart, CCA_RQH_REQ_LIMIT_NONE, 0},
    {"clear_faults", cca_rqh_ClearFaults, CCA_RQH_REQ_LIMIT_NONE, 0},
    {"start", cca_rqh_Start, CCA_RQH_REQ_LIMIT_NONE, 0},
    {"stop", cca_rqh_Stop, CCA_RQH_REQ_LIMIT_NONE, 0},
    {"reinit_com", cca_rqh_ReinitCom, CCA_RQH_REQ_LIMIT_NONE, 0},
    {"close", cca_rqh_Close, CCA_RQH_REQ_LIMIT_NONE, 0},
    {"open", cca_rqh_Open, CCA_RQH_REQ_LIMIT_NONE, 0},
    {"target_act_w", cca_rqh_TargetActW, CCA_RQH_REQ_LIMIT_NONE, 0},
    {"target_react_w", cca_rqh_TargetReactW, CCA_RQH_REQ_LIMIT_NONE, 0},
    {"off_grid_md", cca_rqh_OffGridMd, CCA_RQH_REQ_LIMIT_NONE, 0},
    {"test_alerts", cca_rqh_TestAlerts, CCA_RQH_REQ_LIMIT_NONE, 0},
    {"upd_data", cca_rqh_UpdData, CCA_RQH_REQ_LIMIT_UPD_DATA, 0},
    {"refresh_versions", cca_rqh_RefreshVersions, CCA_RQH_REQ_LIMIT_NONE, 0},
};
FileScop const uint8_t siz_arr_Handlers = sizeof(arr_Handlers) / sizeof(arr_Handlers[0]);

/*********************************************************************************************************************
* @cond                                              MODULE SCOPE                                             @endcond
*********************************************************************************************************************/

ModScop int CCA_Rqh_ValidateRequest(const char *sz_RequestID, cca_HandlerFunction_t *ptr_HandlerFunction)
{
    cca_rqh_Handler_t *ptr_Handler = cca_rqh_HandlerSearch(sz_RequestID);
    if (ptr_Handler != NULL)
    {
        *ptr_HandlerFunction = ptr_Handler->fn_Handler;
    }
    else {
        return -ENOSYS;
    }
    return EXIT_SUCCESS;
}

/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

FileScop cca_rqh_Handler_t *cca_rqh_HandlerSearch(const char *sz_ID)
{
    for (uint8_t i = 0; i < siz_arr_Handlers; i++)
    {
        if (strcmp(arr_Handlers[i].sz_ID, sz_ID) == EXIT_SUCCESS)
        {
            return &arr_Handlers[i];
        }
    }
    return NULL;
}

//------Inline Helper Functions------//

FileScop inline bool cca_rqh_ValidateType(ArduinoJson::JsonDocument &jd_Request, ArduinoJson::JsonDocument &jd_Response,
                                            ArduinoJson::JsonVariant &jv_Type)
{
    if (CCA_RQH_REQ_BODY.containsKey(CCA_sz_JD_BD_TYPE) == false)
    {
        CCA_RQH_RESP_FAULT = CCA_sz_JD_BD_FAULT_NO_TYPE;
        return false;
    }
    else
    {
        jv_Type = CCA_RQH_REQ_TYPE;
        return true;
    }
}

FileScop inline void cca_rqh_ReportStatus(ArduinoJson::JsonDocument &jd_Response, bool bol_st)
{
    if (bol_st)
    {
        CCA_RQH_RESP_ST = CCA_sz_JD_BD_ST_SUCCESS;
    }
    else
    {
        CCA_RQH_RESP_ST = CCA_sz_JD_BD_ST_FAIL;
    }
}

//------Request Handler Functions------//

FileScop void cca_rqh_Restart(ArduinoJson::JsonDocument &jd_Request, ArduinoJson::JsonDocument &jd_Response)
{
    ArduinoJson::JsonVariant jv_Type;
    bool bol_st = cca_rqh_ValidateType(jd_Request, jd_Response, jv_Type);
    if (bol_st)
    {
        if (jv_Type == CCA_sz_UNIT_TYPE_DCDC || jv_Type == CCA_sz_UNIT_TYPE_PCS || jv_Type == CCA_sz_UNIT_TYPE_PDU)
        {
           bol_st = cca_rqh_genericCmd(jd_Request, jd_Response);
        }
        else if (jv_Type == CCA_sz_UNIT_TYPE_DCU || jv_Type == CCA_sz_UNIT_TYPE_PMU)
        {
            if (CCA_Icc_RestartEMS(jv_Type) != EXIT_SUCCESS)
            {
                bol_st = false;
                CCA_RQH_RESP_FAULT = CCA_sz_JD_BD_FAULT_EMS_COMM;
            }
        }
        else
        {
            bol_st = false;
            CCA_RQH_RESP_FAULT = CCA_sz_JD_BD_FAULT_TYPE;
        }
    }
    cca_rqh_ReportStatus(jd_Response, bol_st);
}

FileScop void cca_rqh_ClearFaults(ArduinoJson::JsonDocument &jd_Request, ArduinoJson::JsonDocument &jd_Response)
{
    ArduinoJson::JsonVariant jv_Type;
    bool bol_st = cca_rqh_ValidateType(jd_Request, jd_Response, jv_Type);
    if (bol_st)
    {
        if (jv_Type == CCA_sz_UNIT_TYPE_DCDC || jv_Type == CCA_sz_UNIT_TYPE_PCS || jv_Type == CCA_sz_UNIT_TYPE_PDU)
        {
           bol_st = cca_rqh_genericCmd(jd_Request, jd_Response);
        }
        else
        {
            bol_st = false;
            CCA_RQH_RESP_FAULT = CCA_sz_JD_BD_FAULT_TYPE;
        }
    }
    cca_rqh_ReportStatus(jd_Response, bol_st);
}

FileScop void cca_rqh_Start(ArduinoJson::JsonDocument &jd_Request, ArduinoJson::JsonDocument &jd_Response)
{
    ArduinoJson::JsonVariant jv_Type;
    bool bol_st = cca_rqh_ValidateType(jd_Request, jd_Response, jv_Type);
    if (bol_st)
    {
        if (jv_Type == CCA_sz_UNIT_TYPE_DCDC || jv_Type == CCA_sz_UNIT_TYPE_PCS || jv_Type == CCA_sz_UNIT_TYPE_PDU)
        {
           bol_st = cca_rqh_genericCmd(jd_Request, jd_Response);
        }
        else
        {
            bol_st = false;
            CCA_RQH_RESP_FAULT = CCA_sz_JD_BD_FAULT_TYPE;
        }
    }
    cca_rqh_ReportStatus(jd_Response, bol_st);
}

FileScop void cca_rqh_Stop(ArduinoJson::JsonDocument &jd_Request, ArduinoJson::JsonDocument &jd_Response)
{
    ArduinoJson::JsonVariant jv_Type;
    bool bol_st = cca_rqh_ValidateType(jd_Request, jd_Response, jv_Type);
    if (bol_st)
    {
        if (jv_Type == CCA_sz_UNIT_TYPE_DCDC || jv_Type == CCA_sz_UNIT_TYPE_PCS || jv_Type == CCA_sz_UNIT_TYPE_PDU)
        {
           bol_st = cca_rqh_genericCmd(jd_Request, jd_Response);
        }
        else
        {
            bol_st = false;
            CCA_RQH_RESP_FAULT = CCA_sz_JD_BD_FAULT_TYPE;
        }
    }
    cca_rqh_ReportStatus(jd_Response, bol_st);
}

FileScop void cca_rqh_ReinitCom(ArduinoJson::JsonDocument &jd_Request, ArduinoJson::JsonDocument &jd_Response)
{
    ArduinoJson::JsonVariant jv_Type;
    bool bol_st = cca_rqh_ValidateType(jd_Request, jd_Response, jv_Type);
    if (bol_st)
    {
        if (jv_Type == CCA_sz_UNIT_TYPE_DCDC || jv_Type == CCA_sz_UNIT_TYPE_PCS || jv_Type == CCA_sz_UNIT_TYPE_PDU)
        {
           bol_st = cca_rqh_genericCmd(jd_Request, jd_Response);
        }
        else if (jv_Type == CCA_sz_UNIT_TYPE_DCU)
        {
            if (CCA_Icc_ReinitComDCU() != EXIT_SUCCESS)
            {
                bol_st = false;
                CCA_RQH_RESP_FAULT = CCA_sz_JD_BD_FAULT_EMS_COMM;
            }
        }
        else
        {
            bol_st = false;
            CCA_RQH_RESP_FAULT = CCA_sz_JD_BD_FAULT_TYPE;
        }
    }
    cca_rqh_ReportStatus(jd_Response, bol_st);
}

FileScop void cca_rqh_Close(ArduinoJson::JsonDocument &jd_Request, ArduinoJson::JsonDocument &jd_Response)
{
    ArduinoJson::JsonVariant jv_Type;
    bool bol_st = cca_rqh_ValidateType(jd_Request, jd_Response, jv_Type);
    if (bol_st)
    {
        if (jv_Type == CCA_sz_UNIT_TYPE_PDU)
        {
           bol_st = cca_rqh_genericCmd(jd_Request, jd_Response);
        }
        else
        {
            bol_st = false;
            CCA_RQH_RESP_FAULT = CCA_sz_JD_BD_FAULT_TYPE;
        }
    }
    cca_rqh_ReportStatus(jd_Response, bol_st);
}

FileScop void cca_rqh_Open(ArduinoJson::JsonDocument &jd_Request, ArduinoJson::JsonDocument &jd_Response)
{
    ArduinoJson::JsonVariant jv_Type;
    bool bol_st = cca_rqh_ValidateType(jd_Request, jd_Response, jv_Type);
    if (bol_st)
    {
        if (jv_Type == CCA_sz_UNIT_TYPE_PDU)
        {
           bol_st = cca_rqh_genericCmd(jd_Request, jd_Response);
        }
        else
        {
            bol_st = false;
            CCA_RQH_RESP_FAULT = CCA_sz_JD_BD_FAULT_TYPE;
        }
    }
    cca_rqh_ReportStatus(jd_Response, bol_st);
}

FileScop void cca_rqh_TargetActW(ArduinoJson::JsonDocument &jd_Request, ArduinoJson::JsonDocument &jd_Response)
{
    ArduinoJson::JsonVariant jv_Type;
    bool bol_st = cca_rqh_ValidateType(jd_Request, jd_Response, jv_Type);
    if (bol_st)
    {
        if (jv_Type == CCA_sz_UNIT_TYPE_PCS)
        {
           bol_st = cca_rqh_setPcsValue(jd_Request, jd_Response);
        }
        else
        {
            bol_st = false;
            CCA_RQH_RESP_FAULT = CCA_sz_JD_BD_FAULT_TYPE;
        }
    }
    cca_rqh_ReportStatus(jd_Response, bol_st);
}

FileScop void cca_rqh_TargetReactW(ArduinoJson::JsonDocument &jd_Request, ArduinoJson::JsonDocument &jd_Response)
{
    ArduinoJson::JsonVariant jv_Type;
    bool bol_st = cca_rqh_ValidateType(jd_Request, jd_Response, jv_Type);
    if (bol_st)
    {
        if (jv_Type == CCA_sz_UNIT_TYPE_PCS)
        {
           bol_st = cca_rqh_setPcsValue(jd_Request, jd_Response);
        }
        else
        {
            bol_st = false;
            CCA_RQH_RESP_FAULT = CCA_sz_JD_BD_FAULT_TYPE;
        }
    }
    cca_rqh_ReportStatus(jd_Response, bol_st);
}

FileScop void cca_rqh_OffGridMd(ArduinoJson::JsonDocument &jd_Request, ArduinoJson::JsonDocument &jd_Response)
{
    ArduinoJson::JsonVariant jv_Type;
    bool bol_st = cca_rqh_ValidateType(jd_Request, jd_Response, jv_Type);
    if (bol_st)
    {
        if (jv_Type == CCA_sz_UNIT_TYPE_PCS)
        {
           bol_st = cca_rqh_genericCmd(jd_Request, jd_Response);
        }
        else
        {
            bol_st = false;
            CCA_RQH_RESP_FAULT = CCA_sz_JD_BD_FAULT_TYPE;
        }
    }
    cca_rqh_ReportStatus(jd_Response, bol_st);
}

FileScop void cca_rqh_TestAlerts(ArduinoJson::JsonDocument &jd_Request, ArduinoJson::JsonDocument &jd_Response)
{
    uint8_t u8_UnitIndex = 0;
    ArduinoJson::JsonVariant jv_Type;
    bool bol_st = cca_rqh_ValidateType(jd_Request, jd_Response, jv_Type);
    if (bol_st)
    {
        if (CCA_RQH_REQ_BODY.containsKey(CCA_sz_JD_BD_UNIT))
        {
            u8_UnitIndex = CCA_RQH_REQ_UNIT.as<const uint8_t>();
        }
        else 
        {
            bol_st = false;
            CCA_RQH_RESP_FAULT = CCA_sz_JD_BD_FAULT_NO_UNIT;
        }
        if (bol_st)
        {
            int int_rc = CCA_Icc_TestAlertStatusCallback(jv_Type.as<const char*>(), u8_UnitIndex);
            if (int_rc != EXIT_SUCCESS)
            {
                bol_st = false;
                CCA_RQH_RESP_FAULT = strerror(-int_rc);
            }
        }
    }
    cca_rqh_ReportStatus(jd_Response, bol_st);
}

FileScop void cca_rqh_UpdData(ArduinoJson::JsonDocument &jd_Request, ArduinoJson::JsonDocument &jd_Response)
{
    bool bol_st = cca_rqh_IsMinTimeElapsed(cca_rqh_HandlerSearch(CCA_RQH_REQ_ID.as<const char*>()));
    if (!bol_st)
    {
        CCA_RQH_RESP_FAULT = CCA_sz_JD_BD_FAULT_TIME;
        cca_rqh_ReportStatus(jd_Response, bol_st);
        return;
    }

    ArduinoJson::JsonVariant jv_Type;
    bool bol_IsReqForProductData = !CCA_RQH_REQ_BODY.containsKey(CCA_sz_JD_BD_TYPE);
    if (bol_IsReqForProductData)
    {
        int int_rc = CCA_EMS_UpdateProductData();
        if (int_rc != EXIT_SUCCESS)
        {
            bol_st = false;
            CCA_RQH_RESP_FAULT = strerror(-int_rc);
        }
    }
    else
    {
        bol_st = cca_rqh_ValidateType(jd_Request, jd_Response, jv_Type);
        if (bol_st)
        {
            unit_type_t enm_UnitType;
            if (CCA_Int_UnitTypeStringToEnum(jv_Type.as<const char*>(), &enm_UnitType) == EXIT_SUCCESS)
            {
                //Parse unit index. If none present, update all of that type
                uint8_t u8_UnitIndex = COD_u8_NULL_UNIT_ID;
                if (CCA_RQH_REQ_BODY.containsKey(CCA_sz_JD_BD_UNIT))
                {
                    u8_UnitIndex = CCA_RQH_REQ_UNIT.as<const uint8_t>();
                }

                int (*arr_fp_UpdateComponentData[])(uint8_t) = {
                    CCA_Pcs_UpdateComponentData, CCA_Pdu_UpdateComponentData, CCA_Pmu_UpdateComponentData,
                    CCA_Dcdc_UpdateComponentData, NULL, CCA_Hvac_UpdateComponentData, NULL
                };
                
                if (arr_fp_UpdateComponentData[enm_UnitType] == NULL)
                {
                    bol_st = false;
                    CCA_RQH_RESP_FAULT = CCA_sz_JD_BD_FAULT_TYPE;
                }
                else
                {
                    int int_rc = arr_fp_UpdateComponentData[enm_UnitType](u8_UnitIndex);
                    if (int_rc != EXIT_SUCCESS)
                    {
                        bol_st = false;
                        CCA_RQH_RESP_FAULT = strerror(-int_rc);
                    }
                }
            }
            else
            {
                bol_st = false;
                CCA_RQH_RESP_FAULT = CCA_sz_JD_BD_FAULT_TYPE;
            }
        }
    }
    cca_rqh_ReportStatus(jd_Response, bol_st);
}

FileScop void cca_rqh_RefreshVersions(ArduinoJson::JsonDocument &jd_Request, ArduinoJson::JsonDocument &jd_Response)
{
    bool bol_st = true;
    int int_rc = CCA_Icc_RequestVersionInfo();
    if (int_rc != EXIT_SUCCESS)
    {
        bol_st = false;
        CCA_RQH_RESP_FAULT = strerror(-int_rc);
    }
    cca_rqh_ReportStatus(jd_Response, bol_st);
}

FileScop bool cca_rqh_genericCmd(ArduinoJson::JsonDocument &jd_Request, ArduinoJson::JsonDocument &jd_Response)
{
    uint8_t u8_UnitQty = 0;
    uint8_t u8_UnitIndex = 0;
    ArduinoJson::JsonVariant jv_UnitType = CCA_RQH_REQ_TYPE;
    bool bol_st = true; //Assume good until we run into an issue

    if (CCA_RQH_REQ_BODY.containsKey(CCA_sz_JD_BD_UNIT))
    {
        //If "unit" is specified, send to only that unit
        u8_UnitQty = 1;
        u8_UnitIndex = CCA_RQH_REQ_UNIT.as<const uint8_t>();
        //Check if the unit index is valid
        unit_type_t enm_UnitType;
        (void)CCA_Int_UnitTypeStringToEnum(jv_UnitType.as<const char*>(), &enm_UnitType);
        if (CCA_Int_ValidateUnitIndex(enm_UnitType, u8_UnitIndex) != EXIT_SUCCESS)
        {
            bol_st = false;
            CCA_RQH_RESP_FAULT = CCA_sz_JD_BD_FAULT_UNIT;
        }
    }
    u8_UnitQty = CCA_Int_GetUnitQuantity_sz(CCA_RQH_REQ_TYPE);
    if (bol_st && u8_UnitQty == 0)
    {
        //If we have not already failed and the unit quantity is zero, fail the request.
        //We should not be able to get here under normal circumstances. If we do, there is a bug somewhere.
        LOG_ERR("Could not get unit quantity for type, %s", jv_UnitType.as<const char*>());
        bol_st = false;
        CCA_RQH_RESP_FAULT = CCA_sz_JD_BD_FAULT_PARSE;
    }
    if (bol_st) //If parsing succeeded, send the request to the unit(s)
    {
        // Build the Request with the assumption that the unitIndex is either zero or specified prior to this loop.
        const char *sz_requestKey[1] = { CCA_RQH_REQ_ID.as<const char*>() };
        int32_t s32_requestValue[1] = { 1 };
        int int_requestStatus[u8_UnitQty];
        bool bol_emsCommFailure = false;
        for (uint8_t i = 0; i < u8_UnitQty; i++)
        {
            int_requestStatus[i] = EXIT_FAILURE;
            if (!CCA_Icc_SetQuantifiableData(jv_UnitType.as<const char*>(), u8_UnitIndex, 1, sz_requestKey, 
                                        s32_requestValue, &int_requestStatus[i]) == EXIT_SUCCESS)
            {
                bol_emsCommFailure = true;
                break;
            }
            u8_UnitIndex++;
        }
        if (bol_emsCommFailure)
        {
            bol_st = false;
            CCA_RQH_RESP_FAULT = CCA_sz_JD_BD_FAULT_EMS_COMM;
        }
        else
        {
            // Check if all units were successfully handled
            char sz_error[64] = "comm failure on unit(s)";
            if (u8_UnitQty == 1 && int_requestStatus[0] != EXIT_SUCCESS)
            {
                bol_st = false;
                char sz_unitIndex[CCA_RQH_COMM_FAILURE_UNIT_LEN];
                sprintf(sz_unitIndex, ", %d", CCA_RQH_REQ_UNIT.as<const uint8_t>());
                strcat(sz_error, sz_unitIndex);
            }
            else
            {
                for (uint8_t i = 0; i < u8_UnitQty; i++)
                {
                    if (int_requestStatus[i] != EXIT_SUCCESS) 
                    {
                        bol_st = false;
                        char sz_unitIndex[CCA_RQH_COMM_FAILURE_UNIT_LEN];
                        sprintf(sz_unitIndex, ", %d", i);
                        strcat(sz_error, sz_unitIndex);
                    }
                }
            }
            if (bol_st == false)
            {
                CCA_RQH_RESP_FAULT = sz_error;
            }
        }
    }
    return bol_st;
}

FileScop bool cca_rqh_IsMinTimeElapsed(cca_rqh_Handler_t *ptr_Handler)
{
    time_t t_CurrentTime_s = time(NULL);
    if (t_CurrentTime_s - ptr_Handler->t_LastRequestTime_s < ptr_Handler->t_RequestLimit_s)
    {
        return false;
    }
    ptr_Handler->t_LastRequestTime_s = t_CurrentTime_s;
    return true;
}

FileScop bool cca_rqh_setPcsValue(ArduinoJson::JsonDocument &jd_Request, ArduinoJson::JsonDocument &jd_Response)
{
    ArduinoJson::JsonVariant jv_Type;
    bool bol_st = true; //Assume good until we run into an issue

    if (CCA_RQH_REQ_BODY.containsKey(CCA_sz_JD_BD_VALUE))
    {
        const char *sz_requestedKey[1] = { CCA_RQH_REQ_ID.as<const char*>() };
        ArduinoJson::JsonVariant jv_requestedValue = CCA_RQH_REQ_VALUE;
        int32_t s32_requestedValue[1] = { CCA_RQH_REQ_VALUE.as<const int32_t>() };
        int int_requestStatus[1] = { EXIT_FAILURE };
        if (CCA_Icc_SetQuantifiableData(jv_Type.as<const char*>(), 0, 1, sz_requestedKey, 
                                        s32_requestedValue, int_requestStatus) == EXIT_SUCCESS)
        {
            if (int_requestStatus[0] != EXIT_SUCCESS)
            {
                bol_st = false;
                CCA_RQH_RESP_FAULT = CCA_sz_JD_BD_FAULT_COMM;
            }
        }
        else
        {
            bol_st = false;
            CCA_RQH_RESP_FAULT = CCA_sz_JD_BD_FAULT_EMS_COMM;
        }
    }
    else
    {
        bol_st = false;
        CCA_RQH_RESP_FAULT = CCA_sz_JD_BD_FAULT_NO_VAL;
    }
    return bol_st;
}

//EOF
