#include "include/CCA_Version.h"
#include "lion/log/log.h"
#include "app_version.h"

LOG_MODULE_REGISTER(CCA_version, LOG_LEVEL_INF);

/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

#define STRINGIFY(X) __STRING(X)

#define CCA_sz_VERSION_INFO_JSON_DOC_SIZE CCA_u16_BIG_SETTINGS_JSON_DOC_SIZE
#define CCA_sz_VERSION_INFO_ID_KEY "id"

#define CCA_sz_VERSION_INFO_TYPE_CCU "ccu"

#define CCA_VERSION_INFO_JSON_DOC  cca_jd_VersionInfo
#define CCA_VERSION_INFO_COMPONENTS CCA_VERSION_INFO_JSON_DOC["state"]["reported"]["components"]

/**
 * @brief A JSON document to store the version information of the system. This includes the CCU, DCU, and PMU data.
 * 
 * @details Format:
 *  {
      "state": {
        "reported": {
          "product_id": "<id>",
          "type": "cabinet",
          "components": [
            {
            "type": "ccu",
            "id": "<id>",
            "hw": "<version>",
            "img": "<version>",
            "<app_name>": "<version>-<commit hash>",
            "<app_name>": "<version>-<commit hash>",
            "mac": "<mac_address>",
            "sn": "<serial>"
            },
            {
            "type": "dcu",
            "id": "<id>",
            "hw": "<version>",
            "bl": "<version>",
            "fw": "<version>-<commit hash>",
            "mac": "<mac_address>",
            "sn": "<serial>"
            },
            {
            "type": "pmu",
            "id": "<id>",
            "hw": "<version>",
            "bl": "<version>",
            "fw": "<version>-<commit hash>",
            "mac": "<mac_address>",
            "sn": "<serial>",
            "index": "<index>"
            }
          ]
        }
      }
    }
 */
FileScop ArduinoJson::StaticJsonDocument<CCA_sz_VERSION_INFO_JSON_DOC_SIZE> CCA_VERSION_INFO_JSON_DOC;
FileScop bool bol_VersionInfoInitialized = false;

FileScop bool bol_VersionInfoChanged = false; //Set high when the version info is updated and needs to be sent to AWS

/**
 * @brief Initializes the version info file buffer with CCU info. Other devices will be added later through ICC calls.
 */
FileScop void cca_InitVersionFile(void);

/*********************************************************************************************************************
* @cond                                              MODULE SCOPE                                             @endcond
*********************************************************************************************************************/

ModScop void CCA_Version_Init(void)
{
    if (!bol_VersionInfoInitialized)
    {
        cca_InitVersionFile();
    }
}

ModScop int CCA_Version_InitShadows(void)
{
    int int_rc = EXIT_SUCCESS;

    //Subscribe to Version Shadow sub-topics
    int_rc = CCA_AWS_SubscribeToShadowSubtopics(CCA_sz_VERSION_SHADOW_NAME, cca_sz_version_update_delta_topic, 
                                                cca_sz_version_update_accepted_topic, cca_sz_version_update_rejected_topic);
    //Publish Version Info
    if (int_rc == EXIT_SUCCESS)
    {
        int_rc = CCA_Version_PublishVersionUpdate();
    }
    return int_rc;
}

ModScop int CCA_Version_UpdateVersionInfoWith(ArduinoJson::JsonObject& jo_Obj)
{
    if (!bol_VersionInfoInitialized)
    {
        return -EAGAIN;
    }

    //Get the device ID from the object and check if an object in CCA_VERSION_INFO_JSON_DOC already exists
    if (jo_Obj.containsKey(CCA_sz_VERSION_INFO_ID_KEY))
    {
        const char *sz_ID = jo_Obj[CCA_sz_VERSION_INFO_ID_KEY];
        //Iterate through the array of objects in CCA_VERSION_INFO_JSON_DOC and check if the ID matches
        ArduinoJson::JsonArray ja_Components = CCA_VERSION_INFO_COMPONENTS;
        for (ArduinoJson::JsonObject jo_Component : ja_Components)
        {
            if (jo_Component["id"] == sz_ID && jo_Component["type"] != CCA_sz_VERSION_INFO_TYPE_CCU)
            {
                //Update the object with the new data
                for (ArduinoJson::JsonPair kv : jo_Component)
                {
                    if (jo_Obj.containsKey(kv.key()))
                    {
                        jo_Component[kv.key()] = jo_Obj[kv.key()];
                    }
                }
                LOG_INF("Updated Version Info (%s) for %s", jo_Obj["type"].as<const char*>(), sz_ID);
                bol_VersionInfoChanged = true;
                return EXIT_SUCCESS;
            }
        }
        //If we get here, the object does not exist. Add it
        ja_Components.add(jo_Obj);
        LOG_INF("Added Version Info (%s) for %s", jo_Obj["type"].as<const char*>(), sz_ID);
        bol_VersionInfoChanged = true;
        return EXIT_SUCCESS;
    }
    else
    {
        LOG_ERR("No ID key found in the object, skipping update");
        return -EINVAL;
    }
}

ModScop int CCA_Version_PublishVersionUpdate(void)
{
    int int_rc = EXIT_SUCCESS;
    if (bol_VersionInfoChanged)
    {
        char sz_VersionInfo[CCA_sz_VERSION_INFO_JSON_DOC_SIZE];
        ArduinoJson::serializeJson(CCA_VERSION_INFO_JSON_DOC, sz_VersionInfo);
        LOG_DBG("Publish version info to AWS: %s", sz_VersionInfo);
        int_rc = CCA_AWS_PublishToTopic(&str_MqttContext, cca_sz_version_update_topic, strlen(cca_sz_version_update_topic),
                                        sz_VersionInfo, strlen(sz_VersionInfo));
        if (int_rc == EXIT_SUCCESS)
        {
            bol_VersionInfoChanged = false;
        }
    }
    return int_rc;
}

/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

FileScop void cca_InitVersionFile(void)
{
    CCA_VERSION_INFO_JSON_DOC.clear();
    ArduinoJson::JsonObject jo_Reported = CCA_VERSION_INFO_JSON_DOC["state"].createNestedObject("reported");
    jo_Reported["product_id"] = cca_sz_thing_name;
    jo_Reported["type"] = "cabinet";
    ArduinoJson::JsonArray ja_Components = jo_Reported.createNestedArray("components");
    ArduinoJson::JsonObject jo_CCU = ja_Components.createNestedObject();
    jo_CCU["type"] = CCA_sz_VERSION_INFO_TYPE_CCU;
    jo_CCU["id"] = cca_sz_thing_name;
    jo_CCU["hw"] = ""; //TODO Get the hardware version
    jo_CCU["img"] = ""; //TODO Get the image version
    char sz_AppVnTemp[32];
#if APP_BUILD_DIRTY //This is defined in the CCA CMakeLists.txt
    sprintf(sz_AppVnTemp, "v%s-%.7s-dirty", APP_VERSION_STRING, STRINGIFY(APP_BUILD_VERSION));
#else
    sprintf(sz_AppVnTemp, "v%s-%.7s", APP_VERSION_STRING, STRINGIFY(APP_BUILD_VERSION));
#endif
    jo_CCU["lion-ems-ccu-cca"] = sz_AppVnTemp;
    //TODO how to handle other applications?
    jo_CCU["mac"] = ""; //TODO Get the MAC address
    jo_CCU["sn"] = ""; //TODO Get the serial number

    char sz_VersionInfo[CCA_sz_VERSION_INFO_JSON_DOC_SIZE];
    ArduinoJson::serializeJson(CCA_VERSION_INFO_JSON_DOC, sz_VersionInfo);
    LOG_INF("Initialize Version Info: %s", sz_VersionInfo);
    bol_VersionInfoInitialized = true;
    bol_VersionInfoChanged = true;
}

//EOF
