/**********************************************************************************************************************
* @file            CCA_AWS.h
*
* @brief           This file declares all functions and definitions specific to MQTT AWS cloud communications.
*
* @author          Devan Hone
* @date            06/12/2024
* @copyright       2024 Lion Energy LLC.
*                  All rights reserved
*
*********************************************************************************************************************/
/**
* @addtogroup      CCA
*                  Cloud Communication Application Module 
* 
* @ref             CCA_AWS.h 
*                  
*********************************************************************************************************************/
#include "COD_Compiler_Defines.h"

#ifndef LION_CCA_AWS_H__
#define LION_CCA_AWS_H__

#include "aws_iot/clock.h"
#include "aws_iot/core_json.h"
#include "aws_iot/core_mqtt_config.h"
#include "aws_iot/core_mqtt.h"
#include "aws_iot/core_mqtt_state.h"
#include "aws_iot/backoff_algorithm.h"
#include "aws_iot/openssl_posix.h"
#include "aws_iot/transport_interface.h"
#include "mqtt_subscription_manager.h"

/*********************************************************************************************************************
* @cond                                              MODULE SCOPE                                             @endcond
*********************************************************************************************************************/

#define     CCA_sz_APPLICATION_DATA_DIR "/data/var/lib/app-core/"
#define     REGISTRY_STATUS_FILE        CCA_sz_APPLICATION_DATA_DIR "registry_status"
#define     CLAIM_CERT_PATH             CCA_sz_APPLICATION_DATA_DIR "certs/claim-cert.pem.crt"
#define     CLAIM_PRIVATE_KEY_PATH      CCA_sz_APPLICATION_DATA_DIR "certs/claim-pvt.pem.key"
#define     CCA_sz_DEVICE_CERT_PATH     CCA_sz_APPLICATION_DATA_DIR "certs/deviceCert.pem.crt"
#define     CCA_sz_DEVICE_KEY_PATH      CCA_sz_APPLICATION_DATA_DIR "certs/deviceKey.pem.key"
#define     CCA_sz_OWNERSHIP_TOKEN      CCA_sz_APPLICATION_DATA_DIR "certs/certificateOwnership.pem.key"
#define     CCA_sz_CONFIG_FILE_PATH     CCA_sz_APPLICATION_DATA_DIR "awsConfig.json"

#define     END_OF_CERTIFICATE_STRING          "\",\"privateKey\":\""
#define     END_OF_KEY_STRING                   ",\"certificateOwnershipToken\":"

/* AWS ENVIRONMENT CONFIGURATIONS */
#define     CCA_sz_DEV_ENDPOINT                 "a30rm556obym6i-ats.iot.us-east-1.amazonaws.com"
#define     CCA_sz_STAGE_ENDPOINT               "a2wfv617yfoln8-ats.iot.us-east-1.amazonaws.com"
#define     CCA_sz_PROD_ENDPOINT                "a1qtqnx2h0gfil-ats.iot.us-east-1.amazonaws.com"


/* AWS SHADOW TOPIC DEFINITIONS */
#define     CCA_sz_PCS_SHADOW_NAME              "pcs_settings"
#define     CCA_sz_DCDC_SHADOW_NAME             "dcdc_settings"
#define     CCA_sz_EMS_SHADOW_NAME              "ems_settings"
#define     CCA_sz_PMU_SHADOW_NAME              "pmu_settings"
#define     CCA_sz_HVAC_SHADOW_NAME             "hvac_settings"
#define     CCA_sz_VERSION_SHADOW_NAME          "version"
#define     CCA_sz_SHADOW_PREFIX                "$aws/things/"
#define     CCA_sz_SHADOW_BODY                  "/shadow/name/"
#define     CCA_sz_SHADOW_TOPIC_UPDATE_END      "/update"
#define     CCA_sz_SHADOW_TOPIC_DELTA_END       "/update/delta"
#define     CCA_sz_SHADOW_TOPIC_ACCEPTED_END    "/update/accepted"
#define     CCA_sz_SHADOW_TOPIC_REJECTED_END    "/update/rejected"

#define     CCA_sz_PCS_SHADOW_PREFIX            CCA_sz_SHADOW_BODY CCA_sz_PCS_SHADOW_NAME
#define     CCA_sz_PCS_UPDATE_TOPIC             CCA_sz_PCS_SHADOW_PREFIX CCA_sz_SHADOW_TOPIC_UPDATE_END
#define     CCA_sz_PCS_UPDATE_DELTA_TOPIC       CCA_sz_PCS_SHADOW_PREFIX CCA_sz_SHADOW_TOPIC_DELTA_END
#define     CCA_sz_PCS_UPDATE_ACCEPTED_TOPIC    CCA_sz_PCS_SHADOW_PREFIX CCA_sz_SHADOW_TOPIC_ACCEPTED_END
#define     CCA_sz_PCS_UPDATE_REJECTED_TOPIC    CCA_sz_PCS_SHADOW_PREFIX CCA_sz_SHADOW_TOPIC_REJECTED_END

#define     CCA_sz_DCDC_SHADOW_PREFIX           CCA_sz_SHADOW_BODY CCA_sz_DCDC_SHADOW_NAME
#define     CCA_sz_DCDC_UPDATE_TOPIC            CCA_sz_DCDC_SHADOW_PREFIX CCA_sz_SHADOW_TOPIC_UPDATE_END
#define     CCA_sz_DCDC_UPDATE_DELTA_TOPIC      CCA_sz_DCDC_SHADOW_PREFIX CCA_sz_SHADOW_TOPIC_DELTA_END
#define     CCA_sz_DCDC_UPDATE_ACCEPTED_TOPIC   CCA_sz_DCDC_SHADOW_PREFIX CCA_sz_SHADOW_TOPIC_ACCEPTED_END
#define     CCA_sz_DCDC_UPDATE_REJECTED_TOPIC   CCA_sz_DCDC_SHADOW_PREFIX CCA_sz_SHADOW_TOPIC_REJECTED_END

#define     CCA_sz_EMS_SHADOW_PREFIX            CCA_sz_SHADOW_BODY CCA_sz_EMS_SHADOW_NAME
#define     CCA_sz_EMS_UPDATE_TOPIC             CCA_sz_EMS_SHADOW_PREFIX CCA_sz_SHADOW_TOPIC_UPDATE_END
#define     CCA_sz_EMS_UPDATE_DELTA_TOPIC       CCA_sz_EMS_SHADOW_PREFIX CCA_sz_SHADOW_TOPIC_DELTA_END
#define     CCA_sz_EMS_UPDATE_ACCEPTED_TOPIC    CCA_sz_EMS_SHADOW_PREFIX CCA_sz_SHADOW_TOPIC_ACCEPTED_END
#define     CCA_sz_EMS_UPDATE_REJECTED_TOPIC    CCA_sz_EMS_SHADOW_PREFIX CCA_sz_SHADOW_TOPIC_REJECTED_END

#define     CCA_sz_PMU_SHADOW_PREFIX            CCA_sz_SHADOW_BODY CCA_sz_PMU_SHADOW_NAME
#define     CCA_sz_PMU_UPDATE_TOPIC             CCA_sz_PMU_SHADOW_PREFIX CCA_sz_SHADOW_TOPIC_UPDATE_END
#define     CCA_sz_PMU_UPDATE_DELTA_TOPIC       CCA_sz_PMU_SHADOW_PREFIX CCA_sz_SHADOW_TOPIC_DELTA_END
#define     CCA_sz_PMU_UPDATE_ACCEPTED_TOPIC    CCA_sz_PMU_SHADOW_PREFIX CCA_sz_SHADOW_TOPIC_ACCEPTED_END
#define     CCA_sz_PMU_UPDATE_REJECTED_TOPIC    CCA_sz_PMU_SHADOW_PREFIX CCA_sz_SHADOW_TOPIC_REJECTED_END

#define     CCA_sz_HVAC_SHADOW_PREFIX           CCA_sz_SHADOW_BODY CCA_sz_HVAC_SHADOW_NAME
#define     CCA_sz_HVAC_UPDATE_TOPIC            CCA_sz_HVAC_SHADOW_PREFIX CCA_sz_SHADOW_TOPIC_UPDATE_END
#define     CCA_sz_HVAC_UPDATE_DELTA_TOPIC      CCA_sz_HVAC_SHADOW_PREFIX CCA_sz_SHADOW_TOPIC_DELTA_END
#define     CCA_sz_HVAC_UPDATE_ACCEPTED_TOPIC   CCA_sz_HVAC_SHADOW_PREFIX CCA_sz_SHADOW_TOPIC_ACCEPTED_END
#define     CCA_sz_HVAC_UPDATE_REJECTED_TOPIC   CCA_sz_HVAC_SHADOW_PREFIX CCA_sz_SHADOW_TOPIC_REJECTED_END

#define     CCA_sz_VERSION_SHADOW_PREFIX        CCA_sz_SHADOW_BODY CCA_sz_VERSION_SHADOW_NAME
#define     CCA_sz_VERSION_UPDATE_TOPIC         CCA_sz_VERSION_SHADOW_PREFIX CCA_sz_SHADOW_TOPIC_UPDATE_END
#define     CCA_sz_VERSION_UPDATE_DELTA_TOPIC   CCA_sz_VERSION_SHADOW_PREFIX CCA_sz_SHADOW_TOPIC_DELTA_END
#define     CCA_sz_VERSION_UPDATE_ACCEPTED_TOPIC CCA_sz_VERSION_SHADOW_PREFIX CCA_sz_SHADOW_TOPIC_ACCEPTED_END
#define     CCA_sz_VERSION_UPDATE_REJECTED_TOPIC CCA_sz_VERSION_SHADOW_PREFIX CCA_sz_SHADOW_TOPIC_REJECTED_END

#define     CCA_sz_COMMAND_REQUEST_PREFIX       "powersave/command/request/"
#define     CCA_sz_COMMAND_RESPONSE_PREFIX      "powersave/command/response/"
#define     CCA_sz_ALERTS_TOPIC_PREFIX          "powersave/alerts/"

/**
 * @brief Timeout for MQTT_ProcessLoop function in milliseconds.
 */
#define MQTT_PROCESS_LOOP_TIMEOUT_MS (500U)

ModScop extern MQTTContext_t str_MqttContext;

/**
 * @brief Initializes the CCA AWS module
 * 
 * @returns 0 on success, standard POSIX error code upon failure.
 */
ModScop void CCA_AWS_Init(void);

/**
 * @brief Connect to MQTT broker with reconnection retries.
 *
 * If connection fails, retry is attempted after a back-off period.
 * The back-off period will exponentially increase until either the maximum
 * back-off period is reached or the number of attempts are exhausted.
 *
 * @param[out] ptr_str_NetworkContext The output parameter to return the created network context.
 * @param[out] ptr_str_MqttContext The output to return the created MQTT context.
 *
 * @return EXIT_FAILURE on failure; EXIT_SUCCESS on successful connection.
 */
ModScop int CCA_AWS_ConnectToServerWithBackoffRetries(NetworkContext_t *ptr_str_NetworkContext, 
                                                        MQTTContext_t *ptr_str_MqttContext);

/**
 * @brief Utility to subscribe to the passed topic filter and register
 * a callback for it in the subscription manager.
 *
 * The registered callback will be invoked by the subscription manager
 * when PUBLISH messages on topic(s) that match the registered topic filter
 * are received from the broker.
 *
 * @param[in] ptr_str_Context The MQTT context representing the MQTT connection.
 * @param[in] sz_TopicFilter The topic filter to subscribe to and register a
 * callback for in the subscription manager.
 * @param[in] u16_TopicFilterLength The length of the topic filter, @p sz_TopicFilter.
 * @param[in] fn_Callback The callback to register for the topic filter with the
 * subscription manager.
 *
 * @return EXIT_SUCCESS if subscription and callback registration operations
 * for the topic filter were successfully; EXIT_FAILURE otherwise.
 */
ModScop int CCA_AWS_SubscribeToAndRegisterTopicFilter(MQTTContext_t *ptr_str_Context, const char *sz_TopicFilter, 
                                                        uint16_t u16_TopicFilterLength, 
                                                        SubscriptionManagerCallback_t fn_Callback);

/**
 * @brief Sends an MQTT PUBLISH to the passed topic
 * with the passed message and waits for acknowledgement response
 * from the broker. Also, this function processes an incoming PUBLISH
 * packet of the same message that the broker should send back to us.
 *
 * As this demo subscribes to topic filters that match the topics
 * that it publishes to, the broker should send back to us the same
 * message that we publish to it.
 *
 * @param[in] ptr_str_MqttContext MQTT context pointer.
 * @param[in] sz_Topic The topic on which to send a PUBLISH message.
 * @param[in] u16_TopicLength The length of the topic.
 * @param[in] sz_Message The message payload to PUBLISH.
 * @param[in] u16_MessageLength The length of the message payload.
 *
 * @return EXIT_SUCCESS if PUBLISH was successfully sent and processing
 * of incoming PUBLISH was successful; EXIT_FAILURE otherwise.
 */
ModScop int CCA_AWS_PublishToTopic(MQTTContext_t *ptr_str_MqttContext, const char *sz_Topic, uint16_t u16_TopicLength,
                                    const char *sz_Message, uint16_t u16_MessageLength);

/**
 * @brief  Subsribes to neccessary shadow sub-topics and registers callbacks with the appropriate handlers
 * 
 * @param sz_ShadowName             Name of the AWS shadow 
 * @param sz_UpdateDeltaTopic       Shadow Update/Delta Topic 
 * @param sz_UpdateAcceptedTopic    Shadow Update/Accepted Topic 
 * @param sz_UpdateRejectedTopic    Shadow Update/Rejected Topic 
 * 
 * @returns EXIT_SUCCESS or 0 upon success, EXIT_FAILURE upon failure
 */
ModScop int CCA_AWS_SubscribeToShadowSubtopics(const char* sz_ShadowName, const char *sz_UpdateDeltaTopic, 
                                                const char *sz_UpdateAcceptedTopic, const char *sz_UpdateRejectedTopic);

/**
 * @brief  Subsribes to command request topic and registers callback with the appropriate handler
 * 
 * @returns EXIT_SUCCESS or 0 upon success, EXIT_FAILURE upon failure
 */
ModScop int CCA_AWS_SubscribeToCommandRequestTopic(void);


#endif // LION_CCA_AWS_H__
//EOF
