/**********************************************************************************************************************
* @file            CCA_IccCalls.h
*
* @brief           Handles all the data retreival and setting for all devices across the ICC.
*
* @author          Devan Hone
* @date            06/10/2024
* @copyright       2024 Lion Energy LLC.
*                  All rights reserved
*
*********************************************************************************************************************/
/**
* @addtogroup      CCA
*                  Cloud Communication Application Module   
* 
* @ref             CCA_IccCalls.h 
*                  
*********************************************************************************************************************/
#include "COD_Compiler_Defines.h"

#ifndef LION_CCA_ICC_CALLS_H__
#define LION_CCA_ICC_CALLS_H__

#include "CCA_Internal.h"
#include <lion/ial/ccu/IAL_Control.h>
#include <lion/ial/ccu/IAL_Data.h>

#define CCA_sz_TOU_KEY  IAL_DATA_sz_TOU_KEY

/*********************************************************************************************************************
* @cond                                              MODULE SCOPE                                             @endcond
*********************************************************************************************************************/

/**
 * @brief Initialize the ICC module. This should only be called once and before any other function in the module are called.
 */
ModScop void CCA_Icc_Init(void);

/**
 * @brief Registers all callback functions with the ICC.
 */
ModScop void CCA_Icc_RegisterCallbackHandlers(void);

/**
 * @brief Unregisters all functions with the ICC.
 */
ModScop void CCA_Icc_UnregisterCallbackHandlers(void);

ModScop inline int CCA_Icc_GetQuantifiableData(const char *sz_UnitType, uint8_t u8_UnitIndex, uint8_t u8_NumberOfKeys, 
                                        const char **sz_arr_Keys, int32_t *s32_arr_Responses)
{
    return IAL_Data_GetQuantifiableData(sz_UnitType, u8_UnitIndex, u8_NumberOfKeys, sz_arr_Keys, s32_arr_Responses);
}

ModScop inline int CCA_Icc_SetQuantifiableData(const char *sz_UnitType, uint8_t u8_UnitIndex, uint8_t u8_NumberOfKeys, 
                                        const char **sz_arr_Keys, int32_t *s32_arr_Values, int *int_arr_Statuses)
{
    return IAL_Data_SetQuantifiableData(sz_UnitType, u8_UnitIndex, u8_NumberOfKeys, sz_arr_Keys, s32_arr_Values, 
                                        int_arr_Statuses);
}

ModScop inline int CCA_Icc_GetUnitInfo(const char* sz_UnitType, uint8_t u8_UnitIndex, uint8_t u8_NumberOfKeys,
                                const char** sz_arr_Keys, char** sz_arr_Responses)
{
    return IAL_Data_GetUnitInfo(sz_UnitType, u8_UnitIndex, u8_NumberOfKeys, sz_arr_Keys, sz_arr_Responses);
}

ModScop inline int CCA_Icc_GetUnitCodes(const char *sz_UnitType, uint8_t u8_UnitIndex, const char *sz_CodeType, 
                                    char **sz_arr_Responses)
{
    return IAL_Data_GetUnitCodes(sz_UnitType, u8_UnitIndex, sz_CodeType, sz_arr_Responses);
}

ModScop inline int CCA_Icc_GetBufferedData(const char *sz_UnitType, uint8_t u8_UnitIndex, void *vptr_Buff, 
                                    uint32_t *u32ptr_BuffSize)
{
    return IAL_Data_GetBufferedData(sz_UnitType, u8_UnitIndex, vptr_Buff, u32ptr_BuffSize);
}

ModScop inline int CCA_Icc_GetEMSConfiguration(const char *sz_UnitType, uint8_t u8_UnitIndex, uint8_t u8_NumberOfKeys, 
                                        const char** sz_arr_Keys, char ** sz_arr_Responses)
{
    return IAL_Data_GetEMSConfiguration(sz_UnitType, u8_UnitIndex, u8_NumberOfKeys, sz_arr_Keys, sz_arr_Responses);
}

ModScop inline int CCA_Icc_SetEMSConfiguration(const char *sz_UnitType, uint8_t u8_UnitIndex, uint8_t u8_NumberOfKeys, 
                                        const char **sz_arr_Keys, const char **sz_arr_Values, int *int_arr_Statuses)
{
    return IAL_Data_SetEMSConfiguration(sz_UnitType, u8_UnitIndex, u8_NumberOfKeys, sz_arr_Keys, sz_arr_Values, 
                                        int_arr_Statuses);
}

ModScop inline int CCA_Icc_GetUnitConfiguration(const char *sz_UnitType, uint8_t u8_UnitIndex, uint8_t u8_NumberOfKeys, 
                                            const char** sz_arr_Keys, char ** sz_arr_Responses)
{
    return IAL_Data_GetUnitConfiguration(sz_UnitType, u8_UnitIndex, u8_NumberOfKeys, sz_arr_Keys, sz_arr_Responses);
}

ModScop inline int CCA_Icc_SetUnitConfiguration(const char *sz_UnitType, uint8_t u8_UnitIndex, uint8_t u8_NumberOfKeys, 
                                            const char **sz_arr_Keys, const char **sz_arr_Values, int *int_arr_Statuses)
{
    return IAL_Data_SetUnitConfiguration(sz_UnitType, u8_UnitIndex, u8_NumberOfKeys, sz_arr_Keys, sz_arr_Values, 
                                        int_arr_Statuses);
}

ModScop inline int CCA_Icc_GetTouSettings(uint8_t u8_ProfileID, const char **sz_SettingPath, uint8_t u8_SettingDepth,
                                    char *sz_Response)
{
    return IAL_Data_GetTouSettings(u8_ProfileID, sz_SettingPath, u8_SettingDepth, sz_Response);
}

ModScop inline int CCA_Icc_SetTouSettings(uint8_t u8_ProfileID, const char *sz_SerializedSettings, const char *sz_ArrayName)
{
    return IAL_Data_SetTouSettings(u8_ProfileID, sz_SerializedSettings, sz_ArrayName);
}

ModScop inline int CCA_Icc_TestAlertStatusCallback(const char *sz_UnitType, uint8_t u8_UnitIndex)
{
    return IAL_Data_TestAlertStatusCallback(sz_UnitType, u8_UnitIndex);
}

ModScop inline int CCA_Icc_RestartEMS(const char *sz_UnitType)
{
    return IAL_Ctl_RestartEMS(sz_UnitType);
}

ModScop inline int CCA_Icc_ReinitComDCU(void)
{
    return IAL_Ctl_ReinitComDCU();
}

ModScop inline int CCA_Icc_RequestVersionInfo(void)
{
    return IAL_Ctl_RequestVersionInfo();
}


#endif // LION_CCA_ICC_CALLS_H__
//EOF
