/**********************************************************************************************************************
* @file            CCA_Internal.h
*
* @brief           This component contains common definitions and helper functions used throughout the CCA module.
*
* @author          Devan Hone
* @date            06/07/2024
* @copyright       2024 Lion Energy LLC.
*                  All rights reserved
*
*********************************************************************************************************************/
/**
* @addtogroup      CCA
*                  Cloud Communication Application Module 
* 
* @ref             CCA_Internal.h 
*                  
*********************************************************************************************************************/
#include "COD_Compiler_Defines.h"

#ifndef LION_CCA_INTERNAL_H__
#define LION_CCA_INTERNAL_H__

#include "lion/sch/SCH_AtsSchema.h"
#include "lion/sch/SCH_DcdcSchema.h"
#include "lion/sch/SCH_EmsSchema.h"
#include "lion/sch/SCH_GwSchema.h"
#include "lion/sch/SCH_HvacSchema.h"
#include "lion/sch/SCH_PcsSchema.h"
#include "lion/sch/SCH_PduSchema.h"
#include "lion/sch/SCH_PmuSchema.h"
#include "CCA_AWS.h"
#include <ArduinoJson.h>

/*********************************************************************************************************************
* @cond                                              MODULE SCOPE                                             @endcond
*********************************************************************************************************************/

/* AWS UNIT TYPE KEYS */
#define CCA_sz_UNIT_TYPE_ATS     COD_sz_UNIT_TYPE_ATS
#define CCA_sz_UNIT_TYPE_DCDC    COD_sz_UNIT_TYPE_DCDC
#define CCA_sz_UNIT_TYPE_DCU     COD_sz_UNIT_TYPE_DCU
#define CCA_sz_UNIT_TYPE_EMS     COD_sz_UNIT_TYPE_EMS
#define CCA_sz_UNIT_TYPE_GW      COD_sz_UNIT_TYPE_GW
#define CCA_sz_UNIT_TYPE_HVAC    COD_sz_UNIT_TYPE_HVAC
#define CCA_sz_UNIT_TYPE_PCS     COD_sz_UNIT_TYPE_PCS
#define CCA_sz_UNIT_TYPE_PDU     COD_sz_UNIT_TYPE_PDU
#define CCA_sz_UNIT_TYPE_PMU     COD_sz_UNIT_TYPE_PMU

/* AWS JSON DOCUMENT SIZES */
#define CCA_u16_PRODUCT_JSON_DOC_SIZE       2048
#define CCA_u16_COMPONENT_JSON_DOC_SIZE     2048
#define CCA_u16_SETTINGS_JSON_DOC_SIZE      1024
#define CCA_u16_BIG_SETTINGS_JSON_DOC_SIZE  4096
#define CCA_u16_COMMAND_JSON_DOC_SIZE       1024
#define CCA_u16_CONFIG_JSON_DOC_SIZE        1024
#define CCA_u16_ALERT_JSON_DOC_SIZE         1024
#define CCA_u16_DELTA_JSON_DOC_SIZE         256

/* AWS CONFIGURATION CONSTANTS */
#define CCA_sz_EMS_PARALLEL_DISABLED    "0"
#define CCA_sz_EMS_PARALLEL_SLAVE       "1"
#define CCA_sz_EMS_PARALLEL_MASTER      "2"
#define CCA_u8_MAX_COMPONENT_QTY    3
#define CCA_u8_MAX_EMS_QTY          1
#define CCA_u8_MAX_PMU_QTY          1
#define CCA_u8_MAX_GW_QTY           1
#define CCA_u8_UUID_LENGTH          37
#define CCA_u8_PRODUCT_NAME_LENGTH  15
#define CCA_u8_TOPIC_LENGTH         255
#define CCA_u8_ENDPOINT_LENGTH      50
#define CCA_u8_PARALLEL_CFG_LENGTH  10

#define CCA_u8_STRING_BUFFER_SIZE   64
#define MAX_VALUE_SIZE              20 //Maximum number of characters for a value in the JSON key-value pair

/**
 * @brief Enumerated type for the different unit types in the system.
 */
typedef enum { PCS, PDU, PMU, DCDC, GW, HVAC, EMS } unit_type_t;

/**
 * @brief A mapping of key strings and their corresponding location in the device schema structure.
 */
typedef struct cca_key_to_schema_param_map
{
    const char *sz_key; //Key string
    int32_t* const ptr_s32_Val; //Pointer to the value in the schema structure (for signed values)
    uint32_t* const ptr_u32_Val; //Pointer to the value in the schema structure (for unsigned values)
} CCA_KeyToSchemaParamMap_t;

/**
 * @brief A mapping of key strings and their corresponding location in the device schema structure.
 */
typedef struct cca_key_to_schema_code_map
{
    const char *sz_code; //Key string
    const size_t siz_StrOffset; //Offset of the corresponding code in the device schema structure.
} CCA_KeyToSchemaCodeMap_t;

extern char cca_sz_mqtt_endpoint[CCA_u8_ENDPOINT_LENGTH];
extern char cca_sz_thing_name[CCA_u8_UUID_LENGTH];
extern char cca_sz_product_name[CCA_u8_PRODUCT_NAME_LENGTH];
extern char cca_arr_pcs_component_uuids[CCA_u8_MAX_COMPONENT_QTY][CCA_u8_UUID_LENGTH]; 
extern char cca_arr_pdu_component_uuids[CCA_u8_MAX_COMPONENT_QTY][CCA_u8_UUID_LENGTH];
extern char cca_arr_dcdc_component_uuids[CCA_u8_MAX_COMPONENT_QTY][CCA_u8_UUID_LENGTH];
extern char cca_arr_pmu_component_uuids[CCA_u8_MAX_COMPONENT_QTY][CCA_u8_UUID_LENGTH];
extern char cca_arr_hvac_component_uuids[CCA_u8_MAX_COMPONENT_QTY][CCA_u8_UUID_LENGTH];
extern char cca_sz_pcs_update_topic[CCA_u8_TOPIC_LENGTH];
extern char cca_sz_pcs_update_delta_topic[CCA_u8_TOPIC_LENGTH];
extern char cca_sz_pcs_update_accepted_topic[CCA_u8_TOPIC_LENGTH];
extern char cca_sz_pcs_update_rejected_topic[CCA_u8_TOPIC_LENGTH];
extern char cca_sz_dcdc_update_topic[CCA_u8_TOPIC_LENGTH];
extern char cca_sz_dcdc_update_delta_topic[CCA_u8_TOPIC_LENGTH];
extern char cca_sz_dcdc_update_accepted_topic[CCA_u8_TOPIC_LENGTH];
extern char cca_sz_dcdc_update_rejected_topic[CCA_u8_TOPIC_LENGTH];
extern char cca_sz_ems_update_topic[CCA_u8_TOPIC_LENGTH];
extern char cca_sz_ems_update_delta_topic[CCA_u8_TOPIC_LENGTH];
extern char cca_sz_ems_update_accepted_topic[CCA_u8_TOPIC_LENGTH];
extern char cca_sz_ems_update_rejected_topic[CCA_u8_TOPIC_LENGTH];
extern char cca_sz_pmu_update_topic[CCA_u8_TOPIC_LENGTH];
extern char cca_sz_pmu_update_delta_topic[CCA_u8_TOPIC_LENGTH];
extern char cca_sz_pmu_update_accepted_topic[CCA_u8_TOPIC_LENGTH];
extern char cca_sz_pmu_update_rejected_topic[CCA_u8_TOPIC_LENGTH];
extern char cca_sz_hvac_update_topic[CCA_u8_TOPIC_LENGTH];
extern char cca_sz_hvac_update_delta_topic[CCA_u8_TOPIC_LENGTH];
extern char cca_sz_hvac_update_accepted_topic[CCA_u8_TOPIC_LENGTH];
extern char cca_sz_hvac_update_rejected_topic[CCA_u8_TOPIC_LENGTH];
extern char cca_sz_version_update_topic[CCA_u8_TOPIC_LENGTH];
extern char cca_sz_version_update_delta_topic[CCA_u8_TOPIC_LENGTH];
extern char cca_sz_version_update_accepted_topic[CCA_u8_TOPIC_LENGTH];
extern char cca_sz_version_update_rejected_topic[CCA_u8_TOPIC_LENGTH];
extern char cca_sz_command_request_topic[CCA_u8_TOPIC_LENGTH];
extern char cca_sz_command_response_topic[CCA_u8_TOPIC_LENGTH];
extern char cca_sz_alerts_topic[CCA_u8_TOPIC_LENGTH];
extern char cca_sz_ems_parallel_cfg[CCA_u8_PARALLEL_CFG_LENGTH];

extern const char cca_arr_PCS_SERIALS[][16];
extern const char *cca_sz_PCS_MFR;
extern const char cca_arr_PDU_SERIALS[][16];
extern const char *cca_sz_PDU_MFR;
extern const char *cca_sz_PDU_MODEL;
extern const char *cca_sz_DCDC_MFR;
extern const char *cca_sz_PMU_MFR;

/**
 * @brief Converts a unit type string to the corresponding enumerated type.
 * 
 * @param sz_UnitType The unit type string to convert.
 * @param ptr_UnitType Pointer to the unit type to set.
 * 
 * @return EXIT_SUCCESS or 0 upon success, EXIT_FAILURE upon failure
 */
ModScop int CCA_Int_UnitTypeStringToEnum(const char *sz_UnitType, unit_type_t *ptr_UnitType);

/**
 * @brief Retrieves the device and component quantities of the system 
 * from the DCU EMS settings and initializes the internal variables.
 * 
 * @return EXIT_SUCCESS or 0 upon success, EXIT_FAILURE upon failure
*/
ModScop int CCA_Int_InitUnitQuantities(void);

/**
 * @brief Retrieves the quantity of a specific unit type.
 * 
 * @param enm_UnitType The unit type to get the quantity of.
 * @return Quantity of the specified unit type.
 */
ModScop uint8_t CCA_Int_GetUnitQuantity(unit_type_t enm_UnitType);
ModScop uint8_t CCA_Int_GetUnitQuantity_sz(const char *sz_UnitType);

/**
 * @brief Appends an array of key/value pairs to a JSON document.
 * 
 * @param jd_Doc JSON document to append the device data to.
 * @param arr_KeysToParams Array of key to schema parameter mappings.
 * @param siz_Array Size of the array of key to schema parameter mappings.
 * @param bol_CommFailure Failure state of ICC operation prior to this function call. 
 */
ModScop void CCA_Int_AppendToJsonDoc(ArduinoJson::JsonDocument& jd_Doc, const CCA_KeyToSchemaParamMap_t *arr_KeysToParams, 
                                        size_t siz_Array, bool bol_CommFailure);

/**
 * @brief Appends an array of key/value pairs to a JSON object.
 * 
 * @param jo_Obj JSON object to append the data to.
 * @param arr_KeysToParams Array of key to schema parameter mappings.
 * @param siz_Array Size of the array of key to schema parameter mappings.
 * @param bol_CommFailure Failure state of ICC operation prior to this function call. 
 */
ModScop void CCA_Int_AppendToJsonObject(ArduinoJson::JsonObject& jo_Obj, const CCA_KeyToSchemaParamMap_t *arr_KeysToParams, 
                                        size_t siz_Array, bool bol_CommFailure);

/**
 * @brief Appends an array of key/code pairs to a JSON array.
 * 
 * @param ja_Arr JSON array to append the device data to.
 * @param arr_KeysToCodes Array of key to schema code mappings.
 * @param siz_Array Size of the array of key to schema code mappings.
 * @param vptr_CodesStruct Pointer to the structure containing the codes.
 */
ModScop void CCA_Int_AppendCodesToJsonArray(ArduinoJson::JsonArray& ja_Arr, const CCA_KeyToSchemaCodeMap_t *arr_KeysToCodes, 
                                            size_t siz_Array, void *vptr_CodesStruct);

/**
 * @brief Appends Time of Use (ToU) settings to the given JSON object.
 * 
 * @param jo_Obj JSON object to append the ToU settings to. 
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard POSIX error code upon failure.
 */
ModScop int CCA_Int_AppendTouSettingsToJsonObject(ArduinoJson::JsonObject& jo_Obj);

/**
 * @brief Sets EMS settings that need to be done internally on the CCU. Most settings are handled by the DCU with a few 
 *         exceptions, such as system timezone.
 * 
 * @param u8_numberOfKeys   Size of the array of requested Keys
 * @param sz_arr_keys       Requested keys or settings to be written as an array of strings
 * @param sz_arr_values     Requested values of settings to be written as an array of strings
 *  
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
*/
ModScop int CCA_Int_SetCCUSettings(uint8_t u8_numberOfKeys, const char **sz_arr_keys, const char **sz_arr_values);

/**
 * @brief Helper function to get device type from string
 *
 * @param sz_UnitType String representing device type
 * @param ptr_UnitType Pointer to the unit type to be set
 *
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
ModScop int CCA_Int_ValidateUnitType(const char *sz_UnitType, unit_type_t *ptr_UnitType);

/**
 * @brief Helper function to check if unit index is in range
 *
 * @param enm_UnitType Unit type
 * @param u32_UnitIndex Unit index
 *
 * @returns Return 0 upon success, standard posix errors upon failure
 */
ModScop int CCA_Int_ValidateUnitIndex(unit_type_t enm_UnitType, uint32_t u32_UnitIndex);

#endif // LION_CCA_INTERNAL_H__
//EOF
