#include "COD_Compiler_Defines.h"
#include "ICC_Internal.h"
#include "lion/icc/ICC.h"
#include "lion/log/log.h"
#include "lion/utils/byteorder.h"
#include <errno.h>
#include <string.h>
#ifndef __ZEPHYR__
    #include <stdlib.h>
#endif

/**********************************************************************************************************************
 * @cond                                               FILE SCOPE                                              @endcond
 *********************************************************************************************************************/

LOG_MODULE_DECLARE(icc, CONFIG_ICC_LOG_LEVEL);

bool icc_Event_ProcessPacket(ICC_Context_t *context, ICC_Packet_t *ptr_Packet)
{
    if (ptr_Packet->u16Length >= sizeof(icc_EventHeader_t))
    {
        StackScop icc_EventHeader_t str_EventHeader;
        memcpy(&str_EventHeader, ptr_Packet->arrBuffer, sizeof(icc_EventHeader_t));
        if (str_EventHeader.str_Header.u3Type == icc_u8PACKET_TYPE_EVENT)
        {
            str_EventHeader.u32_EventId = sys_le32_to_cpu(str_EventHeader.u32_EventId);
            uint8_t *ptr_Payload        = &ptr_Packet->arrBuffer[sizeof(icc_EventHeader_t)];
            uint16_t u16_PayloadLength  = ptr_Packet->u16Length - sizeof(icc_EventHeader_t);

            LOG_INF("%p: Event Packet (%p) for 0x%08X with %u bytes", context, ptr_Packet,
                    str_EventHeader.u32_EventId, u16_PayloadLength);
            LOG_HEXDUMP_DBG(ptr_Payload, u16_PayloadLength, "Event Payload");

            StackScop ICC_Event_Handler_t *ptr_Handler;
            StackScop ICC_Event_Handler_t *ptr_Next;
            RTOS_MutexLock(&context->mutex, RTOS_FOREVER);
            SYS_DLIST_FOR_EACH_CONTAINER_SAFE(&context->events, ptr_Handler, ptr_Next, node)
            {
                RTOS_MutexUnlock(&context->mutex);

                if (ptr_Handler->u32_EventID == (ptr_Handler->u32_EventMask & str_EventHeader.u32_EventId))
                {
                    ptr_Handler->fn_Callback(str_EventHeader.u32_EventId, ptr_Payload, u16_PayloadLength,
                                             ptr_Handler->ptr_UserData);
                }
                RTOS_MutexLock(&context->mutex, RTOS_FOREVER);
            }
            RTOS_MutexUnlock(&context->mutex);
        }
    }
    return false;
}

/**********************************************************************************************************************
 * @cond                                              GLOBAL SCOPE                                             @endcond
 *********************************************************************************************************************/

#ifndef __ZEPHYR__
ICC_Event_Handler_t *ICC_Event_AllocateHandler() { return malloc(sizeof(ICC_Event_Handler_t)); }
void                    ICC_Event_FreeHandler(ICC_Event_Handler_t *ptr_Context) { free(ptr_Context); }
#endif

void ICC_Event_InitHandler(ICC_Event_Handler_t *ptr_Handler, uint32_t u32_EventID, uint32_t u32_EventMask,
                           ICC_Event_Callback_t fn_Callback, void *ptr_UserData)
{
    sys_dnode_init(&ptr_Handler->node);
    ptr_Handler->u32_EventID   = u32_EventID;
    ptr_Handler->u32_EventMask = u32_EventMask;
    ptr_Handler->fn_Callback   = fn_Callback;
    ptr_Handler->ptr_UserData  = ptr_UserData;
}

int ICC_Event_RegisterHandler(ICC_Context_t *ptr_ICC, ICC_Event_Handler_t *ptr_Handler)
{
    int errcode = 0;
    if (sys_dnode_is_linked(&ptr_Handler->node))
    {
        errcode = -EALREADY;
    }
    else
    {
        RTOS_MutexLock(&ptr_ICC->mutex, RTOS_FOREVER);
        sys_dlist_append(&ptr_ICC->events, &ptr_Handler->node);
        RTOS_MutexUnlock(&ptr_ICC->mutex);
    }
    return errcode;
}

int ICC_Event_UnregisterHandler(ICC_Context_t *ptr_ICC, ICC_Event_Handler_t *ptr_Handler)
{
    RTOS_MutexLock(&ptr_ICC->mutex, RTOS_FOREVER);
    sys_dlist_remove(&ptr_Handler->node);
    RTOS_MutexUnlock(&ptr_ICC->mutex);
    return 0;
}

int ICC_Event_Dispatch(ICC_Context_t *ptr_ICC, uint32_t u32_EventID, const void *ptr_EventData, size_t siz_DataLength,
                       RTOS_Timeout_t tim_Timeout)
{
    LOG_DBG("Dispatching event 0x%08X with %zu bytes", u32_EventID, siz_DataLength);
    ICC_Packet_t *packet = icc_AllocateTXPacket(ptr_ICC, siz_DataLength + 5U);
    if (packet != NULL)
    {
        packet->ptrSource                 = NULL;
        ptr_ICC->u8_Transaction       = (ptr_ICC->u8_Transaction + 1) % 4;
        icc_EventHeader_t str_EventHeader = {.str_Header.u3Type       = icc_u8PACKET_TYPE_EVENT,
                                             .str_Header.flgEncrypted = 0,
                                                .str_Header.u2Transaction = ptr_ICC->u8_Transaction,
                                             .u32_EventId             = sys_cpu_to_le32(u32_EventID)};
        memcpy(packet->arrBuffer, &str_EventHeader, sizeof(icc_EventHeader_t));
        if (ptr_EventData != NULL)
        {
            memcpy(&packet->arrBuffer[sizeof(icc_EventHeader_t)], ptr_EventData, siz_DataLength);
            packet->u16Length = siz_DataLength + sizeof(icc_EventHeader_t);
        }
        else
        {
            packet->u16Length = sizeof(icc_EventHeader_t);
        }
        icc_EnqueuePacket(ptr_ICC, packet);
        return 0;
    }
    else
    {
        return -EAGAIN;
    }
}