/**********************************************************************************************************************
 * @file            SCH_DcdcSchema.h                                                                                                                                                                                                                                                                                     
 *                                                                                                                   
 * @brief           This file declares all the data structures used for DCDC communications.                         
 *                                                                                                                   
 * @author          Devan Hone                                                                                     
 * @date            05/14/2024                                                                                    
 * @copyright       2024 Lion Energy LLC.                                                                                      
 *                  All rights reserved           
 *                                               
 *********************************************************************************************************************/
/**
 * @addtogroup      SCH
 *                  Identifier Description for the EMS Component (SCH)ema module
 * 
 * @ref             SCH_DcdcSchema.h 
 *                  
 *********************************************************************************************************************/

#ifndef LION_SCH_DCDC_SCHEMA_H__
#define LION_SCH_DCDC_SCHEMA_H__

#include "COD_Compiler_Defines.h"

#define SCH_DCDC_MODEL_LEN 22 //DCDC model number length in bytes

#ifdef __cplusplus
extern "C"
{
#endif

/**********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

/**********************************************************************************************************************
* @cond                                              GLOBAL SCOPE                                             @endcond
*********************************************************************************************************************/

typedef struct __attribute__((packed)) sch_dcdc_settings
{
    int32_t buzz_en; //Buzzer function enable, enum [0, 1]
    int32_t epo_en; //Emergency power off (EPO) enable, enum [0, 1]
    int32_t pid_en; //PID control enable, enum [0, 1]
    int32_t pv_1_wu_v; //PV1 wake-up voltage, 0.1V
    int32_t pv_2_wu_v; //PV2 wake-up voltage, 0.1V
    int32_t pv_3_wu_v; //PV3 wake-up voltage, 0.1V
    int32_t pv_1_v_max; //PV1 maximum voltage, 0.1V
    int32_t pv_2_v_max; //PV2 maximum voltage, 0.1V
    int32_t pv_3_v_max; //PV3 maximum voltage, 0.1V
    int32_t dc_bus_op_v; //DC bus operating voltage, 0.1V
    int32_t dc_bus_p_max; //DC bus maximum power, 0.1kW
    int32_t ems_485_int_to; //EMS RS485 interface timeout, seconds [0, 600]
} SCH_DcdcSettings_t;

typedef struct __attribute__((packed)) sch_dcdc_device_info
{
    char model[SCH_DCDC_MODEL_LEN]; //DCDC model number
    int32_t mb_slave_id; //Modbus slave ID [1, 255]
} SCH_DcdcDeviceInfo_t;

typedef struct __attribute__((packed)) sch_dcdc_component_data
{
    //PV Input Power Summary
    int32_t pv_1_v; //PV1 (solar input) voltage, 0.1V
    int32_t pv_1_i; //PV1 (solar input) current, 0.1A
    int32_t pv_1_p; //PV1 (solar input) power, 0.1kW
    int32_t pv_2_v; //PV2 (solar input) voltage, 0.1V
    int32_t pv_2_i; //PV2 (solar input) current, 0.1A
    int32_t pv_2_p; //PV2 (solar input) power, 0.1kW
    int32_t pv_3_v; //PV3 (solar input) voltage, 0.1V
    int32_t pv_3_i; //PV3 (solar input) current, 0.1A
    int32_t pv_3_p; //PV3 (solar input) power, 0.1kW
    int32_t pv_w; //Total solar input power, 0.1kW
    int32_t pv_i; //Total solar input current, 0.1A
    //PV Input Power Details
    int32_t pv_1_str_1_i; //PV1 string 1 current, 0.1A
    int32_t pv_1_str_1_p; //PV1 string 1 power, 0.1kW
    int32_t pv_1_str_2_i; //PV1 string 2 current, 0.1A
    int32_t pv_1_str_2_p; //PV1 string 2 power, 0.1kW
    int32_t pv_2_str_1_i; //PV2 string 1 current, 0.1A
    int32_t pv_2_str_1_p; //PV2 string 1 power, 0.1kW
    int32_t pv_2_str_2_i; //PV2 string 2 current, 0.1A
    int32_t pv_2_str_2_p; //PV2 string 2 power, 0.1kW
    int32_t pv_3_str_1_i; //PV3 string 1 current, 0.1A
    int32_t pv_3_str_1_p; //PV3 string 1 power, 0.1kW
    int32_t pv_3_str_2_i; //PV3 string 2 current, 0.1A
    int32_t pv_3_str_2_p; //PV3 string 2 power, 0.1kW
    //DC Power
    int32_t dc_bus_v; //External DC bus voltage, 0.1V
    int32_t int_dc_bus_v; //Internal DC bus voltage, 0.1V
    int32_t dc_bus_p; //DC bus power, 0.1kW
    int32_t dc_bus_i; //DC bus current, 0.1A
    int32_t dc_bus_i_max; //DC bus maximum current, 0.1A
    int32_t leak_i; //Leakage current, 0.1A
    //Temperature
    int32_t dc_dc_temp; //DC-DC temperature, degrees C
    int32_t amb_temp; //Ambient temperature, degrees C
} SCH_DcdcComponentData_t;

typedef struct __attribute__((packed)) sch_dcdc_historical_data
{
    uint32_t pv_kwh_total; //Total solar input energy, 0.1kWh
} SCH_DcdcHistoricalData_t;

typedef struct __attribute__((packed)) sch_dcdc_codes_status
{
    bool S1_0; //Fault status
    bool S1_1; //Alert status
    bool S1_2; //On status
    bool S1_5; //Derated
    bool S1_7; //Failed status
    bool S1_8; //Online
    bool S1_20; //Communication failure/powered down
    bool S5_4; //PV1 Alarm
    bool S5_5; //PV2 Alarm
    bool S5_6; //PV3 Alarm
    bool S5_7; //PV1 Fault
    bool S5_8; //PV2 Fault
    bool S5_9; //PV3 Fault
    bool S5_10; //PV1 Failed
    bool S5_11; //PV2 Failed
    bool S5_12; //PV3 Failed
    bool S5_13; //PV1 MPPT Applying
    bool S5_14; //PV2 MPPT Applying
    bool S5_15; //PV3 MPPT Applying
    bool S5_16; //PV1 MPPT Generating
    bool S5_17; //PV2 MPPT Generating
    bool S5_18; //PV3 MPPT Generating
    bool S5_19; //PV1 On status
    bool S5_20; //PV2 On status
    bool S5_21; //PV3 On status
    bool A4_1; //Calibration parameter error
} SCH_DcdcCodes_Status_t;

typedef struct __attribute__((packed)) sch_dcdc_codes_alarm
{
    bool A1_10; //Leakage current
    bool A3_0; //EPO (Emergency Power Off) tripped
    bool A4_3; //Flash error
    bool A4_4; //Derating due to internal overtemperature
    bool A4_5; //Derating due to ambient overtemperature
    bool A4_6; //PV1 current recharge
    bool A4_7; //PV2 current recharge
    bool A4_8; //PV3 current recharge
} SCH_DcdcCodes_Alarm_t;

typedef struct __attribute__((packed)) sch_dcdc_codes_fault
{
    //E1_5; No DC/DC Communication Fault (because nightly comm loss is expected)
    bool E2_2; //Component Configuration Settings Reset
    bool F2_2; //DSP Version Error
    bool F3_4; //RS485 Communication Error
    bool F3_5; //Fan 1 Failure
    bool F3_6; //Fan 2 Failure
    bool F3_7; //Fan 3 Failure
    bool F3_8; //Fan 4 Failure
    bool F3_9; //Auxiliary Power 1 Failure
    bool F3_10; //Auxiliary Power 2 Failure
    bool F3_11; //Auxiliary Power 3 Failure
    bool F3_12; //Over Temperature 1
    bool F3_13; //Over Temperature 2
    bool F3_14; //Over Temperature 3
    bool F3_15; //Over Temperature 4
    bool F3_16; //Temperature Sensor 1 Failure
    bool F3_17; //Temperature Sensor 2 Failure
    bool F3_18; //Temperature Sensor 3 Failure
    bool F3_19; //Temperature Sensor 4 Failure
    bool F3_20; //“Start Frequently”?
    bool F3_21; //Input SPD Fault
    bool F3_22; //Output SPF Fault
    bool F3_23; //PID Check Error
    bool F3_24; //Over Current Fault 1
    bool F3_25; //Over Current Fault 2
    bool F3_26; //Over Current Fault 3
    bool F3_27; //Over Current Fault 4
    bool F3_28; //Over Current Fault 5
    bool F3_29; //Over Current Fault 6
    bool F3_30; //Over Current Fault 7
    bool F3_31; //Current Error 1
    bool F3_32; //Current Error 2
    bool F3_33; //Current Error 3
    bool F3_34; //Current Error 4
    bool F3_35; //Current Error 5
    bool F3_36; //Current Error 6
    bool F3_37; //Current Error 7
    bool F3_38; //PV1 Input Over Voltage
    bool F3_39; //PV2 Input Over Voltage
    bool F3_40; //PV3 Input Over Voltage
    bool F3_44; //PV1 Input Reverse Polarity Connected
    bool F3_45; //PV2 Input Reverse Polarity Connected
    bool F3_46; //PV3 Input Reverse Polarity Connected
    bool F3_47; //Overload Timeout
    bool F3_48; //Dry Contact Open Circuit
    bool F3_49; //Bus Over Voltage
    bool F3_52; //Bus Galvanic Break Open Circuit
    bool F3_53; //Bus Galvanic Break Short Circuit
    bool F3_54; //Bus Galvanic Break Hardware Fault
    bool F3_55; //Bus Galvanic Break Device Fault
    bool F3_56; //Insulation Detection Error 1
    bool F3_57; //Insulation Detection Error 2
    bool F3_58; //Bus Voltage Error
    bool F3_59; //Bus Reverse Polarity Connected
    bool F3_60; //PCS Power Off
    bool F3_61; //Total Current Limit
    bool F3_62; //Current Limit Fault 1
    bool F3_63; //Current Limit Fault 2
    bool F3_64; //Current Limit Fault 3
    bool F3_65; //Current Limit Fault 4
    bool F3_66; //Current Limit Fault 5
    bool F3_67; //Current Limit Fault 6
    bool F3_68; //Aging Test Failure
} SCH_DcdcCodes_Fault_t;

typedef struct __attribute__((packed)) sch_dcdc_codes
{
    SCH_DcdcCodes_Status_t status;
    SCH_DcdcCodes_Alarm_t alarm;
    SCH_DcdcCodes_Fault_t fault;
} SCH_DcdcCodes_t;

typedef struct __attribute__((packed)) sch_dcdc_params
{
    SCH_DcdcSettings_t settings;
    SCH_DcdcDeviceInfo_t deviceInfo;
    SCH_DcdcComponentData_t componentData;
    SCH_DcdcHistoricalData_t historicalData;
    SCH_DcdcCodes_t codes;
} SCH_DcdcParameters_t;

/**
 * @brief DCDC Settings X-Macro used for quickly writing operations
 *        for all parameters in the settings struct. This can be used
 *        to parse string literals as well.
 * @note Example:
 *          //Create a struct with all settings
 *          #define X_DCDC(parameter) int32_t parameter;
 *          DCDC_SETTINGS_XMACRO
 *          #undef X_DCDC
*/
#define DCDC_SETTINGS_XMACRO \
    X_DCDC(buzz_en) \
    X_DCDC(epo_en) \
    X_DCDC(pid_en) \
    X_DCDC(pv_1_wu_v) \
    X_DCDC(pv_2_wu_v) \
    X_DCDC(pv_3_wu_v) \
    X_DCDC(pv_1_v_max) \
    X_DCDC(pv_2_v_max) \
    X_DCDC(pv_3_v_max) \
    X_DCDC(dc_bus_op_v) \
    X_DCDC(dc_bus_p_max) \
    X_DCDC(ems_485_int_to)

/**
 * @brief DCDC Device Info X-Macro used for quickly writing operations
 *        for all parameters in the device info struct. This can be used
 *        to parse string literals as well.
 * @note Example:
 *          //Create a struct with all device info data
 *          #define X_DCDC(parameter) int32_t parameter;
 *          DCDC_DEVICE_INFO_XMACRO
 *          #undef X_DCDC
 * @note All string parameters are not fixed length, so they are not included in the X-Macro
*/
#define DCDC_DEVICE_INFO_XMACRO \
    X_DCDC(mb_slave_id)

/**
 * @brief DCDC Component data X-Macro used for quickly writing operations
 *        for all parameters in the struct. This can be used
 *        to parse string literals as well.
 * @note Example:
 *          //Create a struct with all component data
 *          #define X_DCDC(parameter) int32_t parameter;
 *          DCDC_COMPONENT_DATA_XMACRO
 *          #undef X_DCDC
*/
#define DCDC_COMPONENT_DATA_XMACRO \
    X_DCDC(pv_1_v) \
    X_DCDC(pv_1_i) \
    X_DCDC(pv_1_p) \
    X_DCDC(pv_2_v) \
    X_DCDC(pv_2_i) \
    X_DCDC(pv_2_p) \
    X_DCDC(pv_3_v) \
    X_DCDC(pv_3_i) \
    X_DCDC(pv_3_p) \
    X_DCDC(pv_w) \
    X_DCDC(pv_i) \
    X_DCDC(pv_1_str_1_i) \
    X_DCDC(pv_1_str_1_p) \
    X_DCDC(pv_1_str_2_i) \
    X_DCDC(pv_1_str_2_p) \
    X_DCDC(pv_2_str_1_i) \
    X_DCDC(pv_2_str_1_p) \
    X_DCDC(pv_2_str_2_i) \
    X_DCDC(pv_2_str_2_p) \
    X_DCDC(pv_3_str_1_i) \
    X_DCDC(pv_3_str_1_p) \
    X_DCDC(pv_3_str_2_i) \
    X_DCDC(pv_3_str_2_p) \
    X_DCDC(dc_bus_v) \
    X_DCDC(int_dc_bus_v) \
    X_DCDC(dc_bus_p) \
    X_DCDC(dc_bus_i) \
    X_DCDC(dc_bus_i_max) \
    X_DCDC(leak_i) \
    X_DCDC(dc_dc_temp) \
    X_DCDC(amb_temp)

/**
 * @brief DCDC historical data X-Macro used for quickly writing operations
 *        for all parameters in the historical data struct. This can be used
 *        to parse string literals as well.
 */
#define DCDC_HISTORICAL_DATA_XMACRO \
    X_DCDC(pv_kwh_total)

/**
 * @brief DCDC codes (STATUS) X-Macro used for quickly writing operations
 *        for all parameters in the codes struct. This can be used
 *        to parse string literals as well.
 * @note Example:
 *          //Create a bitfield struct with all codes
 *          #define X_DCDC(parameter) bool parameter: 1;
 *          DCDC_CODES_STATUS_XMACRO
 *          #undef X_DCDC
*/
#define DCDC_CODES_STATUS_XMACRO \
    X_DCDC(S1_0) \
    X_DCDC(S1_1) \
    X_DCDC(S1_2) \
    X_DCDC(S1_5) \
    X_DCDC(S1_7) \
    X_DCDC(S1_8) \
    X_DCDC(S1_20) \
    X_DCDC(S5_4) \
    X_DCDC(S5_5) \
    X_DCDC(S5_6) \
    X_DCDC(S5_7) \
    X_DCDC(S5_8) \
    X_DCDC(S5_9) \
    X_DCDC(S5_10) \
    X_DCDC(S5_11) \
    X_DCDC(S5_12) \
    X_DCDC(S5_13) \
    X_DCDC(S5_14) \
    X_DCDC(S5_15) \
    X_DCDC(S5_16) \
    X_DCDC(S5_17) \
    X_DCDC(S5_18) \
    X_DCDC(S5_19) \
    X_DCDC(S5_20) \
    X_DCDC(S5_21) \
    X_DCDC(A4_1)

/**
 * @brief DCDC codes (ALERT) X-Macro used for quickly writing operations
 *        for all parameters in the codes struct. This can be used
 *        to parse string literals as well.
 * @note Example:
 *          //Create a bitfield struct with all codes
 *          #define X_DCDC(parameter) bool parameter: 1;
 *          DCDC_CODES_ALARM_XMACRO
 *          #undef X_DCDC
*/
#define DCDC_CODES_ALARM_XMACRO \
    X_DCDC(A1_10) \
    X_DCDC(A3_0) \
    X_DCDC(A4_3) \
    X_DCDC(A4_4) \
    X_DCDC(A4_5) \
    X_DCDC(A4_6) \
    X_DCDC(A4_7) \
    X_DCDC(A4_8)

/**
 * @brief DCDC codes (FAULT) X-Macro used for quickly writing operations
 *        for all parameters in the codes struct. This can be used
 *        to parse string literals as well.
 * @note Example:
 *          //Create a bitfield struct with all codes
 *          #define X_DCDC(parameter) bool parameter: 1;
 *          DCDC_CODES_FAULT_XMACRO
 *          #undef X_DCDC
*/
#define DCDC_CODES_FAULT_XMACRO \
    X_DCDC(E2_2) \
    X_DCDC(F2_2) \
    X_DCDC(F3_4) \
    X_DCDC(F3_5) \
    X_DCDC(F3_6) \
    X_DCDC(F3_7) \
    X_DCDC(F3_8) \
    X_DCDC(F3_9) \
    X_DCDC(F3_10) \
    X_DCDC(F3_11) \
    X_DCDC(F3_12) \
    X_DCDC(F3_13) \
    X_DCDC(F3_14) \
    X_DCDC(F3_15) \
    X_DCDC(F3_16) \
    X_DCDC(F3_17) \
    X_DCDC(F3_18) \
    X_DCDC(F3_19) \
    X_DCDC(F3_20) \
    X_DCDC(F3_21) \
    X_DCDC(F3_22) \
    X_DCDC(F3_23) \
    X_DCDC(F3_24) \
    X_DCDC(F3_25) \
    X_DCDC(F3_26) \
    X_DCDC(F3_27) \
    X_DCDC(F3_28) \
    X_DCDC(F3_29) \
    X_DCDC(F3_30) \
    X_DCDC(F3_31) \
    X_DCDC(F3_32) \
    X_DCDC(F3_33) \
    X_DCDC(F3_34) \
    X_DCDC(F3_35) \
    X_DCDC(F3_36) \
    X_DCDC(F3_37) \
    X_DCDC(F3_38) \
    X_DCDC(F3_39) \
    X_DCDC(F3_40) \
    X_DCDC(F3_44) \
    X_DCDC(F3_45) \
    X_DCDC(F3_46) \
    X_DCDC(F3_47) \
    X_DCDC(F3_48) \
    X_DCDC(F3_49) \
    X_DCDC(F3_52) \
    X_DCDC(F3_53) \
    X_DCDC(F3_54) \
    X_DCDC(F3_55) \
    X_DCDC(F3_56) \
    X_DCDC(F3_57) \
    X_DCDC(F3_58) \
    X_DCDC(F3_59) \
    X_DCDC(F3_60) \
    X_DCDC(F3_61) \
    X_DCDC(F3_62) \
    X_DCDC(F3_63) \
    X_DCDC(F3_64) \
    X_DCDC(F3_65) \
    X_DCDC(F3_66) \
    X_DCDC(F3_67) \
    X_DCDC(F3_68)

#ifdef __cplusplus
}
#endif

#endif /* LION_SCH_DCDC_SCHEMA_H__ */