/**********************************************************************************************************************
* @file            SCH_EmsSchema.h
*
* @brief           This file declares all the data structures used for monitoring and controlling the EMS.
*
* @author          Devan Hone
* @date            07/18/2024
* @copyright       2024 Lion Energy LLC.
*                  All rights reserved
*
*********************************************************************************************************************/
/**
* @addtogroup      SCH
*                  Identifier Description for the (SCH)ema module 
* 
* @ref             SCH_EmsSchema.h 
*                  
*********************************************************************************************************************/

#ifndef LION_SCH_EMS_SCHEMA_H__
#define LION_SCH_EMS_SCHEMA_H__

#include "COD_Compiler_Defines.h"


#ifdef __cplusplus
extern "C"
{
#endif

/**********************************************************************************************************************
* @cond                                              GLOBAL SCOPE                                             @endcond
*********************************************************************************************************************/

#define SCH_EMS_IP_ADDR_LEN 16
#define SCH_EMS_NETMASK_LEN 16
#define SCH_EMS_UUID_LEN 40
#define SCH_EMS_TZ_LEN 40

//Maximally defined ToU  profiles (4 profiles x (6 effective dates, 7 days x 6 timeslots)
#define SCH_EMS_TOU_MAX_PROFILES    4
#define SCH_EMS_TOU_MAX_EFF_DATES   6
#define SCH_EMS_TOU_MAX_TIMESLOTS   6
#define SCH_EMS_TOU_DAYS_IN_WEEK    7

typedef struct __attribute__((packed)) sch_ems_settings
{
    //Basic Settings
    int32_t sys_work_md; //System Work Mode, enum [0, 2] U [255]
    int32_t cmsn_md_en; //Commissioning Mode Enable, enum [0, 1]
    int32_t diag_md_en; //Diagnostic Mode Enable, enum [0, 1]
    int32_t on_grid_dod; //On-Grid Depth-of-Discharge, 0.10%, [150, 900]
    int32_t off_grid_dod; //Off-Grid Depth-of-Discharge, 0.10%, [150, 1000]
    int32_t chg_hyst; //Charging Hysteresis, 0.10%, [0, 1000]
    int32_t dchg_hyst; //Discharging Hysteresis, 0.10%, [0, 1000]
    int32_t grid_sell; //Grid Selling, enum [0, 1]
    int32_t grid_bf_kW; //Grid Sell Limit, 0.01kW, [0, 9000]
    int32_t mstr_grid_ch; //Master grid charge Enable, enum [0, 1]
    int32_t sleep_time_lim; //Sleep Time Inactivity Limit, min
    int32_t sleep_md_en; //Sleep Mode Enable, enum [0, 1]
    int32_t pcs_qty; //PCS Quantity, [1, 3]
    int32_t dcdc_qty; //DCDC Quantity, [0, 2]
    int32_t pdu_qty; //PDU Quantity, [1, 2]
    int32_t hvac_qty; //HVAC Quantity, [1, 2]
    int32_t pmu_qty; //PMU Quantity, [1, 2]
    int32_t parallel_cfg; //Parallel EMS Configuration, [0, 2]
    int32_t parallel_qty; //Parallel EMS Quantity, [1, 4]
    int32_t ls_en; //Load Shedding Enable, [0, 1]
    //Peak Shaving Settings
    int32_t ps_sp; //Peak Shaving Power Setpoint, 0.01kW, [0, 65535]
    int32_t ps_prop; //Peak Shaving Algorithm - Proportional (Gain), 0.01
    int32_t ps_integ; //Peak Shaving Algorithm - Integral, 0.01
    int32_t ps_deriv; //Peak Shaving Algorithm - Derivative, 0.01
    //Power Factor Correction Settings
    int32_t pfc_en; //Enable Power Factor Correction, bool, [0, 1]
    int32_t pfc_sp; //Power Factor Correction Setpoint (PF), 0.01, [0, 100]
    int32_t pfc_p_min; //Power Factor Correction - Minimum Power Threshold, 0.01kW, [0, 65535]
    int32_t pfc_prop; //Power Factor Correction Algorithm - Proportional (Gain), 0.01
    int32_t pfc_integ; //Power Factor Correction Algorithm - Integral, 0.01
    int32_t pfc_deriv; //Power Factor Correction Algorithm - Derivative, 0.01
    //Charging Profile Settings
    int32_t ch_i_max; //Maximum Charge Current, 0.1A
    int32_t ch_i_derat_slope; //Current Derating Slope, 0.01%, [0, 10000]
    int32_t ch_pro_cell_v_1; //Charging Profile - Cell Voltage Setpoint 1, mV, [0, 5000]
    int32_t ch_pro_i_1; //Charging Profile - Current Level 1, 0.1A
    int32_t ch_pro_cell_v_2; //Charging Profile - Cell Voltage Setpoint 2, mV, [0, 5000]
    int32_t ch_pro_i_2; //Charging Profile - Current Level 2, 0.1A
    int32_t ch_pro_cell_v_3; //Charging Profile - Cell Voltage Setpoint 3, mV, [0, 5000]
    int32_t ch_pro_i_3; //Charging Profile - Current Level 3, 0.1A
    int32_t ch_pro_prop; //Charging Profile Algorithm - Proportional (Gain), 0.01
    int32_t ch_pro_integ; //Charging Profile Algorithm - Integral, 0.01
    int32_t ch_pro_deriv; //Charging Profile Algorithm - Derivate, 0.01
} SCH_EmsSettings_t;

typedef struct __attribute__((packed)) sch_ems_device_info
{
    char ip_addr[SCH_EMS_IP_ADDR_LEN];
    char netmask[SCH_EMS_NETMASK_LEN];
    char uuid[SCH_EMS_UUID_LEN];
    char timezone[SCH_EMS_TZ_LEN];
} SCH_EmsDeviceInfo_t;

typedef struct __attribute__((packed)) sch_ems_tou_timeslot
{
    uint8_t u8_Hours; //Start time hours, [0, 23]
    uint8_t u8_Minutes; //Start time minutes, [0, 59]
    uint8_t u8_Seconds; //Start time seconds, [0, 59]
    uint8_t u8_MinSOC; //Minimum State of Charge, [0, 100]
    uint8_t u8_GridChg;
} SCH_EmsTouTimeslot_t;

typedef struct __attribute__((packed)) sch_ems_tou_day
{
    uint8_t u8_TimeSlotQty; //Number of used timeslots
    SCH_EmsTouTimeslot_t arr_str_TimeSlots[SCH_EMS_TOU_MAX_TIMESLOTS];
} SCH_EmsTouDay_t;

typedef struct __attribute__((packed)) sch_ems_tou_effdates
{
    uint8_t u8_StartMonth; //Profile effective start month, [0, 11]
    uint8_t u8_StartDay; //Profile effective start day, [1, 31]
    uint8_t u8_StopMonth; //Profile effective stop month, [0, 11]
    uint8_t u8_StopDay; //Profile effective stop day, [1, 31]
} SCH_EmsTouEffDates_t;

typedef struct __attribute__((packed)) sch_ems_tou_profile
{
    uint8_t u8_EffDatesQty; //Number of used effective dates
    SCH_EmsTouEffDates_t arr_str_EffDates[SCH_EMS_TOU_MAX_EFF_DATES];
    SCH_EmsTouDay_t arr_str_Days[SCH_EMS_TOU_DAYS_IN_WEEK]; //Sunday to Saturday
} SCH_EmsTouProfile_t;
typedef struct __attribute__((packed)) sch_ems_tou_settings
{
    uint8_t u8_ProfileQty; //Number of used profiles
    SCH_EmsTouProfile_t arr_str_Profiles[SCH_EMS_TOU_MAX_PROFILES];
} SCH_EmsTouSettings_t;

typedef struct __attribute__((packed)) sch_ems_core_data
{
    int32_t ats_op_st; //ATS operation state, enum [0, 4]
    int32_t bpm_ctl_st; //Battery Priority Mode Control State, enum [0, 6]
    int32_t grid_ctl_st; //Grid Control State, enum [0, 6]
    int32_t lgc_ctl_st; //Limit Grid Consumption Control State, enum [0, 7]
    int32_t pcs_op_st; //PCS Operation State, enum [0, 7]
    int32_t psm_st; //Power Manager State, enum [0, 14]
    int32_t pvc_op_st; //PV Controller Operation State, enum [0, 4]
} SCH_EmsCoreData_t;

typedef struct __attribute__((packed)) sch_ems_codes_alarm
{
    bool A1_19; //Grid Disconnnected
    bool A5_4; //Discharging Derated due to Battery Over-Temperature
    bool A5_5; //Charging Derated due to Battery Over-Temperature
    bool A5_6; //Discharging Derated due to Battery Under-Temperature
    bool A5_7; //Charging Derated due to Battery Under-Temperature
    bool F5_9; //Charging Disabled due to Cell Over-Voltage
} SCH_EmsCodes_Alarm_t;

typedef struct __attribute__((packed)) sch_ems_codes_fault
{
    bool F5_6; //Discharging Disabled due to Battery Over-Temperature
    bool F5_7; //Charging Disabled due to Battery Under-Temperature
    bool F5_8; //Discharging Disabled due to Cell Under-Voltage
    bool F5_10; //High Ambient Temperature Fault (Possible Fire Detected)
} SCH_EmsCodes_Fault_t;

typedef struct __attribute__((packed)) sch_ems_codes
{
    SCH_EmsCodes_Alarm_t alarm;
    SCH_EmsCodes_Fault_t fault;
} SCH_EmsCodes_t;

typedef struct __attribute__((packed)) sch_ems_params
{
    //TODO Integrate CMA_EMS and settings
    // SCH_EmsSettings_t settings;
    // SCH_EmsDeviceInfo_t deviceInfo;
    //TODO Add Tou Settings
    SCH_EmsCoreData_t coreData;
    SCH_EmsCodes_t codes;
} SCH_EmsParameters_t;

/**
 * @brief EMS Settings X-Macro used for quickly writing operations
 *        for all parameters in the settings struct. This can be used
 *        to parse string literals as well.
*/
#define EMS_SETTINGS_XMACRO \
    X_EMS(sys_work_md) \
    X_EMS(cmsn_md_en) \
    X_EMS(diag_md_en) \
    X_EMS(on_grid_dod) \
    X_EMS(off_grid_dod) \
    X_EMS(chg_hyst) \
    X_EMS(dchg_hyst) \
    X_EMS(grid_sell) \
    X_EMS(grid_bf_kW) \
    X_EMS(mstr_grid_ch) \
    X_EMS(sleep_time_lim) \
    X_EMS(sleep_md_en) \
    X_EMS(pcs_qty) \
    X_EMS(dcdc_qty) \
    X_EMS(pdu_qty) \
    X_EMS(hvac_qty) \
    X_EMS(pmu_qty) \
    X_EMS(parallel_cfg) \
    X_EMS(parallel_qty) \
    X_EMS(ls_en) \
    X_EMS(ps_sp) \
    X_EMS(ps_prop) \
    X_EMS(ps_integ) \
    X_EMS(ps_deriv) \
    X_EMS(pfc_en) \
    X_EMS(pfc_sp) \
    X_EMS(pfc_p_min) \
    X_EMS(pfc_prop) \
    X_EMS(pfc_integ) \
    X_EMS(pfc_deriv) \
    X_EMS(ch_i_max) \
    X_EMS(ch_i_derat_slope) \
    X_EMS(ch_pro_cell_v_1) \
    X_EMS(ch_pro_i_1) \
    X_EMS(ch_pro_cell_v_2) \
    X_EMS(ch_pro_i_2) \
    X_EMS(ch_pro_cell_v_3) \
    X_EMS(ch_pro_i_3) \
    X_EMS(ch_pro_prop) \
    X_EMS(ch_pro_integ) \
    X_EMS(ch_pro_deriv)

#define EMS_DEVICE_INFO_XMACRO \
    X_EMS(ip_addr) \
    X_EMS(netmask) \
    X_EMS(uuid) \
    X_EMS(timezone)

/**
 * @brief EMS Component data X-Macro used for quickly writing operations
 *        for all parameters in the struct. This can be used
 *        to parse string literals as well.
*/
#define EMS_CORE_DATA_XMACRO \
    X_EMS(ats_op_st) \
    X_EMS(bpm_ctl_st) \
    X_EMS(grid_ctl_st) \
    X_EMS(lgc_ctl_st) \
    X_EMS(pcs_op_st) \
    X_EMS(psm_st) \
    X_EMS(pvc_op_st)

/**
 * @brief EMS codes (ALERT) X-Macro used for quickly writing operations
 *        for all parameters in the codes struct. This can be used
 *        to parse string literals as well.
*/
#define EMS_CODES_ALARM_XMACRO \
    X_EMS(A1_19) \
    X_EMS(A5_4) \
    X_EMS(A5_5) \
    X_EMS(A5_6) \
    X_EMS(A5_7) \
    X_EMS(F5_9)

/**
 * @brief EMS codes (FAULT) X-Macro used for quickly writing operations
 *        for all parameters in the codes struct. This can be used
 *        to parse string literals as well.
*/
#define EMS_CODES_FAULT_XMACRO \
    X_EMS(F5_6) \
    X_EMS(F5_7) \
    X_EMS(F5_8) \
    X_EMS(F5_10)

#ifdef __cplusplus
}
#endif

#endif /* LION_SCH_EMS_SCHEMA_H__ */