/**********************************************************************************************************************
 * @file            SCH_GwSchema.h                                                                                                                                                                                                                                                                                     
 *                                                                                                                   
 * @brief           This file declares all the data structures used for Gateway communications.                         
 *                                                                                                                   
 * @author          Devan Hone                                                                                     
 * @date            05/17/2024                                                                                    
 * @copyright       2024 Lion Energy LLC.                                                                                      
 *                  All rights reserved           
 *                                               
 *********************************************************************************************************************/
/**
 * @addtogroup      SCH
 *                  Identifier Description for the EMS Component (SCH)ema module
 * 
 * @ref             SCH_GwSchema.h 
 *                  
 *********************************************************************************************************************/

#ifndef LION_SCH_GW_SCHEMA_H__
#define LION_SCH_GW_SCHEMA_H__

#include "COD_Compiler_Defines.h"

#ifdef __cplusplus
extern "C"
{
#endif

/**********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

/**********************************************************************************************************************
* @cond                                              GLOBAL SCOPE                                             @endcond
*********************************************************************************************************************/

typedef struct __attribute__((packed)) sch_gw_component_data
{
    int32_t cab_temp; //Cabinet temperature, 0.1C
    int32_t cab_hum; //Cabinet humidity, 0.1%
} SCH_GwComponentData_t;

typedef struct __attribute__((packed)) sch_gw_codes_status
{
    bool S1_0; //Fault status
    bool S1_1; //Alert status
    bool S1_2; //Online status
} SCH_GwCodes_Status_t;

typedef struct __attribute__((packed)) sch_gw_codes_alarm
{
    bool A3_0; //Emergency Power Off (EPO) tripped
    bool A5_3; //Entrance guard triggered
} SCH_GwCodes_Alarm_t;

typedef struct __attribute__((packed)) sch_gw_codes_fault
{
    bool E1_7; //Gateway communication fault
    bool F5_2; //Single fire alarm
    bool F5_3; //Composite fire alarm
    bool F5_4; //Fire controller failure
    bool F5_5; //Excess hydrogen gas detected
} SCH_GwCodes_Fault_t;

typedef struct __attribute__((packed)) sch_gw_codes
{
    SCH_GwCodes_Status_t status;
    SCH_GwCodes_Alarm_t alarm;
    SCH_GwCodes_Fault_t fault;
} SCH_GwCodes_t;

typedef struct __attribute__((packed)) sch_gw_params
{
    SCH_GwComponentData_t componentData;
    SCH_GwCodes_t codes;
} SCH_GwParameters_t;

/**
 * @brief GW Component data X-Macro used for quickly writing operations
 *        for all parameters in the struct. This can be used
 *        to parse string literals as well.
 * @note Example:
 *          //Create a struct with all component data
 *          #define X_GW(parameter) int32_t parameter;
 *          GW_COMPONENT_DATA_XMACRO
 *          #undef X_GW
*/
#define GW_COMPONENT_DATA_XMACRO \
    X_GW(cab_temp) \
    X_GW(cab_hum)

/**
 * @brief GW codes (STATUS) X-Macro used for quickly writing operations
 *        for all parameters in the codes struct. This can be used
 *        to parse string literals as well.
 * @note Example:
 *          //Create a bitfield struct with all codes
 *          #define X_GW(parameter) bool parameter: 1;
 *          GW_CODES_STATUS_XMACRO
 *          #undef X_GW
*/
#define GW_CODES_STATUS_XMACRO \
    X_GW(S1_0) \
    X_GW(S1_1) \
    X_GW(S1_2)

/**
 * @brief GW codes (ALERT) X-Macro used for quickly writing operations
 *        for all parameters in the codes struct. This can be used
 *        to parse string literals as well.
 * @note Example:
 *          //Create a bitfield struct with all codes
 *          #define X_GW(parameter) bool parameter: 1;
 *          GW_CODES_ALARM_XMACRO
 *          #undef X_GW
*/
#define GW_CODES_ALARM_XMACRO \
    X_GW(A3_0) \
    X_GW(A5_3)

/**
 * @brief GW codes (FAULT) X-Macro used for quickly writing operations
 *        for all parameters in the codes struct. This can be used
 *        to parse string literals as well.
 * @note Example:
 *          //Create a bitfield struct with all codes
 *          #define X_GW(parameter) bool parameter: 1;
 *          GW_CODES_FAULT_XMACRO
 *          #undef X_GW
*/
#define GW_CODES_FAULT_XMACRO \
    X_GW(E1_7) \
    X_GW(F5_2) \
    X_GW(F5_3) \
    X_GW(F5_4) \
    X_GW(F5_5)

#ifdef __cplusplus
}
#endif

#endif /* LION_SCH_GW_SCHEMA_H__ */