/**********************************************************************************************************************
 * @file            SCH_HvacSchema.h                                                                                                                                                                                                                                                                                     
 *                                                                                                                   
 * @brief           This file declares all the data structures used for HVAC communications.                         
 *                                                                                                                   
 * @author          Devan Hone                                                                                     
 * @date            05/17/2024                                                                                    
 * @copyright       2024 Lion Energy LLC.                                                                                      
 *                  All rights reserved           
 *                                               
 *********************************************************************************************************************/
/**
 * @addtogroup      SCH
 *                  Identifier Description for the EMS Component (SCH)ema module
 * 
 * @ref             SCH_HvacSchema.h 
 *                  
 *********************************************************************************************************************/

#ifndef LION_SCH_HVAC_SCHEMA_H__
#define LION_SCH_HVAC_SCHEMA_H__

#include "COD_Compiler_Defines.h"

#ifdef __cplusplus
extern "C"
{
#endif

/**********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

/**********************************************************************************************************************
* @cond                                              GLOBAL SCOPE                                             @endcond
*********************************************************************************************************************/

typedef struct __attribute__((packed)) sch_hvac_settings
{
    int32_t cool_sp; //Cooling setpoint, 0.1C [150, 500]
    int32_t cool_buff; //Cooling buffer, 0.1C [10, 100]
    int32_t heat_sp; //Heating stop setpoint, 0.1C [-150, 150]
    int32_t heat_buff; //Heating buffer, 0.1C [10, 100]
    int32_t high_temp_sp; //High temperature setpoint, 0.1C [250, 800]
    int32_t low_temp_sp; //Low temperature setpoint, 0.1C [-200, 150]
} SCH_HvacSettings_t;

typedef struct __attribute__((packed)) sch_hvac_component_data
{
    int32_t hvac_temp; //HVAC temperature, 0.1C
    int32_t hvac_hum; //HVAC humidity, 0.1%
} SCH_HvacComponentData_t;

typedef struct __attribute__((packed)) sch_hvac_codes_status
{
    bool S1_2; //Online status
} SCH_HvacCodes_Status_t;

typedef struct __attribute__((packed)) sch_hvac_codes_alarm
{
    bool A5_0; //HVAC voltage alarm 1
    bool A5_1; //HVAC voltage alarm 2
    bool A5_2; //HVAC temperature alarm
} SCH_HvacCodes_Alarm_t;

typedef struct __attribute__((packed)) sch_hvac_codes_fault
{
    bool E1_6; //HVAC communication fault
    bool F5_0; //HVAC temperature sensor failure
    bool F5_1; //HVAC compressor failure
} SCH_HvacCodes_Fault_t;

typedef struct __attribute__((packed)) sch_hvac_codes
{
    SCH_HvacCodes_Status_t status;
    SCH_HvacCodes_Alarm_t alarm;
    SCH_HvacCodes_Fault_t fault;
} SCH_HvacCodes_t;

typedef struct __attribute__((packed)) sch_hvac_params
{
    SCH_HvacSettings_t settings;
    SCH_HvacComponentData_t componentData;
    SCH_HvacCodes_t codes;
} SCH_HvacParameters_t;

/**
 * @brief HVAC Settings X-Macro used for quickly writing operations
 *        for all parameters in the settings struct. This can be used
 *        to parse string literals as well.
 * @note Example:
 *          //Create a struct with all settings
 *          #define X_HVAC(parameter) int32_t parameter;
 *          HVAC_SETTINGS_XMACRO
 *          #undef X_HVAC
*/
#define HVAC_SETTINGS_XMACRO \
    X_HVAC(cool_sp) \
    X_HVAC(cool_buff) \
    X_HVAC(heat_sp) \
    X_HVAC(heat_buff) \
    X_HVAC(high_temp_sp) \
    X_HVAC(low_temp_sp)

/**
 * @brief HVAC Component data X-Macro used for quickly writing operations
 *        for all parameters in the struct. This can be used
 *        to parse string literals as well.
 * @note Example:
 *          //Create a struct with all component data
 *          #define X_HVAC(parameter) int32_t parameter;
 *          HVAC_COMPONENT_DATA_XMACRO
 *          #undef X_HVAC
*/
#define HVAC_COMPONENT_DATA_XMACRO \
    X_HVAC(hvac_temp) \
    X_HVAC(hvac_hum)

/**
 * @brief HVAC codes (STATUS) X-Macro used for quickly writing operations
 *        for all parameters in the codes struct. This can be used
 *        to parse string literals as well.
 * @note Example:
 *          //Create a bitfield struct with all codes
 *          #define X_HVAC(parameter) bool parameter: 1;
 *          HVAC_CODES_STATUS_XMACRO
 *          #undef X_HVAC
*/
#define HVAC_CODES_STATUS_XMACRO \
    X_HVAC(S1_2)

/**
 * @brief HVAC codes (ALERT) X-Macro used for quickly writing operations
 *        for all parameters in the codes struct. This can be used
 *        to parse string literals as well.
 * @note Example:
 *          //Create a bitfield struct with all codes
 *          #define X_HVAC(parameter) bool parameter: 1;
 *          HVAC_CODES_ALARM_XMACRO
 *          #undef X_HVAC
*/
#define HVAC_CODES_ALARM_XMACRO \
    X_HVAC(A5_0) \
    X_HVAC(A5_1) \
    X_HVAC(A5_2)

/**
 * @brief HVAC codes (FAULT) X-Macro used for quickly writing operations
 *        for all parameters in the codes struct. This can be used
 *        to parse string literals as well.
 * @note Example:
 *          //Create a bitfield struct with all codes
 *          #define X_HVAC(parameter) bool parameter: 1;
 *          HVAC_CODES_FAULT_XMACRO
 *          #undef X_HVAC
*/
#define HVAC_CODES_FAULT_XMACRO \
    X_HVAC(E1_6) \
    X_HVAC(F5_0) \
    X_HVAC(F5_1)

#ifdef __cplusplus
}
#endif

#endif /* LION_SCH_HVAC_SCHEMA_H__ */