/**********************************************************************************************************************
 * @file            SCH_PcsSchema.h                                                                                                                                                                                                                                                                                     
 *                                                                                                                   
 * @brief           This file declares all the data structures used for PCS communications.                         
 *                                                                                                                   
 * @author          Devan Hone                                                                                     
 * @date            05/02/2024                                                                                    
 * @copyright       2024 Lion Energy LLC.                                                                                      
 *                  All rights reserved           
 *                                               
 *********************************************************************************************************************/
/**
 * @addtogroup      SCH
 *                  Identifier Description for the EMS Component (SCH)ema module
 * 
 * @ref             SCH_PcsSchema.h 
 *                  
 *********************************************************************************************************************/

#ifndef LION_SCH_PCS_SCHEMA_H__
#define LION_SCH_PCS_SCHEMA_H__

#include "COD_Compiler_Defines.h"

#define SCH_PCS_IP_ADDR_LEN 16
#define SCH_PCS_NETMASK_LEN 16
#define SCH_PCS_GW_IP_ADDR_LEN 16
#define SCH_PCS_MAC_ADDR_LEN 8

#ifdef __cplusplus
extern "C"
{
#endif

/**********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

/**********************************************************************************************************************
* @cond                                              GLOBAL SCOPE                                             @endcond
*********************************************************************************************************************/

typedef struct __attribute__((packed)) sch_pcs_settings
{
    int32_t off_grid_v_reg; //Off grid voltage regulation [-800, 120]
    int32_t off_grid_f_reg; //Off grid frequency regulation [-5, 5]
    int32_t lim_i_st_md; //Limit in-rush start mode, enum [0, 1]
    int32_t grid_v; //Rated grid voltage, enum [0, 9]
    int32_t grid_f; //Rated grid frequency, enum [0, 1]
    int32_t bat_ch_v_end; //"Topping" or end charge voltage [1500, 7500]
    int32_t bat_ch_i_max; //Battery max charge current [0, 3200]
    int32_t bat_dch_i_max; //Battery max discharge current [0, 3200]
    int32_t bat_dch_v_min; //Battery discharge voltage minimum (end voltage) [1500, 7500]
    int32_t bat_i_config_max; //Battery current configuration maximimum [0, 3200]
    int32_t dc_low_v; //Battery low voltage threshold [1500, 7500]
    int32_t dc_grid_dch_i_max; //Grid-tied DC discharge current maximum [0, 3200]
    int32_t ctrl_md; //Control mode, enum [0, 2]
    int32_t boot_md; //Boot-up mode, enum [0, 2]
    int32_t grid_rec_delay; //Grid reconnection delay, seconds [0, 600]
    int32_t grid_island; //Anti-islanding enable, enum [0, 1]
    int32_t fvrt_en; //Frequency and Voltage Ride-Through, enum [0, 1]
    int32_t react_w_md; //Reactive power control mode, enum [0, 3]
    int32_t rising_w_md; //Rising power control mode, enum [0, 1]
    int32_t act_w_md; //Active power control mode, enum [0, 3]
    int32_t dcdc_qty; //PV DC/DC Converter access quantity [0, 20]
} SCH_PcsSettings_t;

typedef struct __attribute__((packed)) sch_pcs_device_info
{
    char ip_addr[SCH_PCS_IP_ADDR_LEN];
    char netmask[SCH_PCS_NETMASK_LEN];
    char gw_ip_addr[SCH_PCS_GW_IP_ADDR_LEN];
    char mac_addr[SCH_PCS_MAC_ADDR_LEN];
    int32_t mb_addr; //Modbus address [1, 247]
    int32_t baud; //RTU (RS485) baud, enum [0, 2]
    int32_t ems_485_int_to; //EMS RS485 interface timeout, seconds [0, 600]
    int32_t ems_tcp_int_to; //EMS TCP interface timeout, seconds [0, 600]
    int32_t bms_int_to; //BMS interface timeout, seconds [0, 600]
    int32_t bms_int_prot; //BMS interface protocol, enum [0, 24]
    int32_t par_com_det; //Parallel PCS communication detection, enum [0, 1]
    int32_t pcs_id; //Parallel PCS communication ID [1, 10]
} SCH_PcsDeviceInfo_t;

typedef struct __attribute__((packed)) sch_pcs_component_data
{
    int32_t inv_a_v_rms; //Output voltage RMS, phase A, 0.1V
    int32_t inv_b_v_rms; //Output voltage RMS, phase B, 0.1V
    int32_t inv_c_v_rms; //Output voltage RMS, phase C, 0.1V
    int32_t inv_a_i_rms; //Output current RMS, phase A, 0.1A
    int32_t inv_b_i_rms; //Output current RMS, phase B, 0.1A
    int32_t inv_c_i_rms; //Output current RMS, phase C, 0.1A
    int32_t inv_freq; //Output Frequency, 0.01Hz
    int32_t inv_a_p; //Active power RMS, phase A, 0.01kW
    int32_t inv_b_p; //Active power RMS, phase A, 0.01kW
    int32_t inv_c_p; //Active power RMS, phase A, 0.01kW
    int32_t inv_a_q; //Active reacive power RMS, phase A, 0.01kVAR
    int32_t inv_b_q; //Active reacive power RMS, phase B, 0.01kVAR
    int32_t inv_c_q; //Active reacive power RMS, phase C, 0.01kVAR
    int32_t inv_a_s; //Apparent power, phase A, 0.01kVA
    int32_t inv_b_s; //Apparent power, phase B, 0.01kVA
    int32_t inv_c_s; //Apparent power, phase C, 0.01kVA
    int32_t inv_a_pf; //Power factor, phase A, 0.01
    int32_t inv_b_pf; //Power factor, phase B, 0.01
    int32_t inv_c_pf; //Power factor, phase C, 0.01
    int32_t inv_p; //Acive power (cumulative), 0.01kW
    int32_t inv_q; //Reactive power (cumulative), 0.01kVAR
    int32_t inv_s; //Apparent power (cumulative), 0.01kVA
    int32_t inv_pf; //Power factor (cumlative), 0.01
    int32_t inv_temp; //Temperature (AC Heatsink), celcius
    int32_t amb_temp; //Temperature (Ambient), celcius
    int32_t dc_bus_v; //DC bus voltage, 0.1V
    int32_t dc_p; //DC power, 0.01kW
    int32_t dc_v; //DC voltage, 0.1V
    int32_t dc_i; //DC current, 0.1A
    int32_t leak_i; //Leakage current, mA
    int32_t ac_dch_kwh_total; //Cumulative AC discharged energy, 0.1kWh
    int32_t ac_ch_kwh_total; //Cumulative AC  charged energy, 0.1kWh
    int32_t ac_dch_kwh_day; //Daily AC discharged energy, 0.1kWh
    int32_t ac_ch_kwh_day; //Daily AC charged energy, 0.1kWh
    int32_t dc_dch_kwh_total; //Cumulative DC discharged energy, 0.1kWh
    int32_t dc_ch_kwh_total; //Cumulative DC charged energy, 0.1kWh
    int32_t sys_time_y; //Year [2000, 2199]
    int32_t sys_time_m; //Month [1, 12]
    int32_t sys_time_d; //Day [1, 31]
    int32_t sys_time_hr; //Hour [0, 23]
    int32_t sys_time_min; //Minute [0, 59]
    int32_t sys_time_sec; //Second [0, 59]
} SCH_PcsComponentData_t;

typedef struct __attribute__((packed)) sch_pcs_codes_status
{
    bool S1_0; //Fault status
    bool S1_1; //Alert status
    bool S1_2; //On status
    bool S1_3; //On-grid mode
    bool S1_4; //Off-grid mode
    bool S1_5; //Derated
    bool S1_6; //Checking for grid connection
    bool S1_7; //Failed status
    bool S1_9; //Booting up
    bool S1_10; //Standy mode
    bool S1_11; //Grid recovery delay in progress
    bool S1_12; //Derating due to temperature
    bool S2_2; //Initializing DSP parameters
    bool S3_0; //DC charging
    bool S3_1; //DC discharging
    bool S3_2; //Battery fully charged
    bool S3_3; //Battery empty
    bool S3_4; //Constant voltage charging (topping off)
    bool S3_5; //Drooping
    bool S3_6; //AC glavanic break device closed
    bool S3_7; //DC glavanic break device closed
    bool S3_8; //AC module 1 online
    bool S3_9; //AC module 2 online
    bool S3_10;  //AC module 3 online
    bool S3_11; //AC module 4 online
    bool S3_12; //AC module 5 online
    bool S3_13; //AC module 6 online
    bool S3_14; //AC module 7 online
    bool S3_15; //AC module 8 online
    bool S3_16; //AC module 9 online
    bool S3_17; //AC module 10 online
    bool S3_18; //AC module 11 online
    bool S3_19; //AC module 12 online
    bool S3_20; //AC module 1 ON
    bool S3_21; //AC module 2 ON
    bool S3_22; //AC module 3 ON
    bool S3_23; //AC module 4 ON
    bool S3_24; //AC module 5 ON
    bool S3_25; //AC module 6 ON
    bool S3_26; //AC module 7 ON
    bool S3_27; //AC module 8 ON
    bool S3_28; //AC module 9 ON
    bool S3_29; //AC module 10 ON
    bool S3_30; //AC module 11 ON
    bool S3_31; //AC module 12 ON
    bool S3_32; //AC module 1 Alarm
    bool S3_33; //AC module 2 Alarm
    bool S3_34; //AC module 3 Alarm
    bool S3_35; //AC module 4 Alarm
    bool S3_36; //AC module 5 Alarm
    bool S3_37; //AC module 6 Alarm
    bool S3_38; //AC module 7 Alarm
    bool S3_39; //AC module 8 Alarm
    bool S3_40; //AC module 9 Alarm
    bool S3_41; //AC module 10 Alarm
    bool S3_42; //AC module 11 Alarm
    bool S3_43; //AC module 12 Alarm
    bool S3_44; //AC module 1 Fault
    bool S3_45; //AC module 2 Fault
    bool S3_46; //AC module 3 Fault
    bool S3_47; //AC module 4 Fault
    bool S3_48; //AC module 5 Fault
    bool S3_49; //AC module 6 Fault
    bool S3_50; //AC module 7 Fault
    bool S3_51; //AC module 8 Fault
    bool S3_52; //AC module 9 Fault
    bool S3_53; //AC module 10 Fault
    bool S3_54; //AC module 11 Fault
    bool S3_55; //AC module 12 Fault
    bool S3_56; //"Electrical Room" fan stop
    bool S3_57; //Transformer fan stop
} SCH_PcsCodes_Status_t;

typedef struct __attribute__((packed)) sch_pcs_codes_alarm
{
    bool A1_1; //Overload
    bool A1_4; //Battery low capacity
    bool A1_13; //DC bus under-voltage
    bool A1_14; //DC bus over-voltage
    bool A2_1; //DC bus voltage unbalanced
    bool A3_0; //Emergency power off (EPO) tripped
    bool A4_1; //Calibration parameter error
    bool A4_2; //Sampling zero error
    bool A4_3; //Flash error
    bool A4_4; //Derating due to internal overtemperature
    bool A4_5; //Derating due to ambient overtemperature
    bool A4_9; //Derating due to BMS
    bool A4_10; //Derating due to Voltage
    bool A4_11; //Derating due to Frequency
    bool A4_12; //Module Current limit alarm 1
    bool A4_13; //Module Current limit alarm 2
    bool A4_14; //Module Current limit alarm 3
    bool A4_15; //Module Current limit alarm 4
    bool A4_16; //Module Current limit alarm 5
    bool A4_17; //Module Current limit alarm 6
    bool A4_18; //Module Current limit alarm 7
} SCH_PcsCodes_Alarm_t;

typedef struct __attribute__((packed)) sch_pcs_codes_fault
{
    bool E1_2; //PCS communication fault
    bool F1_0; //DC bus soft start failure
    bool F2_0; //Output overload timeout
    bool F2_2; //DSP version error
    bool F2_3; //CPLD version error
    bool F2_4; //Hardware version error
    bool F2_7; //Fan failure
    bool F2_15; //Sampling conflict
    bool F2_18; //Neutral line disconnected
    bool F2_20; //Wrong connection
    bool F2_27; //BMS connection timeout
    bool F2_28; //EMS connection timeout
    bool F2_34; //ID replication/conflict
    bool F2_35; //Master conflict
    bool F2_36; //DSP parameter mismatch
    bool F2_37; //Monitor parameter mismatch
    bool F2_38; //DRM0 shutdown
    bool F2_39; //Module quantity error
    bool F2_40; //CAN A communication failure
    bool F2_41; //CAN B communication failure
    bool F2_42; //CAN C communication failure
    bool F2_43; //Synchronous signal failure
    bool F2_44; //Auxiliary board communication failure
    bool F2_45; //SPI communication failure
    bool F2_46; //Grid ammeter communication timeout
    bool F2_47; //Ambient over-temperature
    bool F2_48; //Insulation detection error 1
    bool F2_49; //Insulation detection error 2
    bool F2_50; //Module current error 1
    bool F2_51; //Module current error 2
    bool F2_52; //Module current error 3
    bool F2_53; //Module current error 4
    bool F2_54; //Module current error 5
    bool F2_55; //Module current error 6
    bool F2_56; //Module current error 7
    bool F2_57; //BMS shutdown
    bool F3_0; //DSP initialization error
    bool F3_9; //Auxiliary power 1 failure
    bool F3_10; //Auxiliary power 2 failure
    bool F3_11; //Auxiliary power 3 failure
    bool F3_12; //Over temperature 1
    bool F3_13; //Over temperature 2
    bool F4_1; //DC input over-voltage
    bool F4_2; //DC input under-voltage
    bool F4_3; //DC input overload
    bool F4_4; //DC input overload timeout
    bool F4_5; //DC input soft start failure
    bool F4_6; //DC input galvanic break device open circuit
    bool F4_7; //DC input galvanic break device short circuit
    bool F4_8; //DC input galvanic break device error
    bool F4_9; //AC bus over-voltage
    bool F4_10; //AC bus under-voltage
    bool F4_11; //AC bus over-frequency
    bool F4_12; //AC bus under-frequency
    bool F4_13; //AC bus phase reversed
    bool F4_14; //AC bus asymmetric voltage
    bool F4_15; //AC bus  voltage error
    bool F4_16; //AC bus phase lost
    bool F4_17; //Islanding protection
    bool F4_18; //Off grid voltage oscillation
    bool F4_19; //On/Off grid toggling error
    bool F4_20; //AC soft start failure
    bool F4_21; //AC glavanic break device open circuit
    bool F4_22; //AC glavanic break device short circuit
} SCH_PcsCodes_Fault_t;

typedef struct __attribute__((packed)) sch_pcs_codes
{
    SCH_PcsCodes_Status_t status;
    SCH_PcsCodes_Alarm_t alarm;
    SCH_PcsCodes_Fault_t fault;
} SCH_PcsCodes_t;

typedef struct __attribute__((packed)) sch_pcs_params
{
    SCH_PcsSettings_t settings;
    SCH_PcsDeviceInfo_t deviceInfo;
    SCH_PcsComponentData_t componentData;
    SCH_PcsCodes_t codes;
} SCH_PcsParameters_t;

/**
 * @brief PCS Settings X-Macro used for quickly writing operations
 *        for all parameters in the settings struct. This can be used
 *        to parse string literals as well.
 * @note Example:
 *          //Create a struct with all settings
 *          #define X_PCS(parameter) int32_t parameter;
 *          PCS_SETTINGS_XMACRO
 *          #undef X_PCS
*/
#define PCS_SETTINGS_XMACRO \
    X_PCS(off_grid_v_reg) \
    X_PCS(off_grid_f_reg) \
    X_PCS(lim_i_st_md) \
    X_PCS(grid_v) \
    X_PCS(grid_f) \
    X_PCS(bat_ch_v_end) \
    X_PCS(bat_ch_i_max) \
    X_PCS(bat_dch_i_max) \
    X_PCS(bat_dch_v_min) \
    X_PCS(bat_i_config_max) \
    X_PCS(dc_low_v) \
    X_PCS(dc_grid_dch_i_max) \
    X_PCS(ctrl_md) \
    X_PCS(boot_md) \
    X_PCS(grid_rec_delay) \
    X_PCS(grid_island) \
    X_PCS(fvrt_en) \
    X_PCS(react_w_md) \
    X_PCS(rising_w_md) \
    X_PCS(act_w_md) \
    X_PCS(dcdc_qty)

/**
 * @brief PCS Device Info X-Macro used for quickly writing operations
 *        for all parameters in the device info struct. This can be used
 *        to parse string literals as well.
 * @note Example:
 *          //Create a struct with all device info data
 *          #define X_PCS(parameter) int32_t parameter;
 *          PCS_DEVICE_INFO_XMACRO
 *          #undef X_PCS
 * @note All string parameters are not fixed length, so they are not included in the X-Macro
*/
#define PCS_DEVICE_INFO_XMACRO \
    X_PCS(mb_addr) \
    X_PCS(baud) \
    X_PCS(ems_485_int_to) \
    X_PCS(ems_tcp_int_to) \
    X_PCS(bms_int_to) \
    X_PCS(bms_int_prot) \
    X_PCS(par_com_det) \
    X_PCS(pcs_id)

/**
 * @brief PCS Component data X-Macro used for quickly writing operations
 *        for all parameters in the struct. This can be used
 *        to parse string literals as well.
 * @note Example:
 *          //Create a struct with all component data
 *          #define X_PCS(parameter) int32_t parameter;
 *          PCS_COMPONENT_DATA_XMACRO
 *          #undef X_PCS
*/
#define PCS_COMPONENT_DATA_XMACRO \
    X_PCS(inv_a_v_rms) \
    X_PCS(inv_b_v_rms) \
    X_PCS(inv_c_v_rms) \
    X_PCS(inv_a_i_rms) \
    X_PCS(inv_b_i_rms) \
    X_PCS(inv_c_i_rms) \
    X_PCS(inv_freq)  \
    X_PCS(inv_a_p) \
    X_PCS(inv_b_p) \
    X_PCS(inv_c_p) \
    X_PCS(inv_a_q) \
    X_PCS(inv_b_q) \
    X_PCS(inv_c_q) \
    X_PCS(inv_a_s) \
    X_PCS(inv_b_s) \
    X_PCS(inv_c_s) \
    X_PCS(inv_a_pf) \
    X_PCS(inv_b_pf) \
    X_PCS(inv_c_pf) \
    X_PCS(inv_p) \
    X_PCS(inv_q) \
    X_PCS(inv_s) \
    X_PCS(inv_pf) \
    X_PCS(inv_temp) \
    X_PCS(amb_temp) \
    X_PCS(dc_bus_v) \
    X_PCS(dc_p) \
    X_PCS(dc_v) \
    X_PCS(dc_i) \
    X_PCS(leak_i) \
    X_PCS(ac_dch_kwh_total) \
    X_PCS(ac_ch_kwh_total) \
    X_PCS(ac_dch_kwh_day) \
    X_PCS(ac_ch_kwh_day) \
    X_PCS(dc_dch_kwh_total) \
    X_PCS(dc_ch_kwh_total) \
    X_PCS(sys_time_y) \
    X_PCS(sys_time_m) \
    X_PCS(sys_time_d) \
    X_PCS(sys_time_hr) \
    X_PCS(sys_time_min) \
    X_PCS(sys_time_sec)

/**
 * @brief PCS codes (STATUS) X-Macro used for quickly writing operations
 *        for all parameters in the codes struct. This can be used
 *        to parse string literals as well.
 * @note Example:
 *          //Create a bitfield struct with all codes
 *          #define X_PCS(parameter) bool parameter: 1;
 *          PCS_CODES_STATUS_XMACRO
 *          #undef X_PCS
*/
#define PCS_CODES_STATUS_XMACRO \
    X_PCS(S1_0) \
    X_PCS(S1_1) \
    X_PCS(S1_2) \
    X_PCS(S1_3) \
    X_PCS(S1_4) \
    X_PCS(S1_5) \
    X_PCS(S1_6) \
    X_PCS(S1_7) \
    X_PCS(S1_9) \
    X_PCS(S1_10) \
    X_PCS(S1_11) \
    X_PCS(S1_12) \
    X_PCS(S2_2) \
    X_PCS(S3_0) \
    X_PCS(S3_1) \
    X_PCS(S3_2) \
    X_PCS(S3_3) \
    X_PCS(S3_4) \
    X_PCS(S3_5) \
    X_PCS(S3_6) \
    X_PCS(S3_7) \
    X_PCS(S3_8) \
    X_PCS(S3_9) \
    X_PCS(S3_10) \
    X_PCS(S3_11) \
    X_PCS(S3_12) \
    X_PCS(S3_13) \
    X_PCS(S3_14) \
    X_PCS(S3_15) \
    X_PCS(S3_16) \
    X_PCS(S3_17) \
    X_PCS(S3_18) \
    X_PCS(S3_19) \
    X_PCS(S3_20) \
    X_PCS(S3_21) \
    X_PCS(S3_22) \
    X_PCS(S3_23) \
    X_PCS(S3_24) \
    X_PCS(S3_25) \
    X_PCS(S3_26) \
    X_PCS(S3_27) \
    X_PCS(S3_28) \
    X_PCS(S3_29) \
    X_PCS(S3_30) \
    X_PCS(S3_31) \
    X_PCS(S3_32) \
    X_PCS(S3_33) \
    X_PCS(S3_34) \
    X_PCS(S3_35) \
    X_PCS(S3_36) \
    X_PCS(S3_37) \
    X_PCS(S3_38) \
    X_PCS(S3_39) \
    X_PCS(S3_40) \
    X_PCS(S3_41) \
    X_PCS(S3_42) \
    X_PCS(S3_43) \
    X_PCS(S3_44) \
    X_PCS(S3_45) \
    X_PCS(S3_46) \
    X_PCS(S3_47) \
    X_PCS(S3_48) \
    X_PCS(S3_49) \
    X_PCS(S3_50) \
    X_PCS(S3_51) \
    X_PCS(S3_52) \
    X_PCS(S3_53) \
    X_PCS(S3_54) \
    X_PCS(S3_55) \
    X_PCS(S3_56) \
    X_PCS(S3_57)

/**
 * @brief PCS codes (ALERT) X-Macro used for quickly writing operations
 *        for all parameters in the codes struct. This can be used
 *        to parse string literals as well.
 * @note Example:
 *          //Create a bitfield struct with all codes
 *          #define X_PCS(parameter) bool parameter: 1;
 *          PCS_CODES_ALARM_XMACRO
 *          #undef X_PCS
*/
#define PCS_CODES_ALARM_XMACRO \
    X_PCS(A1_1) \
    X_PCS(A1_4) \
    X_PCS(A1_13) \
    X_PCS(A1_14) \
    X_PCS(A2_1) \
    X_PCS(A3_0) \
    X_PCS(A4_1) \
    X_PCS(A4_2) \
    X_PCS(A4_3) \
    X_PCS(A4_4) \
    X_PCS(A4_5) \
    X_PCS(A4_9) \
    X_PCS(A4_10) \
    X_PCS(A4_11) \
    X_PCS(A4_12) \
    X_PCS(A4_13) \
    X_PCS(A4_14) \
    X_PCS(A4_15) \
    X_PCS(A4_16) \
    X_PCS(A4_17) \
    X_PCS(A4_18)

/**
 * @brief PCS codes (FAULT) X-Macro used for quickly writing operations
 *        for all parameters in the codes struct. This can be used
 *        to parse string literals as well.
 * @note Example:
 *          //Create a bitfield struct with all codes
 *          #define X_PCS(parameter) bool parameter: 1;
 *          PCS_CODES_FAULT_XMACRO
 *          #undef X_PCS
*/
#define PCS_CODES_FAULT_XMACRO \
    X_PCS(E1_2) \
    X_PCS(F1_0) \
    X_PCS(F2_0) \
    X_PCS(F2_2) \
    X_PCS(F2_3) \
    X_PCS(F2_4) \
    X_PCS(F2_7) \
    X_PCS(F2_15) \
    X_PCS(F2_18) \
    X_PCS(F2_20) \
    X_PCS(F2_27) \
    X_PCS(F2_28) \
    X_PCS(F2_34) \
    X_PCS(F2_35) \
    X_PCS(F2_36) \
    X_PCS(F2_37) \
    X_PCS(F2_38) \
    X_PCS(F2_39) \
    X_PCS(F2_40) \
    X_PCS(F2_41) \
    X_PCS(F2_42) \
    X_PCS(F2_43) \
    X_PCS(F2_44) \
    X_PCS(F2_45) \
    X_PCS(F2_46) \
    X_PCS(F2_47) \
    X_PCS(F2_48) \
    X_PCS(F2_49) \
    X_PCS(F2_50) \
    X_PCS(F2_51) \
    X_PCS(F2_52) \
    X_PCS(F2_53) \
    X_PCS(F2_54) \
    X_PCS(F2_55) \
    X_PCS(F2_56) \
    X_PCS(F2_57) \
    X_PCS(F3_0) \
    X_PCS(F3_9) \
    X_PCS(F3_10) \
    X_PCS(F3_11) \
    X_PCS(F3_12) \
    X_PCS(F3_13) \
    X_PCS(F4_1) \
    X_PCS(F4_2) \
    X_PCS(F4_3) \
    X_PCS(F4_4) \
    X_PCS(F4_5) \
    X_PCS(F4_6) \
    X_PCS(F4_7) \
    X_PCS(F4_8) \
    X_PCS(F4_9) \
    X_PCS(F4_10) \
    X_PCS(F4_11) \
    X_PCS(F4_12) \
    X_PCS(F4_13) \
    X_PCS(F4_14) \
    X_PCS(F4_15) \
    X_PCS(F4_16) \
    X_PCS(F4_17) \
    X_PCS(F4_18) \
    X_PCS(F4_19) \
    X_PCS(F4_20) \
    X_PCS(F4_21) \
    X_PCS(F4_22)

#ifdef __cplusplus
}
#endif

#endif /* LION_SCH_PCS_SCHEMA_H__ */