/**********************************************************************************************************************
 * @file            SCH_PduSchema.h                                                                                                                                                                                                                                                                                     
 *                                                                                                                   
 * @brief           This file declares all the data structures used for PDU communications.                         
 *                                                                                                                   
 * @author          Sharlotte Holmes                                                                                     
 * @date            05/07/2024                                                                                    
 * @copyright       2024 Lion Energy LLC.                                                                                      
 *                  All rights reserved           
 *                                               
 *********************************************************************************************************************/
/**
 * @addtogroup      SCH
 *                  Identifier Description for the EMS Component (SCH)ema module
 * 
 * @ref             SCH_PduSchema.h 
 *                  
 *********************************************************************************************************************/
#ifndef LION_SCH_PDU_SCHEMA_H__
#define LION_SCH_PDU_SCHEMA_H__
#define MODEL_SIZE 1
#define HW_V_SIZE 3
#define PDU_FW_V_SIZE 5
#define BMS_HW_V_SIZE 3
#define BMS_FW_V_SIZE 7

#include "COD_Compiler_Defines.h"

#ifdef __cplusplus
extern "C" {
#endif

/**********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

/**********************************************************************************************************************
* @cond                                              GLOBAL SCOPE                                             @endcond
*********************************************************************************************************************/
typedef struct __attribute__((packed)) sch_pdu_signals {
    //Power Data
    int32_t pdu_v; //System Voltage, 0.1V
    int32_t pdu_soc; //System SOC, %
    int32_t pdu_i; //System Current, 0.1A
    int32_t pdu_p; //PDU System Power, 0.001kW
    int32_t pdu_soh; //System SOH, %

    //System Info
    int32_t bms_qty; //Total BMS Module Quantity
    int32_t cell_qty; //Total Cell Quantity
    int32_t temp_qty; //Total Temperature Sensor Quantity
    int32_t sys_time_y; //Year [2000, 2099]
    int32_t sys_time_m; //Month [1, 12]
    int32_t sys_time_d; //Day [1, 31]
    int32_t sys_time_hr; //Hour [0, 23]
    int32_t sys_time_min; //Minute [0, 59]
    int32_t sys_time_sec; //Second [0, 59]

    //Summary Energy Data
    int32_t ch_cap_day; //Daily Charging Capacity, 0.1kWh
    int32_t dch_cap_day; //Daily Discharging Capacity, 0.1kWh
    int32_t ch_cap_total; //Total Charging Capacity, 0.1kWh
    int32_t dch_cap_total; //Total Discharging Capacity, 0.1kWh

    //Summary Voltage Data
    int32_t cell_v_avg; //Average Cell Voltage, mV
    int32_t cell_v_max; //Maximum Cell Voltage, mV
    int32_t cell_v_min; //Minimum Cell Voltage, mV
    int32_t cell_v_delta_max;
    int32_t v_max_cell_id; //Maximum Cell Voltage ID
    int32_t v_min_cell_id; //Minimum Cell ID
    int32_t v_max_bms_id; //Maximum BMS Voltage ID
    int32_t v_min_bms_id; //Minimum BMS Voltage ID

    //Summary Temperature Data
    int32_t cell_t_avg; //Average Temperature, 0.1°C
    int32_t cell_t_max; //Maximum Temperature, 0.1°C
    int32_t cell_t_min; //Minimum Temperature, 0.1°C
    int32_t cell_t_delta_max; //Maximum Cell Temperature Difference, 0.1°C
    int32_t t_max_cell_id; //Maximum Cell Temperature ID
    int32_t t_min_cell_id; ///Minimum Cell Temperature ID
    int32_t t_max_term_id; //Maximum Terminal Temperature ID
    int32_t t_min_term_id; //Minimum Terminal Temperature ID
    int32_t t_max_term_bms_id; //Maximum Terminal Temperature BMS ID
    int32_t t_min_term_bms_id; //Minimum Terminal Temperature BMS ID
    int32_t t_max_bms_id; //Maximum Temperature BMS ID
    int32_t t_min_bms_id; //Minimum Temperature BMS ID
    int32_t term_t_max; //Maximum Terminal Temperature, 0.1°C
    int32_t term_t_min; //Minimum Terminal Temperature, 0.1°C
    int32_t term_t_avg; //Average Terminal Temperature, 0.1°C
    int32_t term_t_delta_max; //Maximum Terminal Temperature Difference

    //Other Data
    int32_t ch_p_max_set; //Maximum Charging Power Setting, 0.1kW
    int32_t dch_p_max_set; //Maximum Discharging Power Setting, 0.1kW
    int32_t ch_i_max_set; //Maximum Charging Current Setting, 0.1A
    int32_t dch_i_max_set; //Maximum Discharge Current Setting, 0.1A
    int32_t insul_r_neg; //Insulation Resistance (R-), 1kΩ
    int32_t insul_r_pos; //Insulation Resistance (R+), 1kΩ
} SCH_PduSignals_t;

typedef struct __attribute__((packed)) sch_pdu_system_info {
    char model[MODEL_SIZE]; //Product Model Number/Name
    char hw_v[HW_V_SIZE]; //Hardware Version
    char pdu_fw_v[PDU_FW_V_SIZE]; //PDU Firmware Version
    char bms_hw_v[BMS_HW_V_SIZE]; //BMS Hardware Version
    char bms_fw_v[BMS_FW_V_SIZE]; //BMS Firmware Version
} SCH_PduSystemInfo_t;

typedef struct __attribute__((packed)) sch_pdu_codes_status {
    bool S1_0; //Fault Status
    bool S1_2; //On Status
    bool S1_9; //Booting Up
    bool S1_13; //Performing Self-Check
    bool S1_14; //Performing Fault Recovery
    bool S1_15; //Forced Control Test Mode
    bool S1_16; //Maintenance Mode
    bool S1_17; //Powering Down
    bool S1_18; //Powered Down
    bool S1_19; //Ready
    bool S5_0; //Normal Status
    bool S5_1; //Alert Status (Level 1)
    bool S5_2; //Alert Status (Level 2)
    bool S5_3; //Alert Status (Level 3)
    bool S6_0; //Open Circuit
    bool S6_1; //Charging
    bool S6_2; //Discharging
    bool S6_3; //Charging Disabled
    bool S6_4; //Discharging Disabled
    bool S6_5; //Charging and Discharging Disabled
} SCH_PduCodes_Status_t;

typedef struct __attribute__((packed)) sch_pdu_codes_alarm {
    bool B2_0; //Cell Over-Voltage (Level 1)
    bool B2_1; //Cell Under-Voltage (Level 1)
    bool B2_2; //Over-Temperature (Level 1)
    bool B2_4; //Over-Differential Temperature (Level 1)
    bool B2_5; //Over-Differential Voltage (Level 1)
    bool B2_10; //Low SOC (Level 1)
    bool B2_13; //Temperature Rise (Level 1)
    bool B2_14; //Discharge Over-Current (Level 1)
    bool B2_15; //Power Supply Under-Voltage (Level 1)
    bool B2_16; //Power Supply Over-Voltage (Level 1)
    bool B2_17; //PDU Under-Voltage (Level 1)
    bool B2_18; //PDU Over-Voltage (Level 1)
    bool B2_19; //Quick Charge Over-Current (Level 1)
    bool B2_20; //Terminal Over-Temperature (Level 1)
    bool B2_21; //Charge Over-Current (Level 1)
    bool B2_22; //Low Insulation (Level 1)
    bool B2_23; //High SOC (Level 1)
    bool B2_24; //Charging Over-Temperature (Level 1)
    bool B2_25; //Charging Under-Temperature (Level 1)
    bool B2_26; //Discharging Over-Temperature (Level 1)
    bool B2_27; //Discharging Under-Temperature (Level 1)
    bool B2_28; //SOC Over-Differential  (Level 1)
    bool B2_29; //BMS Over-Voltage (Level 1)
    bool B2_30; //BMS Under-Voltage  (Level 1)
    bool B3_0; //Cell Over-Voltage (Level 2)
    bool B3_1; //Cell Under-Voltage (Level 2)
    bool B3_2; //Over-Temperature (Level 2)
    bool B3_4; //Over-Differential Temperature (Level 2)
    bool B3_5; //Over-Differential Voltage (Level 2)
    bool B3_10; //Low SOC  (Level 2)
    bool B3_13; //Temperature Rise (Level 2)
    bool B3_14; //Discharge Over-Current (Level 2)
    bool B3_15; //Power Supply Under-Voltage (Level 2)
    bool B3_16; //Power Supply Over-Voltage (Level 2)
    bool B3_17; //PDU Under-Voltage (Level 2)
    bool B3_18; //PDU Over-Voltage (Level 2)
    bool B3_19; //Quick Charge Over-Current (Level 2)
    bool B3_20; //Terminal Over-Temperature (Level 2)
    bool B3_21; //Charge Over-Current (Level 2)
    bool B3_22; //Low Insulation (Level 2)
    bool B3_23; //High SOC (Level 2)
    bool B3_24; //Charging Over-Temperature (Level 2)
    bool B3_25; //Charging Under-Temperature (Level 2)
    bool B3_26; //Discharging Over-Temperature (Level 2)
    bool B3_27; //Discharging Under-Temperature (Level 2)
    bool B3_28; //SOC Over-Differential  (Level 2)
    bool B3_29; //BMS Over-Voltage (Level 2)
    bool B3_30; //BMS Under-Voltage  (Level 2)
} SCH_PduCodes_Alarm_t;

typedef struct __attribute__((packed)) sch_pdu_codes_fault {
    bool E1_3; //PDU Communication Fault
    bool B4_0; //Cell Over-Voltage (Level 3)
    bool B4_1; //Cell Under-Voltage (Level 3)
    bool B4_2; //Over-Temperature (Level 3)
    bool B4_4; //Cell Over-Differential Temperature (Level 3)
    bool B4_5; //Cell Over-Differential Voltage (Level 3)
    bool B4_10; //Low SOC  (Level 3)
    bool B4_13; //Temperature Rise (Level 3)
    bool B4_14; //Discharge Over-Current (Level 3)
    bool B4_15; //Power Supply Under-Voltage (Level 3)
    bool B4_16; //Power Supply Over-Voltage (Level 3)
    bool B4_17; //PDU Under-Voltage (Level 3)
    bool B4_18; //PDU Over-Voltage (Level 3)
    bool B4_19; //Quick Charge Over-Current (Level 3)
    bool B4_20; //Terminal Over-Temperature (Level 3)
    bool B4_21; //Charge Over-Current (Level 3)
    bool B4_22; //Low Insulation (Level 3)
    bool B4_23; //High SOC (Level 3)
    bool B4_24; //Charging Over-Temperature (Level 3)
    bool B4_25; //Charging Under-Temperature (Level 3)
    bool B4_26; //Discharging Over-Temperature (Level 3)
    bool B4_27; //Discharging Under-Temperature (Level 3)
    bool B4_28; //SOC Over-Differential  (Level 3)
    bool B4_29; //BMS Over-Voltage (Level 3)
    bool B4_30; //BMS Under-Voltage  (Level 3)
    bool B4_6; //Communication Fault
    bool B4_7; //Voltage Sense Failure
    bool B4_8; //Temperature Sense Failure
    bool B4_31; //Display Control Failure
    bool B4_32; //Module Over-Differential Voltage
    bool B4_33; //PDU Tripped
    bool B4_34; //BMS Fault Detected
    bool B4_35; //Battery Empty
    bool B4_36; //Inconsistency Fault
    bool B4_37; //Communication with PCS Failure
    bool B4_38; //CAN Hall Sensor Failure
    bool B4_39; //CAN Hall Sensor Communication Failure
    bool B4_40; //Hardware Self-Test Failed
    bool B4_41; //BMS[1] Communication Failure
    bool B4_42; //BMS[2] Communication Failure
    bool B4_43; //BMS[3] Communication Failure
    bool B4_44; //BMS[4] Communication Failure
    bool B4_45; //BMS[5] Communication Failure
    bool B4_46; //BMS[6] Communication Failure
    bool B4_47; //BMS[7] Communication Failure
    bool B4_48; //BMS[8] Communication Failure
    bool B4_49; //BMS[9] Communication Failure
    bool B4_50; //BMS[10] Communication Failure
    bool B4_51; //BMS[11] Communication Failure
    bool B4_52; //BMS[12] Communication Failure
    bool B4_53; //BMS[13] Communication Failure
    bool B4_54; //BMS[14] Communication Failure
    bool B4_55; //BMS[15] Communication Failure
    bool B4_56; //BMS[16] Communication Failure
    bool B4_57; //BMS[17] Communication Failure
    bool B4_58; //BMS[18] Communication Failure
    bool B4_59; //BMS[19] Communication Failure
    bool B4_60; //BMS[20] Communication Failure
    bool B4_61; //BMS[21] Communication Failure
    bool B4_62; //BMS[22] Communication Failure
    bool B4_63; //BMS[23] Communication Failure
    bool B4_64; //BMS[24] Communication Failure
    bool B4_65; //BMS[25] Communication Failure
    bool B4_66; //BMS[26] Communication Failure
    bool B4_67; //BMS[27] Communication Failure
    bool B4_68; //BMS[28] Communication Failure
    bool B4_69; //BMS[29] Communication Failure
    bool B4_70; //BMS[30] Communication Failure
    bool B4_71; //BMS[31] Communication Failure
    bool B4_72; //BMS[32] Communication Failure
    bool B4_73; //BMS[33] Communication Failure
    bool B4_74; //BMS[34] Communication Failure
    bool B4_75; //BMS[35] Communication Failure
    bool B4_76; //BMS[36] Communication Failure
    bool B4_77; //BMS[37] Communication Failure
    bool B4_78; //BMS[38] Communication Failure
    bool B4_79; //BMS[39] Communication Failure
    bool B4_80; //BMS[40] Communication Failure
    bool B4_81; //BMS[41] Communication Failure
    bool B4_82; //BMS[42] Communication Failure
    bool B4_83; //BMS[43] Communication Failure
    bool B4_84; //BMS[44] Communication Failure
    bool B4_85; //BMS[45] Communication Failure
    bool B4_86; //BMS[46] Communication Failure
    bool B4_87; //BMS[47] Communication Failure
    bool B4_88; //BMS[48] Communication Failure
    bool B4_89; //BMS[49] Communication Failure
    bool B4_90; //BMS[50] Communication Failure
    bool B4_91; //BMS[51] Communication Failure
    bool B4_92; //BMS[52] Communication Failure
    bool B4_93; //BMS[53] Communication Failure
    bool B4_94; //BMS[54] Communication Failure
    bool B4_95; //BMS[55] Communication Failure
    bool B4_96; //BMS[56] Communication Failure
} SCH_PduCodes_Fault_t;

typedef struct __attribute__((packed)) sch_pdu_codes {
    SCH_PduCodes_Status_t status;
    SCH_PduCodes_Alarm_t alarm;
    SCH_PduCodes_Fault_t fault;
} SCH_PduCodes_t;

typedef struct __attribute__((packed)) sch_pdu_params {
    SCH_PduSignals_t signals;
    SCH_PduSystemInfo_t systemInfo;
    SCH_PduCodes_t codes;
} SCH_PduParameters_t;

/**
 * @brief PDU Component data X-Macro used for quickly writing operations
 *        for all parameters in the struct. This can be used
 *        to parse string literals as well.
 * @note Example:
 *          //Create a struct with all component data
 *          #define X_PDU(parameter) int32_t parameter;
 *          PDU_SIGNALS_XMACRO
 *          #undef X_PDU
*/
#define PDU_SIGNALS_XMACRO \
    X_PDU(pdu_v) \
    X_PDU(pdu_soc) \
    X_PDU(pdu_i) \
    X_PDU(pdu_p) \
    X_PDU(pdu_soh) \
    X_PDU(bms_qty) \
    X_PDU(cell_qty) \
    X_PDU(temp_qty) \
    X_PDU(sys_time_y) \
    X_PDU(sys_time_m) \
    X_PDU(sys_time_d) \
    X_PDU(sys_time_hr) \
    X_PDU(sys_time_min) \
    X_PDU(sys_time_sec) \
    X_PDU(ch_cap_day) \
    X_PDU(dch_cap_day) \
    X_PDU(ch_cap_total) \
    X_PDU(dch_cap_total) \
    X_PDU(cell_v_avg) \
    X_PDU(cell_v_max) \
    X_PDU(cell_v_min) \
    X_PDU(cell_v_delta_max) \
    X_PDU(v_max_cell_id) \
    X_PDU(v_min_cell_id) \
    X_PDU(v_max_bms_id) \
    X_PDU(v_min_bms_id) \
    X_PDU(cell_t_avg) \
    X_PDU(cell_t_max) \
    X_PDU(cell_t_min) \
    X_PDU(cell_t_delta_max) \
    X_PDU(t_max_cell_id) \
    X_PDU(t_min_cell_id) \
    X_PDU(t_max_term_id) \
    X_PDU(t_min_term_id) \
    X_PDU(t_max_term_bms_id) \
    X_PDU(t_min_term_bms_id) \
    X_PDU(t_max_bms_id) \
    X_PDU(t_min_bms_id) \
    X_PDU(term_t_max) \
    X_PDU(term_t_min) \
    X_PDU(term_t_avg) \
    X_PDU(term_t_delta_max) \
    X_PDU(ch_p_max_set) \
    X_PDU(dch_p_max_set) \
    X_PDU(ch_i_max_set) \
    X_PDU(dch_i_max_set) \
    X_PDU(insul_r_neg) \
    X_PDU(insul_r_pos)

/**
 * @brief PDU SYSTEM INFO X-Macro used for quickly writing operations
 *        for all parameters in the codes struct. This can be used
 *        to parse string literals as well.
 * @note Example:
 *          //Create a bitfield struct with all codes
 *          #define X_PDU(parameter) char parameter: 1;
 *          PDU_CODES_SYSTEM_INFO_XMACRO
 *          #undef X_PDU
*/
#define PDU_SYSTEM_INFO_XMACRO \
    X_PDU(model) \
    X_PDU(hw_v) \
    X_PDU(pdu_fw_v) \
    X_PDU(bms_hw_v) \
    X_PDU(bms_fw_v)

/**
 * @brief PDU codes (STATUS) X-Macro used for quickly writing operations
 *        for all parameters in the codes struct. This can be used
 *        to parse string literals as well.
 * @note Example:
 *          //Create a bitfield struct with all codes
 *          #define X_PDU(parameter) bool parameter: 1;
 *          PDU_CODES_STATUS_XMACRO
 *          #undef X_PDU
*/
#define PDU_CODES_STATUS_XMACRO \
    X_PDU(S1_0) \
    X_PDU(S1_2) \
    X_PDU(S1_9) \
    X_PDU(S1_13) \
    X_PDU(S1_14) \
    X_PDU(S1_15) \
    X_PDU(S1_16) \
    X_PDU(S1_17) \
    X_PDU(S1_18) \
    X_PDU(S1_19) \
    X_PDU(S5_0) \
    X_PDU(S5_1) \
    X_PDU(S5_2) \
    X_PDU(S5_3) \
    X_PDU(S6_0) \
    X_PDU(S6_1) \
    X_PDU(S6_2) \
    X_PDU(S6_3) \
    X_PDU(S6_4) \
    X_PDU(S6_5)

/**
 * @brief PDU codes (ALARM) X-Macro used for quickly writing operations
 *        for all parameters in the codes struct. This can be used
 *        to parse string literals as well.
 * @note Example:
 *          //Create a bitfield struct with all codes
 *          #define X_PDU(parameter) bool parameter: 1;
 *          PDU_CODES_ALARM_XMACRO
 *          #undef X_PDU
*/
#define PDU_CODES_ALARM_XMACRO \
    X_PDU(B2_0) \
    X_PDU(B2_1) \
    X_PDU(B2_2) \
    X_PDU(B2_4) \
    X_PDU(B2_5) \
    X_PDU(B2_10) \
    X_PDU(B2_13) \
    X_PDU(B2_14) \
    X_PDU(B2_15) \
    X_PDU(B2_16) \
    X_PDU(B2_17) \
    X_PDU(B2_18) \
    X_PDU(B2_19) \
    X_PDU(B2_20) \
    X_PDU(B2_21) \
    X_PDU(B2_22) \
    X_PDU(B2_23) \
    X_PDU(B2_24) \
    X_PDU(B2_25) \
    X_PDU(B2_26) \
    X_PDU(B2_27) \
    X_PDU(B2_28) \
    X_PDU(B2_29) \
    X_PDU(B2_30) \
    X_PDU(B3_0) \
    X_PDU(B3_1) \
    X_PDU(B3_2) \
    X_PDU(B3_4) \
    X_PDU(B3_5) \
    X_PDU(B3_10) \
    X_PDU(B3_13) \
    X_PDU(B3_14) \
    X_PDU(B3_15) \
    X_PDU(B3_16) \
    X_PDU(B3_17) \
    X_PDU(B3_18) \
    X_PDU(B3_19) \
    X_PDU(B3_20) \
    X_PDU(B3_21) \
    X_PDU(B3_22) \
    X_PDU(B3_23) \
    X_PDU(B3_24) \
    X_PDU(B3_25) \
    X_PDU(B3_26) \
    X_PDU(B3_27) \
    X_PDU(B3_28) \
    X_PDU(B3_29) \
    X_PDU(B3_30)

/**
 * @brief PDU codes (FAULT) X-Macro used for quickly writing operations
 *        for all parameters in the codes struct. This can be used
 *        to parse string literals as well.
 * @note Example:
 *          //Create a bitfield struct with all codes
 *          #define X_PDU(parameter) bool parameter: 1;
 *          PDU_CODES_FAULT_XMACRO
 *          #undef X_PDU
*/
#define PDU_CODES_FAULT_XMACRO \
    X_PDU(E1_3) \
    X_PDU(B4_0) \
    X_PDU(B4_1) \
    X_PDU(B4_2) \
    X_PDU(B4_4) \
    X_PDU(B4_5) \
    X_PDU(B4_10) \
    X_PDU(B4_13) \
    X_PDU(B4_14) \
    X_PDU(B4_15) \
    X_PDU(B4_16) \
    X_PDU(B4_17) \
    X_PDU(B4_18) \
    X_PDU(B4_19) \
    X_PDU(B4_20) \
    X_PDU(B4_21) \
    X_PDU(B4_22) \
    X_PDU(B4_23) \
    X_PDU(B4_24) \
    X_PDU(B4_25) \
    X_PDU(B4_26) \
    X_PDU(B4_27) \
    X_PDU(B4_28) \
    X_PDU(B4_29) \
    X_PDU(B4_30) \
    X_PDU(B4_6) \
    X_PDU(B4_7) \
    X_PDU(B4_8) \
    X_PDU(B4_31) \
    X_PDU(B4_32) \
    X_PDU(B4_33) \
    X_PDU(B4_34) \
    X_PDU(B4_35) \
    X_PDU(B4_36) \
    X_PDU(B4_37) \
    X_PDU(B4_38) \
    X_PDU(B4_39) \
    X_PDU(B4_40) \
    X_PDU(B4_41) \
    X_PDU(B4_42) \
    X_PDU(B4_43) \
    X_PDU(B4_44) \
    X_PDU(B4_45) \
    X_PDU(B4_46) \
    X_PDU(B4_47) \
    X_PDU(B4_48) \
    X_PDU(B4_49) \
    X_PDU(B4_50) \
    X_PDU(B4_51) \
    X_PDU(B4_52) \
    X_PDU(B4_53) \
    X_PDU(B4_54) \
    X_PDU(B4_55) \
    X_PDU(B4_56) \
    X_PDU(B4_57) \
    X_PDU(B4_58) \
    X_PDU(B4_59) \
    X_PDU(B4_60) \
    X_PDU(B4_61) \
    X_PDU(B4_62) \
    X_PDU(B4_63) \
    X_PDU(B4_64) \
    X_PDU(B4_65) \
    X_PDU(B4_66) \
    X_PDU(B4_67) \
    X_PDU(B4_68) \
    X_PDU(B4_69) \
    X_PDU(B4_70) \
    X_PDU(B4_71) \
    X_PDU(B4_72) \
    X_PDU(B4_73) \
    X_PDU(B4_74) \
    X_PDU(B4_75) \
    X_PDU(B4_76) \
    X_PDU(B4_77) \
    X_PDU(B4_78) \
    X_PDU(B4_79) \
    X_PDU(B4_80) \
    X_PDU(B4_81) \
    X_PDU(B4_82) \
    X_PDU(B4_83) \
    X_PDU(B4_84) \
    X_PDU(B4_85) \
    X_PDU(B4_86) \
    X_PDU(B4_87) \
    X_PDU(B4_88) \
    X_PDU(B4_89) \
    X_PDU(B4_90) \
    X_PDU(B4_91) \
    X_PDU(B4_92) \
    X_PDU(B4_93) \
    X_PDU(B4_94) \
    X_PDU(B4_95) \
    X_PDU(B4_96)

#ifdef __cplusplus
	}
#endif
#endif