/**********************************************************************************************************************
 * @file            SCH_PmuSchema.h                                                                                                                                                                                                                                                                                     
 *                                                                                                                   
 * @brief           This file declares all the data structures used for PMU communications.                         
 *                                                                                                                   
 * @author          Sharlotte Holmes                                                                                     
 * @date            05/17/2024                                                                                    
 * @copyright       2024 Lion Energy LLC.                                                                                      
 *                  All rights reserved           
 *                                               
 *********************************************************************************************************************/
#ifndef LION_SCH_PMU_SCHEMA_H__
#define LION_SCH_PMU_SCHEMA_H__


#include "COD_Compiler_Defines.h"
#include "lion/sch/SCH_PmuSchema.h"

#ifdef __cplusplus
extern "C" {
#endif

/**********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

/**********************************************************************************************************************
* @cond                                              GLOBAL SCOPE                                             @endcond
*********************************************************************************************************************/
typedef struct __attribute__((packed)) sch_pmu_component_data {
    //Grid Power Data
    int32_t grid_a_v;  //Grid Voltage - Phase A, 0.1V
    int32_t grid_a_i;  //Grid Current - Phase A, 0.1A
    int32_t grid_a_p;  //Grid Active Power (P) - Phase A, 0.01kW
    int32_t grid_b_v;  //Grid Voltage - Phase B, 0.1V
    int32_t grid_b_i;  //Grid Current - Phase B, 0.1A
    int32_t grid_b_p;  //Grid Active Power (P) - Phase B, 0.01kW
    int32_t grid_c_v;  //Grid DVoltage - Phase C, 0.1V
    int32_t grid_c_i;  //Grid Current - Phase C, 0.1A
    int32_t grid_c_p;  //Grid Active Power (P) - Phase C, 0.01kW
    int32_t grid_p;  //Cumulative Grid Active Power (P), 0.01kW
    int32_t grid_a_s;  //Grid Apparent Power (S) - Phase A, 0.01kVa
    int32_t grid_b_s;  //Grid Apparent Power (S) - Phase B, 0.01kVa
    int32_t grid_c_s;  //Grid Apparent Power (S) - Phase C, 0.01kVa
    int32_t grid_s;  //Cumulative Grid Apparent Power (S), 0.01kVa
    int32_t grid_a_q;  //Grid Reactive Power (Q) - Phase A, 0.01kVAR
    int32_t grid_b_q;  //Grid Reactive Power (Q) - Phase B, 0.01kVAR
    int32_t grid_c_q;  //Grid Reactive Power (Q) - Phase C, 0.01kVAR
    int32_t grid_q;  //Cumulative Grid Reactive Power (Q), 0.01kVAR
    int32_t grid_v;  //Cumulative Grid Voltage, 0.1V
    int32_t grid_i;  //Cumulative Grid Current, 0.1A

    //Load Power Data
    int32_t load_a_v;  //Load Voltage - Phase A, 0.1V
    int32_t load_a_i;  //Load Current - Phase A, 0.1A
    int32_t load_a_p;  //Load Active Power (P) - Phase A, 0.01kW
    int32_t load_b_v;  //Load Voltage - Phase B, 0.1V
    int32_t load_b_i;  //Load Current - Phase B, 0.1A
    int32_t load_b_p;  //Load Active Power (P) - Phase B, 0.01kW
    int32_t load_c_v;  //Load DVoltage - Phase C, 0.1V
    int32_t load_c_i;  //Load Current - Phase C, 0.1A
    int32_t load_c_p;  //Load Active Power (P) - Phase C, 0.01kW
    int32_t load_p;  //Cumulative Load Active Power (P), 0.01kW
    int32_t load_a_s;  //Load Apparent Power (S) - Phase A, 0.01kVa
    int32_t load_b_s;  //Load Apparent Power (S) - Phase B, 0.01kVa
    int32_t load_c_s;  //Load Apparent Power (S) - Phase C, 0.01kVa
    int32_t load_s;  //Cumulative Load Apparent Power (S), 0.01kVa
    int32_t load_a_q;  //Load Reactive Power (Q) - Phase A, 0.01kVAR
    int32_t load_b_q;  //Load Reactive Power (Q) - Phase B, 0.01kVAR
    int32_t load_c_q;  //Load Reactive Power (Q) - Phase C, 0.01kVAR
    int32_t load_q;  //Cumulative Load Reactive Power (Q), 0.01kVAR
    int32_t load_v;  //Cumulative Load Voltage, 0.1V
    int32_t load_i;  //Cumulative Load Current, 0.1A

    //Power Factor
    int32_t grid_a_pf;  //Grid Power Factor - Phase A, 0.01
    int32_t grid_b_pf;  //Grid Power Factor - Phase B, 0.01
    int32_t grid_c_pf;  //Grid Power Factor - Phase C, 0.01
    int32_t grid_pf;  //Cumulative Grid Power Factor, 0.01
    int32_t load_a_pf;  //Load Power Factor - Phase A, 0.01
    int32_t load_b_pf;  //Load Power Factor - Phase B, 0.01
    int32_t load_c_pf;  //Load Power Factor - Phase C, 0.01
    int32_t load_pf;  //Cumulative Load Power Factor, 0.01
    int32_t min_pf;  //Minimum Power Factor, 0.01

} SCH_PmuComponentData_t;

typedef struct __attribute__((packed)) sch_pmu_codes_fault {
    bool E1_4;  //PMU Communication Fault
} SCH_PmuCodes_Fault_t;

typedef struct __attribute__((packed)) sch_pmu_codes {
    SCH_PmuCodes_Fault_t fault;
} SCH_PmuCodes_t;
typedef struct __attribute__((packed)) sch_pmu_params {
    SCH_PmuComponentData_t componentData;
    SCH_PmuCodes_t codes;
} SCH_PmuParameters_t;

/**
 * @brief PMU COMPONENT DATA X-Macro used for quickly writing operations
 *        for all parameters in the struct. This can be used
 *        to parse string literals as well.
 * @note Example:
 *          //Create a struct with all component data
 *          #define X_PMU(parameter) int32_t parameter;
 *          PMU_COMPONENT_DATA_XMACRO
 *          #undef X_PMU
*/
#define PMU_COMPONENT_DATA_XMACRO \
    X_PMU(grid_a_v) \
    X_PMU(grid_a_i) \
    X_PMU(grid_a_p) \
    X_PMU(grid_b_v) \
    X_PMU(grid_b_i) \
    X_PMU(grid_b_p) \
    X_PMU(grid_c_v) \
    X_PMU(grid_c_i) \
    X_PMU(grid_c_p) \
    X_PMU(grid_p) \
    X_PMU(grid_a_s) \
    X_PMU(grid_b_s) \
    X_PMU(grid_c_s) \
    X_PMU(grid_s) \
    X_PMU(grid_a_q) \
    X_PMU(grid_b_q) \
    X_PMU(grid_c_q) \
    X_PMU(grid_q) \
    X_PMU(grid_v) \
    X_PMU(grid_i) \
    X_PMU(load_a_v) \
    X_PMU(load_a_i) \
    X_PMU(load_a_p) \
    X_PMU(load_b_v) \
    X_PMU(load_b_i) \
    X_PMU(load_b_p) \
    X_PMU(load_c_v) \
    X_PMU(load_c_i) \
    X_PMU(load_c_p) \
    X_PMU(load_p) \
    X_PMU(load_a_s) \
    X_PMU(load_b_s) \
    X_PMU(load_c_s) \
    X_PMU(load_s) \
    X_PMU(load_a_q) \
    X_PMU(load_b_q) \
    X_PMU(load_c_q) \
    X_PMU(load_q) \
    X_PMU(load_v) \
    X_PMU(load_i) \
    X_PMU(grid_a_pf) \
    X_PMU(grid_b_pf) \
    X_PMU(grid_c_pf) \
    X_PMU(grid_pf) \
    X_PMU(load_a_pf) \
    X_PMU(load_b_pf) \
    X_PMU(load_c_pf) \
    X_PMU(load_pf) \
    X_PMU(min_pf) \

/**
 * @brief PMU CODES FAULT X-Macro used for quickly writing operations
 *        for all parameters in the codes struct. This can be used
 *        to parse string literals as well.
 * @note Example:
 *          //Create a bitfield struct with all codes
 *          #define X_PMU(parameter) char parameter: 1;
 *          PMU_CODES_FAULT_XMACRO
 *          #undef X_PMU
*/
#define PMU_CODES_FAULT_XMACRO \
    X_PMU(E1_4)

#ifdef __cplusplus
	}
#endif
#endif