/**********************************************************************************************************************
* @file            ICC_Client.h
*
* @brief           This service processes messages and events from the ICC.
*
* @author          Abram Early
* @date            06/15/2023
* @copyright       2023-2024 Lion Energy LLC.
*                  All rights reserved
*
*********************************************************************************************************************/
/**
* @addtogroup      ICC
*                  Identifier Description for the InterController Communication (ICC) Module  
* 
* @ref             ICC_Client.h 
*                  
*********************************************************************************************************************/
#include "COD_Compiler_Defines.h"

#ifndef ICC_CLIENT_H__
#define ICC_CLIENT_H__

/*********************************************************************************************************************
* @cond                                              GLOBAL SCOPE                                             @endcond
*********************************************************************************************************************/

#include "lion/icc/ICC.h"
#include "lion/icc/transport/ICC_Stream.h"
#include "lion/rtos/RTOS_CondVar.h"
#include <stdbool.h>

#ifdef __cplusplus
extern "C"
{
#endif

    typedef enum ICC_Client_State_t {
        ICC_CLIENT_DISABLED,
        ICC_CLIENT_DISCONNECTED,
        ICC_CLIENT_CONNECTED
    } ICC_Client_State_t;

    typedef struct ICC_Client_Context_t
    {
        pthread_t            fd_Thread;
        ICC_Context_t       *ptr_ICC;
        ICC_Stream_Context_t str_Stream;
        volatile bool        bol_Broken;
        volatile bool        bol_KeepRunning;
        const char          *sz_SocketPath;
        int                  fd_Socket;
        ICC_Client_State_t   enm_State;
        RTOS_Mutex_t         str_StateLock;
        RTOS_CondVar_t       str_StateChanged;

    } ICC_Client_Context_t;

    /**
     * @brief Initialize the ICC Client.
     * 
     * @param sz_SocketPath The path to the socket file.
     * 
     * @returns A pointer to the newly created ICC Context.
     */
    GlobScop ICC_Context_t *ICC_Client_Init(const char *sz_SocketPath);

    /**
     * @brief Starts the ICC Client.
     * 
     * @param tim_Timeout Timeout to wait for client to connect, RTOS_NO_WAIT to return immediately.
     * @returns 0 on success, -EALREADY if already running.
     */
    GlobScop int ICC_Client_Start(RTOS_Timeout_t tim_Timeout);

    /**
     * @brief Stops the ICC Client.
     * 
     * @returns 0 on success, -EALREADY if already stopped.
     */
    GlobScop int ICC_Client_Stop(void);

    /**
     * @brief Get the current state of the ICC Client.
     * 
     * @returns ICC_Client_State_t enum value.
     */
    GlobScop ICC_Client_State_t ICC_Client_GetState(void);

    int ICC_Client_WaitForState(ICC_Client_Context_t *ptr_Context, ICC_Client_State_t enm_State, RTOS_Timeout_t tim_Timeout);

    static inline int ICC_Client_WaitForConnected(ICC_Client_Context_t *ptr_Context, RTOS_Timeout_t tim_Timeout) {
        return ICC_Client_WaitForState(ptr_Context, ICC_CLIENT_CONNECTED, tim_Timeout);
    }

#ifdef __cplusplus
}
#endif

#endif /* ICC_CLIENT_H__ */