/**********************************************************************************************************************
* @file            IAL_Data.h
*
* @brief           Provides APIs for data related ICC functions and events for the CCU.
*                  
*
* @author          Devan Hone
* @date            07/25/2024
* @copyright       2024 Lion Energy LLC.
*                  All rights reserved
*
*********************************************************************************************************************/
/**
* @addtogroup      IAL
*                  Identifier Description for the ICC Abstraction Library (IAL) Module  
* 
* @ref             IAL.h 
*                  
*********************************************************************************************************************/
#include "COD_Compiler_Defines.h"

#ifndef LION_IAL_DATA_H__
#define LION_IAL_DATA_H__

/*********************************************************************************************************************
* @cond                                              GLOBAL SCOPE                                             @endcond
*********************************************************************************************************************/

#include <lion/ial/IAL.h>

#ifdef __cplusplus
extern "C"
{
#endif

#define IAL_DATA_u8_ALERT_BUFFER_SIZE   128
#define IAL_DATA_sz_TOU_KEY             "tou_prf"

/**
 * @brief  Retreieves data on the EMS from the Domain Controller Unit (DCU) using the Inter-Controller Communication
 *         (ICC) Service. Expects to only receive keys supported by the Data Source Type, and to only request for one
 *         data source or device per request. 
 * 
 * @param sz_UnitType       "pcs", "pdu", "dcdc", "pmu", "hvac", "gw"
 * @param u8_UnitIndex      (0-2), 255 (Null)
 * @param u8_NumberOfKeys   Size of the array of requested Keys
 * @param sz_arr_Keys       Requested keys as an array of strings
 * @param s32_arr_Responses Empty array of size (u8_NumberOfKeys) to be filled with retrieved values corresponding 
 *                          the sz_arr_Keys 
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
GlobScop int IAL_Data_GetQuantifiableData(const char *sz_UnitType, uint8_t u8_UnitIndex, uint8_t u8_NumberOfKeys, 
                                        const char **sz_arr_Keys, int32_t *s32_arr_Responses);

/**
 * @brief  Sets parameters of the EMS from the Domain Controller Unit (DCU) using the Inter-Controller Communication
 *         (ICC) Service. Expects to only receive keys supported by the Data Source Type, and to only request for one
 *         data source or device per request. 
 * 
 * @param sz_UnitType       "pcs", "pdu", "dcdc", "pmu", "hvac", "gw"
 * @param u8_UnitIndex      (0-2), 255 (Null)
 * @param u8_NumberOfKeys   Size of the array of requested Keys
 * @param sz_arr_Keys       Requested keys or settings to be written as an array of strings
 * @param s32_arr_Values    Requested values of settings to be written as an array of strings
 * @param int_arr_Statuses  Empty array of size (u8_NumberOfKeys) to be filled with the status to writing the requested
 *                          value to the device corresponding to the sz_arr_Keys (0 upon success)
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
GlobScop int IAL_Data_SetQuantifiableData(const char *sz_UnitType, uint8_t u8_UnitIndex, uint8_t u8_NumberOfKeys, 
                                        const char **sz_arr_Keys, int32_t *s32_arr_Values, int *int_arr_Statuses);

/**
 * @brief  Retreieves device information on the EMS from the Domain Controller Unit (DCU) using the Inter-Controller 
 *         Communication (ICC) Service. Expects to only receive keys supported by the Data Source Type, and to only 
 *         request for one data source or device per request. 
 * 
 * @param sz_UnitType       "pcs", "pdu", "dcdc", "pmu", "hvac", "gw"
 * @param u8_UnitIndex      (0-2), 255 (Null)
 * @param u8_NumberOfKeys   Size of the array of requested Keys
 * @param sz_arr_Keys       Requested keys as an array of strings
 * @param sz_arr_Responses  Empty array of size (u8_NumberOfKeys) to be filled with retrieved values corresponding 
 *                          the sz_arr_Keys 
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
GlobScop int IAL_Data_GetUnitInfo(const char* sz_UnitType, uint8_t u8_UnitIndex, uint8_t u8_NumberOfKeys,
                                const char** sz_arr_Keys, char** sz_arr_Responses);

/**
 * @brief  Retreieves Fault, Alarm, and Status "Code" data on the EMS from the Domain Controller Unit (DCU) using the 
 *         Inter-Controller Communication (ICC) Service. Expects to only receive keys supported by the Data Source Type,
 *         and to only request for one data source or device per request. 
 * 
 * @param sz_UnitType       "pcs", "pdu", "dcdc", "pmu", "hvac", "gw"
 * @param u8_UnitIndex      (0-2), 255 (Null)
 * @param sz_CodeType       "fault", "alarm", or "status"
 * @param sz_arr_Responses  Empty array of size (u8_NumberOfKeys) to be filled with retrieved values 
 *                          corresponding the sz_arr_Keys 
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
GlobScop int IAL_Data_GetUnitCodes(const char *sz_UnitType, uint8_t u8_UnitIndex, const char *sz_CodeType, 
                                    char **sz_arr_Responses);

/**
 * @brief Retrieves bufferred component data from the DCU (Domain Controller Unit) 
 *        using the Inter-Controller Communication (ICC) Service.
 * 
 * @param sz_UnitType Unit type data to retrieve ["pcs", "pdu", "dcdc", "pmu", "hvac", "gw", "ats"]
 * @param u8_UnitIndex Unit device index
 * @param vptr_Buff Destination buffer to store the data
 * @param u32ptr_BuffSize Size of the destination buffer. Also, when the function returns 
 *                        this parameter will be set to the size of the data retrieved.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
*/
GlobScop int IAL_Data_GetBufferedData(const char *sz_UnitType, uint8_t u8_UnitIndex, void *vptr_Buff, 
                                    uint32_t *u32ptr_BuffSize);

/**
 * @brief  Retreieves Configuration Settings on the EMS from the Domain Controller Unit (DCU) using the Inter-Controller
 *         Communication (ICC) Service.
 * 
 * @param sz_UnitType       Configuration Type ("pcs", "pdu", "dcdc", "pmu", "hvac", "gw")
 * @param u8_UnitIndex      Unit ID (when applicable, otherwise send COD_u8_NULL_UNIT_ID)
 * @param u8_NumberOfKeys   Size of the array of requested Keys
 * @param sz_arr_Keys       Requested keys as an array of strings
 * @param sz_arr_Responses  Empty array of size (u8_NumberOfKeys) to be filled with retrieved values corresponding 
 *                          the sz_arr_Keys 
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
GlobScop int IAL_Data_GetEMSConfiguration(const char *sz_UnitType, uint8_t u8_UnitIndex, uint8_t u8_NumberOfKeys, 
                                        const char** sz_arr_Keys, char ** sz_arr_Responses);

/**
 * @brief  Sets Configuration Settings of the EMS from the Domain Controller Unit (DCU) using the Inter-Controller
 *         Communication (ICC) Service. 
 * 
 * @param sz_UnitType       Configuration Type ("pcs", "pdu", "dcdc", "pmu", "hvac", "gw")
 * @param u8_UnitIndex      Unit ID (when applicable, otherwise send COD_u8_NULL_UNIT_ID)
 * @param u8_NumberOfKeys   Size of the array of requested Keys
 * @param sz_arr_Keys       Requested keys or settings to be written as an array of strings
 * @param sz_arr_Values     Requested values of settings to be written as an array of strings
 * @param int_arr_Statuses  Empty array of size (u8_NumberOfKeys) to be filled with the status to writing the requested
 *                          value to the device corresponding to the sz_arr_Keys (0 upon success)
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
GlobScop int IAL_Data_SetEMSConfiguration(const char *sz_UnitType, uint8_t u8_UnitIndex, uint8_t u8_NumberOfKeys, 
                                        const char **sz_arr_Keys, const char **sz_arr_Values, int *int_arr_Statuses);

/**
 * @brief  Retreieves Configuration Settings of the various Units/Components from the Domain Controller Unit (DCU)
 *         using the Inter-Controller Communication (ICC) Service.
 * 
 * @param sz_UnitType       Configuration Type ("pcs", "pdu", "dcdc", "pmu", "hvac", "gw")
 * @param u8_UnitIndex      Unit ID (when applicable, otherwise send COD_u8_NULL_UNIT_ID)
 * @param u8_NumberOfKeys   Size of the array of requested Keys
 * @param sz_arr_Keys       Requested keys as an array of strings
 * @param sz_arr_Responses  Empty array of size (u8_NumberOfKeys) to be filled with retrieved values corresponding 
 *                          the sz_arr_Keys 
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
GlobScop int IAL_Data_GetUnitConfiguration(const char *sz_UnitType, uint8_t u8_UnitIndex, uint8_t u8_NumberOfKeys, 
                                            const char** sz_arr_Keys, char ** sz_arr_Responses);

/**
 * @brief  Sets Configuration Settings of the various Units/Components from the Domain Controller Unit (DCU) using the 
 *         Inter-Controller Communication (ICC) Service. 
 * 
 * @param sz_UnitType       Configuration Type ("pcs", "pdu", "dcdc", "pmu", "hvac", "gw")
 * @param u8_UnitIndex      Unit ID (when applicable, otherwise send COD_u8_NULL_UNIT_ID)
 * @param u8_NumberOfKeys   Size of the array of requested Keys
 * @param sz_arr_Keys       Requested keys or settings to be written as an array of strings
 * @param sz_arr_Values     Requested values of settings to be written as an array of strings
 * @param int_arr_Statuses  Empty array of size (u8_NumberOfKeys) to be filled with the status to writing the requested
 *                          value to the device corresponding to the sz_arr_Keys (0 upon success)
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
GlobScop int IAL_Data_SetUnitConfiguration(const char *sz_UnitType, uint8_t u8_UnitIndex, uint8_t u8_NumberOfKeys, 
                                            const char **sz_arr_Keys, const char **sz_arr_Values, int *int_arr_Statuses);

/**
 * @brief  Retreieves Time of Use Settings on the EMS from the Domain Controller Unit (DCU) using the Inter-Controller 
 *          Communication (ICC) Service.
 * 
 * @param u8_ProfileID          Profile ID
 * @param sz_SettingPath        Name of the desired setting to get
 * @param u8_SettingDepth       Depth of Setting Path (0 indexed)
 * @param sz_Response           Retrieved setting value (either a value or an serialized JSON Array)
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
GlobScop int IAL_Data_GetTouSettings(uint8_t u8_ProfileID, const char **sz_SettingPath, uint8_t u8_SettingDepth,
                                    char *sz_Response);

/**
 * @brief  Sets Time of Use Settings of the EMS from the Domain Controller Unit (DCU) using the Inter-Controller
 *         Communication (ICC) Service. 
 * 
 * @param u8_ProfileID          Profile ID
 * @param sz_SerializedSettings Serailized JSON Array of settings to be configured
 * @param sz_ArrayName          Name of the array serialized in the previous argument
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
GlobScop int IAL_Data_SetTouSettings(uint8_t u8_ProfileID, const char *sz_SerializedSettings, const char *sz_ArrayName);

/**
 * @brief Test function used to validate the send alert status function via ICC commands
 * 
 * @param sz_UnitType Unit type
 * @param u8_UnitIndex Unit index
 *
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
GlobScop int IAL_Data_TestAlertStatusCallback(const char *sz_UnitType, uint8_t u8_UnitIndex);

#ifdef __cplusplus
}
#endif

#endif /* LION_IAL_DATA_H__ */
//EOF
