#include <lion/ial/ccu/IAL_Data.h>
#include <lion/log/log.h>
#include <lion/sch/SCH_AtsSchema.h>
#include <lion/sch/SCH_DcdcSchema.h>
#include <lion/sch/SCH_EmsSchema.h>
#include <lion/sch/SCH_GwSchema.h>
#include <lion/sch/SCH_HvacSchema.h>
#include <lion/sch/SCH_PcsSchema.h>
#include <lion/sch/SCH_PduSchema.h>
#include <lion/sch/SCH_PmuSchema.h>
#include <ArduinoJson.h>
#include <zcbor_decode.h>
#include <zcbor_encode.h>

LOG_MODULE_REGISTER(ial_data, LOG_LEVEL_INF);

/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

/**
 * @brief Gives the size of the buffer required to store data structs of the given unit type.
 * 
 * @param[in] sz_UnitType The unit type to get the size of the data struct for.
 * 
 * @return The size of the data struct for the given unit type.
*/
FileScop size_t ial_data_BufferedDataSize(const char *sz_UnitType);


/*********************************************************************************************************************
* @cond                                              MODULE SCOPE                                             @endcond
*********************************************************************************************************************/

ModScop extern ICC_Context_t *ial_ptr_IccCtx; //Defined in IAL.c

/*********************************************************************************************************************
* @cond                                              GLOBAL SCOPE                                             @endcond
*********************************************************************************************************************/

GlobScop int IAL_Data_GetQuantifiableData(const char *sz_UnitType, uint8_t u8_UnitIndex, uint8_t u8_NumberOfKeys, 
                                        const char **arr_sz_Keys, int32_t *arr_s32_Responses)
{
    // Build the Request
    uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBORState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 1);
    zcbor_tstr_encode_ptr(CBORState, sz_UnitType, strlen(sz_UnitType)+1);
    if (u8_UnitIndex != COD_u8_NULL_UNIT_ID) {
        zcbor_uint32_put(CBORState, (u8_UnitIndex));
    }
    zcbor_list_start_encode(CBORState, u8_NumberOfKeys);
    for (uint8_t i = 0; i < u8_NumberOfKeys; i++)
    {
        zcbor_tstr_encode_ptr(CBORState, arr_sz_Keys[i], strlen(arr_sz_Keys[i]) + 1);
    }
    zcbor_list_end_encode(CBORState, u8_NumberOfKeys);

    // Send the Request
    uint32_t u32_RequestLength = (CBORState[0].payload - u8_RequestBuffer);
    uint8_t arr_s32_EncodedResponses[2 + u8_NumberOfKeys * 5] = {0};
    uint32_t u32_ResponseLength = sizeof(arr_s32_EncodedResponses);
    int int_Rc = ICC_Function_Call(ial_ptr_IccCtx, IAL_GET_QUANTIFIABLE_DATA_ID, u8_RequestBuffer, u32_RequestLength, 
                                    arr_s32_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
    if (int_Rc < 0) {
        LOG_INF("ICC_Function_Call Error %d \"%s\", Attempting Retry", int_Rc, strerror(-int_Rc));
        // Try again
        u32_ResponseLength = sizeof(arr_s32_EncodedResponses);
        int_Rc = ICC_Function_Call(ial_ptr_IccCtx, IAL_GET_QUANTIFIABLE_DATA_ID, u8_RequestBuffer, u32_RequestLength, 
                                    arr_s32_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
        if (int_Rc < 0) {
            LOG_INF("ICC_Function_Call Retry Failed with Error %d \"%s\"", int_Rc, strerror(-int_Rc));
            return int_Rc;
        }
    }
    

    // Decode the response 
    ZCBOR_STATE_D(CBORDecodeState, 1, arr_s32_EncodedResponses, u32_ResponseLength, 3);
    if (zcbor_list_start_decode(CBORDecodeState) == 0)
    {
        LOG_INF("Decoding CBOR Error");
    }
    else 
    {
        for (uint32_t i = 0; i < u8_NumberOfKeys; i++)
        {
            zcbor_int32_decode(CBORDecodeState, &arr_s32_Responses[i]);
        }
    }
    zcbor_list_end_decode(CBORDecodeState);

    LOG_INF("Call for %s %d data: %d, %u bytes", sz_UnitType, u8_UnitIndex, int_Rc, u32_ResponseLength);
    return int_Rc;
}

GlobScop int IAL_Data_SetQuantifiableData(const char *sz_UnitType, uint8_t u8_UnitIndex, uint8_t u8_NumberOfKeys, 
                                        const char **arr_sz_Keys,int32_t *s32_arr_Values, int *arr_int_Statuses)
{
    // Build the Request
    uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBORState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 1);
    zcbor_tstr_encode_ptr(CBORState, sz_UnitType, strlen(sz_UnitType)+1);
    if (u8_UnitIndex != COD_u8_NULL_UNIT_ID) {
        zcbor_uint32_put(CBORState, (u8_UnitIndex));
    }
    const uint8_t u8_MAP_WIDTH = 2; // Keys | Values
    zcbor_map_start_encode(CBORState, u8_NumberOfKeys*u8_MAP_WIDTH);
    for (uint8_t i = 0; i < u8_NumberOfKeys; i++)
    {
        zcbor_tstr_encode_ptr(CBORState, arr_sz_Keys[i], strlen(arr_sz_Keys[i]) + 1);
        zcbor_int32_put(CBORState, s32_arr_Values[i]);
    }
    zcbor_map_end_encode(CBORState, u8_NumberOfKeys*u8_MAP_WIDTH);

    // Send the Request
    uint32_t u32_RequestLength = (CBORState[0].payload - u8_RequestBuffer);
    FileScop uint8_t u8_EncodedResponses[IAL_u16_ICC_RESPONSE_BUFFER_SIZE] = {0};
    uint32_t u32_ResponseLength = sizeof(u8_EncodedResponses);
    int int_Rc = ICC_Function_Call(ial_ptr_IccCtx, IAL_SET_QUANTIFIABLE_DATA_ID, u8_RequestBuffer, u32_RequestLength, 
                                    u8_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
    if (int_Rc < 0) {
        LOG_INF("ICC_Function_Call Error %d \"%s\"", int_Rc, strerror(-int_Rc));
        return int_Rc;
    }

    // Decode the response 
    ZCBOR_STATE_D(CBORDecodeState, 1, u8_EncodedResponses, u32_ResponseLength, 3);
    if (zcbor_list_start_decode(CBORDecodeState) == 0)
    {
        LOG_INF("Decoding CBOR Error");
    }
    else 
    {
        for (uint32_t i = 0; i < u8_NumberOfKeys; i++)
        {
            zcbor_int32_decode(CBORDecodeState, &arr_int_Statuses[i]);
        }
    }
    zcbor_list_end_decode(CBORDecodeState);

    LOG_INF("Call for %s %d data: %d, %u bytes", sz_UnitType, u8_UnitIndex, int_Rc, u32_ResponseLength);
    return int_Rc;
}

GlobScop int IAL_Data_GetUnitInfo(const char* sz_UnitType, uint8_t u8_UnitIndex, uint8_t u8_NumberOfKeys, 
                                const char** arr_sz_Keys, char** arr_sz_Responses)
{
    // Build the Request
    uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBORState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 1);
    zcbor_tstr_encode_ptr(CBORState, sz_UnitType, strlen(sz_UnitType)+1);
    if (u8_UnitIndex != COD_u8_NULL_UNIT_ID) {
        zcbor_uint32_put(CBORState, (u8_UnitIndex));
    }
    zcbor_list_start_encode(CBORState, u8_NumberOfKeys);
    for (uint8_t i = 0; i < u8_NumberOfKeys; i++)
    {
        zcbor_tstr_encode_ptr(CBORState, arr_sz_Keys[i], strlen(arr_sz_Keys[i]) + 1);
    }
    zcbor_list_end_encode(CBORState, u8_NumberOfKeys);
    
    // Send the Request
    uint32_t u32_RequestLength = (CBORState[0].payload - u8_RequestBuffer);
    uint8_t u8_EncodedResponses[IAL_u16_ICC_RESPONSE_BUFFER_SIZE] = {0};
    uint32_t u32_ResponseLength = sizeof(u8_EncodedResponses);
    int int_Rc = ICC_Function_Call(ial_ptr_IccCtx, IAL_GET_UNIT_DATA_ID, u8_RequestBuffer, u32_RequestLength, 
                                    u8_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
    if (int_Rc < 0) {
        LOG_INF("ICC_Function_Call Error %d \"%s\"", int_Rc, strerror(-int_Rc));
        return int_Rc;
    }

    // Decode the response 
    ZCBOR_STATE_D(CBORDecodeState, 1, u8_EncodedResponses, u32_ResponseLength, 3);
    if (zcbor_list_start_decode(CBORDecodeState) == 0)
    {
        LOG_INF("Decoding CBOR Error");
    }
    else 
    {
        struct zcbor_string sz_zcborKey; 
        for (uint32_t i = 0; zcbor_bstr_decode(CBORDecodeState, &sz_zcborKey); i++)
        {
            arr_sz_Responses[i] = (char *)sz_zcborKey.value;
        }
    }
    zcbor_list_end_decode(CBORDecodeState);

    LOG_INF("Call for %s %d data: %d, %u bytes", sz_UnitType, u8_UnitIndex, int_Rc, u32_ResponseLength);
    return int_Rc;
}

GlobScop int IAL_Data_GetUnitCodes(const char *sz_UnitType, uint8_t u8_UnitIndex, const char *sz_CodeType, 
                                    char **arr_sz_Responses)
{
    // Build the Request
    uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBORState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 1);
    zcbor_tstr_encode_ptr(CBORState, sz_UnitType, strlen(sz_UnitType)+1);
    if (u8_UnitIndex != COD_u8_NULL_UNIT_ID) {
        zcbor_uint32_put(CBORState, (u8_UnitIndex));
    }
    zcbor_tstr_encode_ptr(CBORState, sz_CodeType, strlen(sz_CodeType)+1);
    
    // Send the Request
    uint32_t u32_RequestLength = (CBORState[0].payload - u8_RequestBuffer);
    FileScop uint8_t arr_s32_EncodedResponses[IAL_u16_ICC_RESPONSE_BUFFER_SIZE] = {0};
    uint32_t u32_ResponseLength = sizeof(arr_s32_EncodedResponses);
    int int_Rc = ICC_Function_Call(ial_ptr_IccCtx, IAL_GET_UNIT_CODES_ID, u8_RequestBuffer, u32_RequestLength, 
                                    arr_s32_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
    
    if (int_Rc < 0) {
        LOG_INF("ICC_Function_Call Error %d \"%s\"", int_Rc, strerror(-int_Rc));
        return int_Rc;
    }

    // Decode the response 
    ZCBOR_STATE_D(CBORDecodeState, 1, arr_s32_EncodedResponses, u32_ResponseLength, 3);
    if (zcbor_list_start_decode(CBORDecodeState) == 0)
    {
        LOG_INF("Decoding EMS Codes CBOR Error");
    }
    else 
    {
        struct zcbor_string sz_zcborKey; 
        for (uint32_t i = 0; zcbor_tstr_decode(CBORDecodeState, &sz_zcborKey); i++)
        {
            arr_sz_Responses[i] = (char *)sz_zcborKey.value;
        }
        zcbor_list_end_decode(CBORDecodeState);
    }

    LOG_INF("Call for %s %d %s codes: %d, %u bytes", sz_UnitType, u8_UnitIndex, sz_CodeType, int_Rc, u32_ResponseLength);
    return int_Rc;
}

GlobScop int IAL_Data_GetBufferedData(const char *sz_UnitType, uint8_t u8_UnitIndex, void *vptr_Buff, 
                                    uint32_t *ptr_u32_BuffSize)
{
    //Ensure buffer can hold the data structure
    if (*ptr_u32_BuffSize < ial_data_BufferedDataSize(sz_UnitType)) {
        LOG_ERR("Destination buffer is too small for %s data", sz_UnitType);
        return -ENOMEM;
    }

    //Build the request
    uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBORState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 1);
    zcbor_tstr_encode_ptr(CBORState, sz_UnitType, strlen(sz_UnitType)+1);
    if (u8_UnitIndex != COD_u8_NULL_UNIT_ID) {
        zcbor_uint32_put(CBORState, (u8_UnitIndex));
    }
    
    //Send the Request
    uint32_t u32_RequestLength = (CBORState[0].payload - u8_RequestBuffer);
    int int_Rc = ICC_Function_Call(ial_ptr_IccCtx, IAL_GET_BUFFERED_DATA, u8_RequestBuffer, u32_RequestLength, 
                                    vptr_Buff, ptr_u32_BuffSize, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
    //No need to decode response as the data is just the serialized data structure for the unit.
    
    if (int_Rc < 0) {
        LOG_INF("ICC_Function_Call Error %d \"%s\"", int_Rc, strerror(-int_Rc));
        return int_Rc;
    }
    return 0;
}

GlobScop int IAL_Data_GetEMSConfiguration(const char *sz_UnitType, uint8_t u8_UnitIndex, uint8_t u8_NumberOfKeys, 
                                        const char** arr_sz_Keys, char ** arr_sz_Responses)
{
    // Build the Request
    uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBORState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 1);
    zcbor_tstr_encode_ptr(CBORState, sz_UnitType, strlen(sz_UnitType)+1);
    if (u8_UnitIndex != COD_u8_NULL_UNIT_ID) {
        zcbor_uint32_put(CBORState, (u8_UnitIndex));     
    }
    zcbor_list_start_encode(CBORState, u8_NumberOfKeys);
    for (uint8_t i = 0; i < u8_NumberOfKeys; i++)
    {
        zcbor_tstr_encode_ptr(CBORState, arr_sz_Keys[i], strlen(arr_sz_Keys[i]) + 1);
    }
    zcbor_list_end_encode(CBORState, u8_NumberOfKeys);

    // Determine Function Code based on requested Unit Type (EMS or PMU)
    uint32_t u32_FunctionCode;
    if (strcmp(sz_UnitType, IAL_sz_UNIT_TYPE_PMU) == 0)
    {
        u32_FunctionCode = IAL_GET_PMU_CONFIG_ID;
    }
    else 
    {
        u32_FunctionCode = IAL_GET_EMS_CONFIG_ID;
    }

    // Send the Request
    uint32_t u32_RequestLength = (CBORState[0].payload - u8_RequestBuffer);
    uint8_t u8_EncodedResponses[2 + u8_NumberOfKeys * 17] = {0};
    uint32_t u32_ResponseLength = sizeof(u8_EncodedResponses);
    int int_Rc = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionCode, u8_RequestBuffer, u32_RequestLength, 
                                    u8_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CONFIG_CALL_TIMEOUT_S));
    if (int_Rc < 0) {
        LOG_INF("ICC_Function_Call Error %d \"%s\", Attempting Retry", int_Rc, strerror(-int_Rc));
        // Try again
        u32_ResponseLength = sizeof(u8_EncodedResponses);
        int_Rc = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionCode, u8_RequestBuffer, u32_RequestLength, 
                                    u8_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CONFIG_CALL_TIMEOUT_S));
        if (int_Rc < 0) {
            LOG_INF("ICC_Function_Call Retry Failed with Error %d \"%s\"", int_Rc, strerror(-int_Rc));
            return int_Rc;
        }
    }

    // Decode the response 
    ZCBOR_STATE_D(CBORDecodeState, 1, u8_EncodedResponses, u32_ResponseLength, 3);
    if (zcbor_list_start_decode(CBORDecodeState) == 0)
    {
        LOG_INF("Decoding CBOR Error");
    }
    else 
    {
        struct zcbor_string sz_zcborKey; 
        for (uint32_t i = 0; zcbor_tstr_decode(CBORDecodeState, &sz_zcborKey); i++)
        {
            arr_sz_Responses[i] = (char *)sz_zcborKey.value;
        }
        zcbor_list_end_decode(CBORDecodeState);
    }
    zcbor_list_end_decode(CBORDecodeState);

    LOG_INF("Call for %s %d data: %d, %u bytes", sz_UnitType, u8_UnitIndex, int_Rc, u32_ResponseLength);
    return int_Rc;
}

GlobScop int IAL_Data_SetEMSConfiguration(const char *sz_UnitType, uint8_t u8_UnitIndex, uint8_t u8_NumberOfKeys, 
                                        const char **arr_sz_Keys, const char **arr_sz_Values, int *arr_int_Statuses) 
{
    // Build the Request
    uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBORState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 1);
    zcbor_tstr_encode_ptr(CBORState, sz_UnitType, strlen(sz_UnitType)+1);
    if (u8_UnitIndex != COD_u8_NULL_UNIT_ID) {
        zcbor_uint32_put(CBORState, (u8_UnitIndex));
    }
    const uint8_t u8_MAP_WIDTH = 2; // Keys | Values
    zcbor_map_start_encode(CBORState, u8_NumberOfKeys*u8_MAP_WIDTH);
    for (uint8_t i = 0; i < u8_NumberOfKeys; i++)
    {
        zcbor_tstr_encode_ptr(CBORState, arr_sz_Keys[i], strlen(arr_sz_Keys[i]) + 1);
        zcbor_tstr_encode_ptr(CBORState, arr_sz_Values[i], strlen(arr_sz_Values[i]) + 1);
    }
    zcbor_map_end_encode(CBORState, u8_NumberOfKeys*u8_MAP_WIDTH);

    // Determine Function Code based on requested Unit Type (EMS or PMU)
    uint32_t u32_FunctionCode;
    if (strcmp(sz_UnitType, IAL_sz_UNIT_TYPE_PMU) == 0)
    {
        u32_FunctionCode = IAL_SET_PMU_CONFIG_ID;
    }
    else 
    {
        u32_FunctionCode = IAL_SET_EMS_CONFIG_ID;
    }

    // Send the Request
    uint32_t u32_RequestLength = (CBORState[0].payload - u8_RequestBuffer);
    FileScop uint8_t arr_s32_EncodedResponses[IAL_u16_ICC_RESPONSE_BUFFER_SIZE] = {0};
    uint32_t u32_ResponseLength = sizeof(arr_s32_EncodedResponses);
    int int_Rc = ICC_Function_Call(ial_ptr_IccCtx, u32_FunctionCode, u8_RequestBuffer, u32_RequestLength, 
                                    arr_s32_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CONFIG_CALL_TIMEOUT_S));
    if (int_Rc < 0) {
        LOG_INF("ICC_Function_Call Error %d \"%s\"", int_Rc, strerror(-int_Rc));
        return int_Rc;
    }

    // Decode the response 
    ZCBOR_STATE_D(CBORDecodeState, 1, arr_s32_EncodedResponses, u32_ResponseLength, 3);
    if (zcbor_list_start_decode(CBORDecodeState) == 0)
    {
        LOG_INF("Decoding CBOR Error");
    }
    else 
    {
        for (uint32_t i = 0; i < u8_NumberOfKeys; i++)
        {
            zcbor_int32_decode(CBORDecodeState, &arr_int_Statuses[i]);
        }
    }
    zcbor_list_end_decode(CBORDecodeState);

    LOG_INF("Call for %s %d data: %d, %u bytes", sz_UnitType, u8_UnitIndex, int_Rc, u32_ResponseLength);
    return int_Rc;
}

GlobScop int IAL_Data_GetUnitConfiguration(const char *sz_UnitType, uint8_t u8_UnitIndex, uint8_t u8_NumberOfKeys, 
                                            const char** arr_sz_Keys, char ** arr_sz_Responses)
{
    // Build the Request
    uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBORState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 1);
    zcbor_tstr_encode_ptr(CBORState, sz_UnitType, strlen(sz_UnitType)+1);
    if (u8_UnitIndex != COD_u8_NULL_UNIT_ID) {
        zcbor_uint32_put(CBORState, (u8_UnitIndex));     
    }
    zcbor_list_start_encode(CBORState, u8_NumberOfKeys);
    for (uint8_t i = 0; i < u8_NumberOfKeys; i++)
    {
        zcbor_tstr_encode_ptr(CBORState, arr_sz_Keys[i], strlen(arr_sz_Keys[i]) + 1);
    }
    zcbor_list_end_encode(CBORState, u8_NumberOfKeys);

    // Send the Request
    uint32_t u32_RequestLength = (CBORState[0].payload - u8_RequestBuffer);
    uint8_t u8_EncodedResponses[IAL_u16_ICC_RESPONSE_BUFFER_SIZE] = {0};
    uint32_t u32_ResponseLength = sizeof(u8_EncodedResponses);
    int int_Rc = ICC_Function_Call(ial_ptr_IccCtx, IAL_GET_UNIT_CONFIG_ID, u8_RequestBuffer, u32_RequestLength, 
                                    u8_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
    if (int_Rc < 0) {
        LOG_INF("ICC_Function_Call Error %d \"%s\"", int_Rc, strerror(-int_Rc));
        return int_Rc;
    }

    // Decode the response 
    ZCBOR_STATE_D(CBORDecodeState, 1, u8_EncodedResponses, u32_ResponseLength, 3);
    if (zcbor_list_start_decode(CBORDecodeState) == 0)
    {
        LOG_INF("Decoding CBOR Error");
    }
    else 
    {
        struct zcbor_string sz_zcborKey; 
        for (uint32_t i = 0; zcbor_tstr_decode(CBORDecodeState, &sz_zcborKey); i++)
        {
            arr_sz_Responses[i] = (char *)sz_zcborKey.value;
        }
        zcbor_list_end_decode(CBORDecodeState);
    }
    zcbor_list_end_decode(CBORDecodeState);

    LOG_INF("Call for %s %d data: %d, %u bytes", sz_UnitType, u8_UnitIndex, int_Rc, u32_ResponseLength);
    return int_Rc;
}

GlobScop int IAL_Data_SetUnitConfiguration(const char *sz_UnitType, uint8_t u8_UnitIndex, uint8_t u8_NumberOfKeys, 
                                            const char **arr_sz_Keys, const char **arr_sz_Values, int *arr_int_Statuses)
{
    // Build the Request
    uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBORState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 1);
    zcbor_tstr_encode_ptr(CBORState, sz_UnitType, strlen(sz_UnitType)+1);
    if (u8_UnitIndex != COD_u8_NULL_UNIT_ID) {
        zcbor_uint32_put(CBORState, (u8_UnitIndex));
    }
    const uint8_t u8_MAP_WIDTH = 2; // Keys | Values
    zcbor_map_start_encode(CBORState, u8_NumberOfKeys*u8_MAP_WIDTH);
    for (uint8_t i = 0; i < u8_NumberOfKeys; i++)
    {
        zcbor_tstr_encode_ptr(CBORState, arr_sz_Keys[i], strlen(arr_sz_Keys[i]) + 1);
        zcbor_tstr_encode_ptr(CBORState, arr_sz_Values[i], strlen(arr_sz_Values[i]) + 1);
    }
    zcbor_map_end_encode(CBORState, u8_NumberOfKeys*u8_MAP_WIDTH);

    // Send the Request
    uint32_t u32_RequestLength = (CBORState[0].payload - u8_RequestBuffer);
    FileScop uint8_t arr_s32_EncodedResponses[IAL_u16_ICC_RESPONSE_BUFFER_SIZE] = {0};
    uint32_t u32_ResponseLength = sizeof(arr_s32_EncodedResponses);
    int int_Rc = ICC_Function_Call(ial_ptr_IccCtx, IAL_SET_UNIT_CONFIG_ID, u8_RequestBuffer, u32_RequestLength, 
                                    arr_s32_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
    if (int_Rc < 0) {
        LOG_INF("ICC_Function_Call Error %d \"%s\"", int_Rc, strerror(-int_Rc));
        return int_Rc;
    }

    // Decode the response 
    ZCBOR_STATE_D(CBORDecodeState, 1, arr_s32_EncodedResponses, u32_ResponseLength, 3);
    if (zcbor_list_start_decode(CBORDecodeState) == 0)
    {
        LOG_INF("Decoding CBOR Error");
    }
    else 
    {
        for (uint32_t i = 0; i < u8_NumberOfKeys; i++)
        {
            zcbor_int32_decode(CBORDecodeState, &arr_int_Statuses[i]);
        }
    }
    zcbor_list_end_decode(CBORDecodeState);

    LOG_INF("Call for %s %d data: %d, %u bytes", sz_UnitType, u8_UnitIndex, int_Rc, u32_ResponseLength);
    return int_Rc;
}

GlobScop int IAL_Data_GetTouSettings(uint8_t u8_ProfileID, const char **sz_SettingPath, uint8_t u8_SettingDepth, 
                                    char *sz_Response)
{
    const char *sz_PathPrefix = IAL_DATA_sz_TOU_KEY;
    // Build the Request
    uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBORState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 1);
    zcbor_list_start_encode(CBORState, 0);                                              // Start List
    zcbor_tstr_encode_ptr(CBORState, sz_PathPrefix, strlen(sz_PathPrefix)+1);               // Path Prefix
    if (u8_ProfileID != COD_u8_NULL_UNIT_ID)
    {
        zcbor_uint32_put(CBORState, u8_ProfileID);                                          // Profile ID
    }
    if (u8_SettingDepth > 0) 
    {
        for (uint8_t i = 0; i < u8_SettingDepth; i++)
        {
            zcbor_tstr_encode_ptr(CBORState, sz_SettingPath[i], strlen(sz_SettingPath[i])); // Setting/Extended Path Name
        }
        
    }
    zcbor_list_end_encode(CBORState, 0);                                                // End List
    zcbor_uint32_put(CBORState, 0);

    // Send the Request
    uint32_t u32_RequestLength = (CBORState[0].payload - u8_RequestBuffer);
    uint8_t u8_EncodedResponses[IAL_u16_ICC_RESPONSE_BUFFER_SIZE] = {0};
    uint32_t u32_ResponseLength = sizeof(u8_EncodedResponses);
    int int_Rc = ICC_Function_Call(ial_ptr_IccCtx, IAL_GET_TOU_ID, u8_RequestBuffer, u32_RequestLength, 
                                    u8_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
    if (int_Rc < 0) 
    {
        LOG_INF("ICC_Function_Call Error %d \"%s\"", int_Rc, strerror(-int_Rc));
        u32_ResponseLength = sizeof(u8_EncodedResponses);
        int_Rc = ICC_Function_Call(ial_ptr_IccCtx, IAL_GET_TOU_ID, u8_RequestBuffer, u32_RequestLength, 
                                    u8_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
        if (int_Rc < 0) 
        {
            LOG_INF("ICC_Function_Call Retry Error %d \"%s\"", int_Rc, strerror(-int_Rc));
            return int_Rc;
        }
    }

    // Decode the response 
    ZCBOR_STATE_D(CBORDecodeState, 1, u8_EncodedResponses, u32_ResponseLength, 3);
    if (zcbor_map_start_decode(CBORDecodeState) == 0)
    {
        LOG_INF("Decoding CBOR Error");
    }
    else 
    {
        ArduinoJson::StaticJsonDocument<IAL_u16_ICC_RESPONSE_BUFFER_SIZE> jd_response;
        struct zcbor_string sz_zcborKey; 
        for (uint32_t i = 0; zcbor_tstr_decode(CBORDecodeState, &sz_zcborKey); i++)
        {
            
            // Store those key value pairs in the Response JSON
            const char *sz_key = (char *)sz_zcborKey.value;
            struct zcbor_string sz_zcborValue;
            zcbor_tstr_decode(CBORDecodeState, &sz_zcborValue);
            const char *sz_value = (char *)sz_zcborValue.value;
            jd_response[sz_key] = sz_value;
        }
        zcbor_map_end_decode(CBORDecodeState);
        char sz_buffer[1024];
        ArduinoJson::serializeJson(jd_response, sz_buffer);
        strcpy(sz_Response, sz_buffer);
        LOG_INF("JSON Response: %s", sz_Response);
    }

    LOG_INF("Call for ToU Get: %s %d data: %d, %u bytes", sz_PathPrefix, u8_ProfileID, int_Rc, u32_ResponseLength);
    return int_Rc;
}

GlobScop int IAL_Data_SetTouSettings(uint8_t u8_ProfileID, const char *sz_SerializedSettings, const char *sz_ArrayName)
{
    const char *sz_PathPrefix = IAL_DATA_sz_TOU_KEY;
    const uint8_t u8_MAP_WIDTH = 2; // Keys | Values
    // Convert the serialized array of settings to a Json Doc, then extract the array
    ArduinoJson::StaticJsonDocument<IAL_u16_ICC_RESPONSE_BUFFER_SIZE> jd_SettingsDoc;
    ArduinoJson::deserializeJson(jd_SettingsDoc, sz_SerializedSettings);
    if (!jd_SettingsDoc.containsKey(sz_ArrayName))
    {
        LOG_ERR("Failed to set ToU settings due to invalid arrayName: %s", sz_ArrayName);
        return EINVAL;
    }
    ArduinoJson::JsonArray ja_Settings = jd_SettingsDoc[sz_ArrayName];
    // Start building the Request
    uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBORState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 1);
    zcbor_list_start_encode(CBORState, 0);                                      // Start List
    zcbor_tstr_encode_ptr(CBORState, sz_PathPrefix, strlen(sz_PathPrefix));         // Path Prefix
    if (u8_ProfileID != COD_u8_NULL_UNIT_ID)
    {
        zcbor_uint32_put(CBORState, u8_ProfileID);                                  // Profile ID
    }
    zcbor_list_end_encode(CBORState, 0);                                        // End List
    if (!ja_Settings.isNull())
    {
        // Build the Request
        zcbor_map_start_encode(CBORState, 0);                                   // Start Map
        zcbor_tstr_encode_ptr(CBORState, sz_ArrayName, strlen(sz_ArrayName));       // Array/List Name
        zcbor_list_start_encode(CBORState, 0);                                      // Start List
        for (uint8_t i = 0; i < ja_Settings.size(); i++)    // Iterate through List
        {
            ArduinoJson::JsonObject jo_arrayElement = ja_Settings[i];
            zcbor_map_start_encode(CBORState, 0);                                       // Start Map
            for (ArduinoJson::JsonPair keyValue : jo_arrayElement)  // Iterate Through Map
            {  
                const char *sz_key = keyValue.key().c_str();
                ArduinoJson::JsonVariant jv_value = keyValue.value();
                const char *sz_value = jv_value.as<const char*>();
                zcbor_tstr_encode_ptr(CBORState, sz_key, strlen(sz_key));                   // Key
                zcbor_tstr_encode_ptr(CBORState, sz_value, strlen(sz_value));               // Value
            }
            zcbor_map_end_encode(CBORState, 0);                                         // End Map
        }
        zcbor_list_end_encode(CBORState, 0);                                        // End List
        zcbor_map_end_encode(CBORState, 0);                                     // End Map
    }
    else // Non Array type requests
    {
         zcbor_map_start_encode(CBORState, 0);                                   // Start Map
        for (ArduinoJson::JsonPair jp_keyValue : jd_SettingsDoc.as<ArduinoJson::JsonObject>()) 
        {  
            const char *sz_key = jp_keyValue.key().c_str();
            ArduinoJson::JsonVariant jv_value = jp_keyValue.value();
            char sz_value[64];
            sprintf(sz_value, "%i", jv_value.as<int32_t>());
            zcbor_tstr_encode_ptr(CBORState, sz_key, strlen(sz_key));
            zcbor_tstr_encode_ptr(CBORState, sz_value, strlen(sz_value));
        }
        zcbor_map_end_encode(CBORState, 0);                                     // End Map
    }
    // Send the Request
    uint32_t u32_RequestLength = (CBORState[0].payload - u8_RequestBuffer);
    FileScop uint8_t arr_s32_EncodedResponses[IAL_u16_ICC_RESPONSE_BUFFER_SIZE] = {0};
    uint32_t u32_ResponseLength = sizeof(arr_s32_EncodedResponses);
    int int_Rc = ICC_Function_Call(ial_ptr_IccCtx, IAL_SET_TOU_ID, u8_RequestBuffer, u32_RequestLength, 
                                    arr_s32_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
    if (int_Rc < 0) 
    {
        LOG_INF("ICC_Function_Call Error %d \"%s\"", int_Rc, strerror(-int_Rc));
        u32_ResponseLength = sizeof(arr_s32_EncodedResponses);
        int_Rc = ICC_Function_Call(ial_ptr_IccCtx, IAL_SET_TOU_ID, u8_RequestBuffer, u32_RequestLength, 
                                    arr_s32_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
        if (int_Rc < 0) 
        {
            LOG_INF("ICC_Function_Call Retry Error %d \"%s\"", int_Rc, strerror(-int_Rc));
            return int_Rc;
        }
    }
    LOG_INF("Call for ToU Settings: %s %d data: %d, %u bytes", sz_PathPrefix, u8_ProfileID, int_Rc, u32_ResponseLength);
    return int_Rc;
}

GlobScop int IAL_Data_TestAlertStatusCallback(const char *sz_UnitType, uint8_t u8_UnitIndex)
{
    uint8_t u8_NewCodeQty = 2;
    const char *arr_sz_NewCodes[u8_NewCodeQty] = { "A1_2", "F2_7" };
    uint8_t u8_ClearedCodeQty = 2;
    const char *arr_sz_ClearedCodes[u8_ClearedCodeQty] = { "A1_3", "F2_8" };
    // Build the Request
    uint8_t u8_RequestBuffer[IAL_u16_ICC_REQUEST_BUFFER_SIZE];
    ZCBOR_STATE_E(CBORState, 1, u8_RequestBuffer, sizeof(u8_RequestBuffer), 1);
    // Encode Unit Type
    zcbor_tstr_encode_ptr(CBORState, sz_UnitType, strlen(sz_UnitType)+1);
    // Encode Unit ID
    zcbor_uint32_put(CBORState, (u8_UnitIndex));
    // Encode New codes
    zcbor_list_start_encode(CBORState, u8_NewCodeQty);
    for (uint8_t i = 0; i < u8_NewCodeQty; i++)
    {
        zcbor_tstr_encode_ptr(CBORState, arr_sz_NewCodes[i], strlen(arr_sz_NewCodes[i]) + 1);
    }
    zcbor_list_end_encode(CBORState, u8_NewCodeQty);
    // Encode Cleared codes
    zcbor_list_start_encode(CBORState, u8_ClearedCodeQty);
    for (uint8_t i = 0; i < u8_ClearedCodeQty; i++)
    {
        zcbor_tstr_encode_ptr(CBORState, arr_sz_ClearedCodes[i], strlen(arr_sz_ClearedCodes[i]) + 1);
    }
    zcbor_list_end_encode(CBORState, u8_ClearedCodeQty);

    // Send the Request
    uint32_t u32_RequestLength = (CBORState[0].payload - u8_RequestBuffer);
    FileScop uint8_t arr_s32_EncodedResponses[IAL_u16_ICC_RESPONSE_BUFFER_SIZE] = {0};
    uint32_t u32_ResponseLength = sizeof(arr_s32_EncodedResponses);
    int int_Rc = ICC_Function_Call(ial_ptr_IccCtx, IAL_ALERT_STATUS_ID, u8_RequestBuffer, u32_RequestLength, 
                                    arr_s32_EncodedResponses, &u32_ResponseLength, RTOS_SECONDS(IAL_u8_ICC_CALL_TIMEOUT_S));
    
    if (int_Rc < 0) {
        LOG_INF("ICC_Function_Call Error %d \"%s\"", int_Rc, strerror(-int_Rc));
        return int_Rc;
    }
    LOG_INF("AWS SEND ALERT TEST RESULT %d \"%s\"", int_Rc, strerror(-int_Rc));
    return int_Rc;
}

/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

FileScop size_t ial_data_BufferedDataSize(const char *sz_UnitType)
{
    if (strcmp(sz_UnitType, IAL_sz_UNIT_TYPE_ATS) == 0)
    {
        return sizeof(SCH_AtsParameters_t);
    }
    else if (strcmp(sz_UnitType, IAL_sz_UNIT_TYPE_DCDC) == 0)
    {
        return sizeof(SCH_DcdcParameters_t);
    }
    else if (strcmp(sz_UnitType, IAL_sz_UNIT_TYPE_EMS) == 0)
    {
        return sizeof(SCH_EmsParameters_t);
    }
    else if (strcmp(sz_UnitType, IAL_sz_UNIT_TYPE_GW) == 0)
    {
        return sizeof(SCH_GwParameters_t);
    }
    else if (strcmp(sz_UnitType, IAL_sz_UNIT_TYPE_HVAC) == 0)
    {
        return sizeof(SCH_HvacParameters_t);
    }
    else if (strcmp(sz_UnitType, IAL_sz_UNIT_TYPE_PCS) == 0)
    {
        return sizeof(SCH_PcsParameters_t);
    }
    else if (strcmp(sz_UnitType, IAL_sz_UNIT_TYPE_PDU) == 0)
    {
        return sizeof(SCH_PduParameters_t);
    }
    else if (strcmp(sz_UnitType, IAL_sz_UNIT_TYPE_PMU) == 0)
    {
        return sizeof(SCH_PmuParameters_t);
    }
    else if (strcmp(sz_UnitType, IAL_sz_TYPE_TOU) == 0)
    {
        return sizeof(SCH_EmsTouSettings_t);
    }
    else
    {
        return INT_MAX; //Invalid Unit Type, return very large number
    }
}

//EOF
